/*  This file is part of the KDE Libraries
 *  Copyright (C) 2006-2007 by Olivier Goffart (ogoffart at kde.org)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "kcolorsettingsitem.h"
#include "kcolorsettings.h"

#include <QColor>
#include <QFont>
#include <QPixmap>
#include <kconfigbase.h>

#include <kdebug.h>

class KColorSettingsItem::KColorSettingsItemPrivate : public QSharedData
{ public:
		QColor foregroundColor, backgroundColor;
		//QColor selectedFgColor, selectedBgColor;
		QFont font;
		//QPixmap bgImage;
		StyleFlags mask;
		const KColorSettingsItem *parent;
};


KColorSettingsItem::KColorSettingsItem()
	: d(new KColorSettingsItemPrivate)
{
	d->parent=0l;
}


KColorSettingsItem::KColorSettingsItem( const KColorSettingsItem &i )
	: d(i.d)
{
	//if we don't implement copy contructor and operator= , the compilation will fail
	// because the inline constructor require to know the structure of Private.
}

KColorSettingsItem & KColorSettingsItem::operator =( const KColorSettingsItem &i )
{
	this->d=i.d;
	return *this;
}



KColorSettingsItem::~KColorSettingsItem()
{
//	delete d;
}



KColorSettingsItem::StyleFlags KColorSettingsItem::mask() const
{
	return d->mask;
}

QColor KColorSettingsItem::foregroundColor() const
{
	if(d->mask & ForegroundColor)
		return d->foregroundColor;
	else if(d->parent)
		return d->parent->foregroundColor();
	else 
		return  QColor();
}


QColor KColorSettingsItem::backgroundColor() const
{
	if(d->mask & BackgroundColor)
		return d->backgroundColor;
	else if(d->parent)
		return d->parent->backgroundColor();
	else
		return QColor();
}
/*
QColor KColorSettingsItem::selectedForegroundColor() const
{
	if(d->mask & SelectedForegroundColor)
		return d->selectedFgColor;
	else if(d->parent)
		return d->parent->selectedForegroundColor();
	else
		return QColor();
}


QColor KColorSettingsItem::selectedBackgroundColor() const
{
	if(d->mask & SelectedBackgroundColor)
		return d->selectedBgColor;
	else if(d->parent)
		return d->parent->selectedBackgroundColor();
	else 
		return QColor();
}

QPixmap KColorSettingsItem::backgroundImage() const
{
	if(d->mask & BackgroundImage)
		return d->bgImage;
	else if(d->parent)
		return d->parent->backgroundImage();
	else
		return QPixmap();
}

*/

QFont KColorSettingsItem::font() const
{
	if(d->mask & Font)
		return d->font;
	else if(d->parent)
		return d->parent->font();
	else 
		return QFont();
}


void KColorSettingsItem::setForegroundColor(const QColor &value)
{
	if(!value.isValid())
		d->mask &= ~ForegroundColor;
	else
		d->mask |= ForegroundColor;
	d->foregroundColor = value;
}

void KColorSettingsItem::setBackgroundColor(const QColor &value)
{
	if(!value.isValid())
		d->mask &= ~BackgroundColor;
	else
		d->mask |= BackgroundColor;
	d->backgroundColor = value;
}

/*void KColorSettingsItem::setSelectedForegroundColor(const QColor &value)
{
	if(!value.isValid())
		d->mask &= ~SelectedForegroundColor;
	else
		d->mask |= SelectedForegroundColor;
	d->selectedFgColor = value;
}

void KColorSettingsItem::setSelectedBackgroundColor(const QColor &value)
{
	if(value.isValid())
		d->mask &= ~SelectedBackgroundColor;
	else
		d->mask |= SelectedBackgroundColor;
	d->selectedBgColor = value;
}



void KColorSettingsItem::setBackgroundImage(const QPixmap &value)
{
	if(value.isNull())
		d->mask &= ~BackgroundImage;
	else
		d->mask |= BackgroundImage;
	d->bgImage = value;
}
*/


void KColorSettingsItem::setFont(const QFont &value)
{
	if(value == QFont())
		d->mask &= ~Font;
	else
		d->mask |= Font;
	d->font = value;
}


void KColorSettingsItem::setParent(const KColorSettingsItem* parent)
{
	d->parent=parent;
}

const KColorSettingsItem* KColorSettingsItem::parent() const
{
	return d->parent;
}

void KColorSettingsItem::save(const QString &key , KConfigGroup *config) const
{
//	kDebug() << endl;
	config->writeEntry( key + "-ForegroundColor", d->foregroundColor);
	config->writeEntry(  key + "-BackgroundColor", d->backgroundColor);
//	config.writeEntry(  key + "SelectedForegroundColor", d->selectedFgColor);
//	config.writeEntry(  key + "SelectedBackgroundColor", d->selectedBgColor);
	config->writeEntry(  key + "Font", d->font);
//	config->writeEntry( key+ "-BackgroundImage", d->bgImage);
}


void KColorSettingsItem::load(const QString &key ,KConfigGroup *configLocal, KConfigGroup *configDef)
{
#define READ_ENTRY( ROLE , TYPE , GETROLE )  do { \
			TYPE c=configLocal->readEntry( key + "-" #ROLE ,  TYPE () ); \
			if(c == TYPE () && configDef) \
				c=configDef->readEntry( #ROLE,  TYPE () ); \
			if(c == TYPE () && configDef) \
			{ \
				QString parent=configDef->readEntry( "Parent" ,  QString() ); \
				if(!parent.isEmpty()) \
				{ \
					KColorSettings dset("kde"); \
					c = dset.color(parent) . GETROLE (); \
				} \
			} \
			set##ROLE ( c ); \
		} while(false)
	

	
	READ_ENTRY(ForegroundColor , QColor , foregroundColor);
	READ_ENTRY(BackgroundColor , QColor , backgroundColor);
//	READ_ENTRY(Font , QFont);
	
	
	
//	 setForegroundColor(  config->readEntry(  "ForegroundColor",  def.d->foregroundColor ) );
//	 setBackgroundColor(  config->readEntry(  "BackgroundColor",  def.d->backgroundColor ) );
//	 setSelectedForegroundColor(  config->readEntry(  key+ "-SelectedForegroundColor",  def.d->selectedFgColor ) );
//	 setSelectedBackgroundColor(  config->readEntry(  key+ "-SelectedBackgroundColor",  def.d->selectedBgColor ) );
//	 setFont(  config->readEntry(  key+ "Font",  def.d->font ) );
//	 d->bgImage = config->readEntry(  key+ "-BackgroundImage",  def.d->bgImage );
}


