/*  This file is part of the KDE project
    Copyright (C) 2006 Kevin Ottens <ervin@kde.org>
    Copyright (C) 2009 Dario Freddi <drf54321@gmail.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.

*/

#include "solidbackendchooser.h"

#include "ui_solidbackendchooser.h"

#include <kservicetypeprofile.h>
#include <kservicetype.h>
#include <kservicetypetrader.h>
#include <kconfig.h>
#include <QStringList>
#include <QListWidgetItem>
#include <kapplication.h>
#include <kicon.h>
#include <QList>
#include <QHash>
#include <QString>

#include <kservice.h>

static bool equals(const KService::List &list1, const KService::List &list2)
{
    if (list1.size() != list2.size()) {
        return false;
    }

    for (int i = 0; i < list1.size(); ++i) {
        if (list1[i]->name() != list2[i]->name()) {
            return false;
        }
    }

    return true;
}

class SolidBackendChooserPrivate
{

public:
    explicit SolidBackendChooserPrivate(const QString &b)
            : backendType(b) {}

    void loadServicesInView(const KService::List &offers);
    KService::List servicesFromView();
    void load();

    void slotSelectionChanged();
    void slotUpClicked();
    void slotDownClicked();

    Q_DECLARE_PUBLIC(SolidBackendChooser)
    SolidBackendChooser *q_ptr;

    Ui::BackendChooser ui;
    KService::List initServices;

    QString backendType;
    QHash<QString, KService::Ptr> services;
};


SolidBackendChooser::SolidBackendChooser(const QString &backendType, QWidget *parent)
        : QWidget(parent)
        , d_ptr(new SolidBackendChooserPrivate(backendType))
{
    Q_D(SolidBackendChooser);

    d->ui.setupUi(this);

    connect(d->ui.listView, SIGNAL(itemSelectionChanged()),
            this, SLOT(slotSelectionChanged()));
    connect(d->ui.upButton, SIGNAL(clicked()),
            this, SLOT(slotUpClicked()));
    connect(d->ui.downButton, SIGNAL(clicked()),
            this, SLOT(slotDownClicked()));

    KServiceType::Ptr service_type = KServiceType::serviceType(backendType);

    QString title = service_type->comment();

    if (title.isEmpty()) {
        title = backendType;
    }

    d->ui.groupBox->setTitle(title);

    d->load();
}

SolidBackendChooser::~SolidBackendChooser()
{
    delete d_ptr;
}

void SolidBackendChooser::setBackendType(const QString &type)
{
    Q_D(SolidBackendChooser);

    d->backendType = type;
    d->load();
    emit backendTypeChanged(d->backendType);
}

QString SolidBackendChooser::backendType() const
{
    Q_D(const SolidBackendChooser);

    return d->backendType;
}

void SolidBackendChooserPrivate::load()
{
    initServices = KServiceTypeTrader::self()->query(backendType, "Type == 'Service'");

    loadServicesInView(initServices);
}

void SolidBackendChooserPrivate::loadServicesInView(const KService::List &s)
{
    services.clear();
    ui.listView->clear();

    foreach(KService::Ptr service, s) {
        ui.listView->addItem(service->name());
        services[service->name()] = service;
    }

    ui.listView->setItemSelected(ui.listView->item(0), true);
}

KService::List SolidBackendChooserPrivate::servicesFromView()
{
    KService::List serv;

    int count = ui.listView->count();

    for (int i = 0; i < count; ++i) {
        QListWidgetItem *item = ui.listView->item(i);
        KService::Ptr service = services[item->text()];
        serv.append(service);
    }

    return serv;
}

void SolidBackendChooser::save()
{
    Q_D(SolidBackendChooser);

    if (!equals(d->initServices, d->servicesFromView())) {
        d->initServices = d->servicesFromView();
        KServiceTypeProfile::writeServiceTypeProfile(d->backendType, d->initServices);
        emit currentSelectionChanged(false);
    }
}

void SolidBackendChooser::defaults()
{
    Q_D(SolidBackendChooser);

    KService::List currentServices = d->servicesFromView();
    KService::List defaultServices = KServiceTypeTrader::self()->defaultOffers(d->backendType, "Type == 'Service'");

    if (!equals(currentServices, defaultServices)) {
        d->loadServicesInView(KServiceTypeTrader::self()->defaultOffers(d->backendType, "Type == 'Service'"));
        emit currentSelectionChanged(!equals(d->initServices, d->servicesFromView()));
    }
}

KService::Ptr SolidBackendChooser::selectedBackend()
{
    QString text = d->ui.listView->currentItem()->text();
    return services[text];
}

void SolidBackendChooserPrivate::slotSelectionChanged()
{
    KService::Ptr service;

    for (int i = 0; i < ui.listView->count(); ++i) {
        QListWidgetItem *item = ui.listView->item(i);

        if (ui.listView->isItemSelected(item)) {
            service = services[item->text()];
            break;
        }
    }

    if (service) {
        ui.iconLabel->setPixmap(KIcon(service->icon()).pixmap(32));
        ui.nameLabel->setText(service->name());
        ui.commentLabel->setText(service->comment());
        ui.versionLabel->setText(service->property("X-KDE-SolidBackendInfo-Version").toString());
    }
}

void SolidBackendChooserPrivate::slotUpClicked()
{
    Q_Q(SolidBackendChooser);

    QList<QListWidgetItem *> selectedList = ui.listView->selectedItems();

    foreach(QListWidgetItem *selected, selectedList) {
        int row = ui.listView->row(selected);

         if (row > 0) {
            QListWidgetItem *taken = ui.listView->takeItem(row - 1);
            ui.listView->insertItem(row, taken);

            emit q->currentSelectionChanged(!equals(initServices, servicesFromView()));
            return;
        }
    }
}

void SolidBackendChooserPrivate::slotDownClicked()
{
    Q_Q(SolidBackendChooser);

    QList<QListWidgetItem *> selectedList = ui.listView->selectedItems();

    foreach(QListWidgetItem *selected, selectedList) {
        int row = ui.listView->row(selected);

        if (row + 1 < ui.listView->count()) {
            QListWidgetItem *taken = ui.listView->takeItem(row + 1);
            ui.listView->insertItem(row, taken);

            emit q->currentSelectionChanged(!equals(initServices, servicesFromView()));
            return;
        }
    }
}

#include "solidbackendchooser.moc"
