/*
 * KDE Composite Manager, 
 * Copyright (C) 2003-2005 Fredrik Höglund <fredrik@kde.org>
 *
 * Based on xcompmgr,
 * Copyright (C) 2003 Keith Packard <keithp@keithp.com>
 *
 * Internal structure and naming style inspired by KWin,
 * Copyright (C) 1999, 2000 Matthias Ettrich <ettrich@kde.org>
 * Copyright (C) 2003 Lubos Lunak <l.lunak@kde.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Fredrik Höglund not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Fredrik Höglund makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * FREDRIK HÖGLUND DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL FREDRIK HÖGLUND BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <QRect>
#include <QTimer>

#include "workspace.h"
#include "workspace.moc"

#include "shadowpainter.h"
#include "client.h"
#include "predicates.h"
#include "utils.h"
#include "extensions.h"

#include "settings.h"
#include "animation.h"

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/extensions/Xcomposite.h>

#include <iostream>
#include <cmath>

Workspace *Workspace::sInst = NULL;

// Declared and interned in main.cpp
extern Atom net_wm_window_opacity;
extern Atom net_wm_window_type;
extern Atom net_wm_desktop;
extern Atom net_current_desktop;
extern Atom net_active_window;
extern Atom xa_esetroot_pmap_id;
extern Atom xa_xrootpmap_id;
extern Atom xa_xsetroot_id;

extern Display *dpy;


Workspace::Workspace()
	: QObject( 0, 0 ),
	  mActiveClient( NULL ), mDamage( None ), mRootTile( None ),
	  mWaitForClients( false ), mInitialRepaint( true )
{
	new KWinAdaptor( "org.kde.kwin", "/CMInterface", QDBus::sessionBus(), this );
    Workspace::sInst = this;

	XGrabServer( dpy );

	// Redirect all toplevel window contents to offscreen storage
	XCompositeRedirectSubwindows( dpy, rootId(), CompositeRedirectManual );

	// Initialize the client list
	initClientList();

	// Request notification about toplevel window state changes
	XSelectInput( dpy, rootId(),
			SubstructureNotifyMask | ExposureMask |
			StructureNotifyMask | PropertyChangeMask );

	XUngrabServer( dpy );

	// Find the picture format for the root window
	mFormat = XRenderFindVisualFormat( dpy, visual() );

	// Create an unclipped picture for drawing on the root window
	XRenderPictureAttributes pa;
	pa.subwindow_mode = IncludeInferiors;
	mFrontbuffer = XRenderCreatePicture( dpy, rootId(), format(), CPSubwindowMode, &pa );

	createBackbuffer();

	connect( &mAnimationTimer, SIGNAL( timeout() ), SLOT( updateAnimations() ) );
}


Workspace::~Workspace()
{
	// Stop the animation timer
	mAnimationTimer.stop();

	// Delete all running animations
	qDeleteAll( mAnimationList );
	mAnimationList.clear();

	// Delete all clients
	qDeleteAll( mClientList );
	mClientList.clear();

	releasePicture( mRootTile );
	releasePicture( mBackbuffer );
	releasePicture( mFrontbuffer );

	if ( mDamage != None )
		XFixesDestroyRegion( dpy, mDamage );
}


// Part of the DCOP interface -- called by the KCM when a setting has been
// changed by the user.
void Workspace::reconfigure()
{
	// Reparse the config file
	Settings::self()->config()->reparseConfiguration();
	Settings::self()->readConfig();

	foreach ( Client *client, mClientList )
		client->reconfigure();

	// Damage and repaint the whole desktop
	XserverRegion r = createRegionFromQRect( rect() );
	addDamage( r );

	if ( !XEventsQueued( dpy, QueuedAfterFlush ) )
	{
		repaint();
		XSync( dpy, false );
	}
}


void Workspace::initClientList()
{
	// Get the active window ID
	Window activeWindow = None;
	readWindowProperty( rootId(), net_active_window, &activeWindow );

	// Get a list of all toplevel windows from the X server, sorted bottom to top
	uint nwindows;
	Window root, parent, *windows;
	XQueryTree( dpy, rootId(), &root, &parent, &windows, &nwindows );

	// Create a client object for each window and insert it into the client
	// list, which is sorted top to bottom. (the opposite of the order returned
	// by XQueryTree()).
	for ( uint i = 0; i < nwindows; i++ )
	{
		XWindowAttributes attr;
		if ( !XGetWindowAttributes( dpy, windows[i], &attr ) )
			continue;

		// Create the client object for the window
		Client *client = new Client( windows[i], attr );

		if ( client->clientId() == activeWindow )
			mActiveClient = client;

		mClientList.prepend( client );
	}

	XFree( windows );

	// If we've identified the active client, tell it that it's focused
	if ( mActiveClient )
		mActiveClient->focusIn();
}


void Workspace::updateActiveClient()
{
	// Get the value of the _NET_ACTIVE_WINDOW property
	Window activeWindow = None;
	readWindowProperty( rootId(), net_active_window, &activeWindow );

	// If the active window was changed
	if ( ( mActiveClient && mActiveClient->clientId() != activeWindow )
			  || ( !mActiveClient && activeWindow != None ) )
	{
		// Tell the currently active client it's lost focus
		if ( mActiveClient )
			mActiveClient->focusOut();

		mActiveClient = NULL;

		// If a new window has been focused, try to find it in the client list.
		if ( activeWindow != None )
		{
			mActiveClient = findClient( ClientIdMatchPredicate( activeWindow ) );

			// If we found the window, tell the client it's been focused
			if ( mActiveClient )
				mActiveClient->focusIn();
		}
	}
}


void Workspace::createBackbuffer()
{
	Pixmap pixmap = XCreatePixmap( dpy, rootId(), width(), height(), depth() );
	mBackbuffer = XRenderCreatePicture( dpy, pixmap, format(), 0, 0 );
	XFreePixmap( dpy, pixmap ); // The picture owns the pixmap now
}


void Workspace::deleteClient( Client *client )
{
	QMutableLinkedListIterator<Client *> i( mClientList );
	if ( i.findNext( client ) )
		i.remove();

	delete client;
}


// Finds a client using a predicate function
template <typename T>
Client *Workspace::findClient( T predicate ) const
{
	foreach ( Client *client, mClientList )
		if ( predicate( client ) )
			return client;

	return NULL;
}


void Workspace::restack( Client *client, Window belowClient )
{
	// ### Should we deal with the possibility that belowClient isn't in the list?

	// Unlink the client from the client list
	QMutableLinkedListIterator<Client*> i( mClientList );
	if ( i.findNext( client ) )
		i.remove();

	if ( belowClient != None )
	{
		// Search the client list bottom->top for belowClient
		i.toBack();
		while ( i.hasPrevious() )
		{
			if ( i.previous()->winId() == belowClient )
			{
				// The iterator should now be pointing at the position immediately
				// above belowClient, which is where we'll reinsert client.
				i.insert( client );
				break;
			}
		}
	} else
		// The window was moved to the bottom of the stacking order
		mClientList.append( client );
}


// Adds the region to the total damage for the workspace and destroys the region
void Workspace::addDamage( XserverRegion &region )
{
	if ( mDamage ) {
		XFixesUnionRegion( dpy, mDamage, mDamage, region );
		XFixesDestroyRegion( dpy, region );
	} else
		mDamage = region;

	region = None;
}


void Workspace::createRootTile()
{
	Pixmap pixmap = None;

	// Try to find a root window property with a background pixmap ID
	Atom properties[] = { xa_xrootpmap_id, xa_esetroot_pmap_id, xa_xsetroot_id };
	for ( int i = 0; i < 3; i++ )
		if ( readPixmapProperty( rootId(), properties[i], &pixmap ) )
			break;

	XRenderPictureAttributes pa;
	pa.repeat = true;

	// Create the root tile picture for the pixmap, if one was found
	if ( pixmap != None ) {
		mRootTile = XRenderCreatePicture( dpy, pixmap, format(), CPRepeat, &pa );
	} else {
		// If no background pixmap is set, create a simple 1x1 tile
		pixmap    = XCreatePixmap( dpy, rootId(), 1, 1, depth() );
		mRootTile = XRenderCreatePicture( dpy, pixmap, format(), CPRepeat, &pa );
		XFreePixmap( dpy, pixmap );

		// Fill it with a gray color
		XRenderColor col = { 0x8080, 0x8080, 0x8080, 0xffff };
		XRenderFillRectangle( dpy, PictOpSrc, mRootTile, &col, 0, 0, 1, 1 );
	}
}


bool Workspace::allVisibleClientsPainted() const
{
	foreach ( Client *client, mClientList )
	{
		if ( !client->isVisible() || client->isPainted() )
			continue;

		// We only care about windows that are inside the workspace boundaries
		if ( !client->geometry( WindowAndBorder ).intersects( rect() ) )
			continue;

		return false;
	}

	return true;
}


// This function creates the screen presentation, by drawing the root window background
// and compositing all top-level windows onto the root window.
void Workspace::repaint()
{
	// If there's no damage, update the whole display
	if ( mDamage == None || mInitialRepaint )
	{
		if ( mDamage )
			XFixesDestroyRegion( dpy, mDamage );

		mDamage = createRegionFromQRect( rect() );
		mInitialRepaint = false;
	}

	// Use the damage region as the clip region for the root window
	XFixesSetPictureClipRegion( dpy, frontBuffer(), 0, 0, mDamage );

	// List of clients that are either translucent or have a shadow
	QList<Client *> translucents;

	// Draw each opaque window top to bottom, subtracting the bounding rect of
	// each window from the clip region after it's been drawn.
	foreach ( Client *client, mClientList )
	{
		if ( !client->isVisible() || !client->isPainted() )
			continue;

		// Update the region with the area the window was last rendered at.
		client->updateOnScreenRegion();

		// Only draw the window if it's opaque
		if ( client->isOpaque() ) {
			// Set the clip region for the backbuffer to the damage region, and
			// subtract the clients shape from the damage region
			XFixesSetPictureClipRegion( dpy, backBuffer(), 0, 0, mDamage );
			XFixesSubtractRegion( dpy, mDamage, mDamage, client->shape() );

			XRenderComposite( dpy, PictOpSrc, client->picture(),
					None, backBuffer(), 0, 0, 0, 0,
					client->x(), client->y(),
					client->width() + client->borderWidth() * 2,
					client->height() + client->borderWidth() * 2 );
		}

		// Save the clip region before the next client shape is subtracted from it.
		// We need to restore it later when we're drawing the shadow.
		client->saveClipRegion( mDamage );
		translucents.prepend( client );
	}

	// Fill any areas of the root window not covered by windows with the root tile
	XFixesSetPictureClipRegion( dpy, backBuffer(), 0, 0, mDamage );
	XRenderComposite( dpy, PictOpSrc, rootTile(), None, backBuffer(),
					  0, 0, 0, 0, 0, 0, width(), height() );

	// Now walk the list backwards, drawing translucent windows and shadows.
	// That we draw bottom to top is important now since we're drawing translucent objects.
	foreach ( Client *client, translucents )
	{
		// Restore the previously saved clip region
		XFixesSetPictureClipRegion( dpy, backBuffer(), 0, 0, client->savedClipRegion() );

		// Draw a drop shadow if the window doesn't have an alpha channel
		// and isn't a desktop window.
		if ( Settings::drawShadows() && !client->hasAlphaChannel()
				   && client->windowType() != NET::Desktop )
		{
			//ShadowPainter::drawShadow( backBuffer(), client->geometry( Shadow ),
			//							 client->shadowRadius(), client->opacity() );
			ShadowPainter::drawShadowPicture( backBuffer(), client->geometry( Shadow ),
										 client->shadow(), client->opacity() );
		}

		// Only draw the window if it's translucent
		// (we drew the opaque ones in the previous loop)
		if ( !client->isOpaque() )
			XRenderComposite( dpy, PictOpOver, client->picture(),
				    client->alphaMask(), backBuffer(), 0, 0, 0, 0,
					client->x(), client->y(),
					client->width() + client->borderWidth() * 2,
					client->height() + client->borderWidth() * 2 );

		// We don't need the clip region anymore
		client->destroySavedClipRegion();
	}

	translucents.clear();

	// Destroy the damage region
	XFixesDestroyRegion( dpy, mDamage );
	mDamage = None;

	// Copy the back buffer contents to the root window
	XFixesSetPictureClipRegion( dpy, backBuffer(), 0, 0, None );
	XRenderComposite( dpy, PictOpSrc, backBuffer(), None, frontBuffer(),
					  0, 0, 0, 0, 0, 0, width(), height() );
}


// Searches the animation list for an animation for the specified client
Animation *Workspace::findAnimation( const Client *client )
{
	foreach ( Animation *animation, mAnimationList )
		if ( animation->client() == client )
			return animation;

	return NULL;
}


// Creates an animation of the specified type for a client, reusing an existing one
// if possible.
void Workspace::createAnimation( Client *client, AnimationType type, uint flags )
{
	// Try to find an existing animation for the client
	Animation *animation = findAnimation( client );
	if ( animation )
	{
		// If we're asked to create a FadeOut animation, and a FadeOut animation is
		// already running, just make sure all the flags are set and leave it running.
		// The scenario here is that the client was unmapped, and then destroyed before
		// the resulting FadeOut animation was completed.
		if ( animation->type() == type && type == FadeOut ) {
			animation->setFlags( flags );
			return;
		}

		// If an unmapped client is mapped again before its FadeOut animation is
		// finished, the clients window pixmap is stale and has to be renewed.
		// 
		// Unfortunately renewing it at this point may result in visible flicker if
		// the client hasn't finished painting the new window pixmap by the time we
		// enter the paint cycle.
		// 
		// The alternative would be to assume that the window will look exactly the
		// same as it did when it was unmapped, and continue using the stale window
		// pixmap until the animation is finished, and only then renew it. Doing so
		// would give the client more time to paint the new window pixmap.
		if ( animation->type() == FadeOut && type == FadeIn )
			client->renewWindowPixmap();

		animation->setType( type );
		animation->clearFlags( AllFlags );
		animation->setFlags( flags );
		return;
	}

	if ( client->width() * client->height() > (384 * 384) )
	{
		client->setAnimationOpacity( 1.0 );

		if ( flags & HideClientWhenDone )
			client->hide();

		if ( flags & DeleteClientWhenDone )
			deleteClient( client );

		return;
	}

	// Create the new animation, and enqueue it
	animation = new Animation( client, type, flags );
	mAnimationList.append( animation );

	// Start the animation timer if it isn't already running
	if ( !mAnimationTimer.isActive() )
		mAnimationTimer.start( 1000 / Settings::framesPerSec() );
}


void Workspace::updateAnimations()
{
	QMutableLinkedListIterator<Animation*> i( mAnimationList );
	while ( i.hasNext() )
	{
		Animation *animation = i.next();
		animation->update();

		if ( animation->isDone() ) {
			i.remove();
			delete animation;
		}
	}

	// Stop the timer if all animations have finished
	if ( mAnimationList.count() == 0 )
		mAnimationTimer.stop();

	// Update the display if the X event queue is empty
	if ( mDamage && QLength( dpy ) == 0 )
		repaint();
}


void Workspace::deleteAnimation( Animation *animation )
{
	QMutableLinkedListIterator<Animation*> i( mAnimationList );
	if ( i.findNext( animation ) )
		i.remove();

	delete animation;

	if ( mAnimationList.count() == 0 )
		mAnimationTimer.stop();
}


void Workspace::createNotifyEvent( const XCreateWindowEvent *event )
{
	XWindowAttributes attr;
	if ( !XGetWindowAttributes( dpy, event->window, &attr ) )
		return;

	// ### Workaround for a QApplication bug:
	// QApplication swallows the reparent event that occurs when a kwin style
	// creates the decoration QWidget and kwin reparents it from the root window
	// to the frame window, thus preventing us from seeing the event.
	//
	// We work around this by adding a synchronous X call to check if the parent
	// window still is the root window, before creating the client object. The
	// decoration widget is always created as initially visible, which allows for
	// a small optimization. 
	if ( attr.map_state == IsViewable && getParentWindow( event->window ) != rootId() )
		return;

	// Create a client for the the new window, and insert it at the top of the
	// stacking order
	Client *client = new Client( event->window, attr );
	mClientList.prepend( client );

	if ( !client->isInputOnly() && client->isVisible() && Settings::fadeWindows() )
		createAnimation( client, FadeIn );
}


void Workspace::destroyNotifyEvent( const XDestroyWindowEvent *event )
{
	Client *client = findClient( WindowIdMatchPredicate( event->window ) );
	if ( client )
	{
		if ( Settings::fadeWindows() && !client->isInputOnly() && client->isVisible() )
			createAnimation( client, FadeOut, DeleteClientWhenDone );
		else
			deleteClient( client );
	}
}


void Workspace::mapNotifyEvent( const XMapEvent *event )
{
	Client *client = findClient( WindowIdMatchPredicate( event->window ) );
	if ( client )
	{
		client->show();

		if ( Settings::fadeWindows() && !client->isInputOnly() )
			createAnimation( client, FadeIn );
	}
}


void Workspace::unmapNotifyEvent( const XUnmapEvent *event )
{
	Client *client = findClient( WindowIdMatchPredicate( event->window ) );
	if ( client )
	{
		if ( Settings::fadeWindows() && !client->isInputOnly() )
			createAnimation( client, FadeOut, HideClientWhenDone );
		else
			client->hide();
	}
}


// Move, resize, or stacking order change
void Workspace::configureNotifyEvent( const XConfigureEvent *event )
{
	Client *client = findClient( WindowIdMatchPredicate( event->window ) );

	if ( client ) {
		QRect geometry = QRect( event->x, event->y, event->width, event->height );
		int borderWidth = event->border_width;
		Window above = event->above;

		// Compress configure events. There's no need to union each of these rects with
		// the damage region, since there are only two visibly affected rects; the rect
		// where the window is currently drawn, and the rect in the last configure event,
		// which is where we'll draw the window in the next paint cycle.
		XConfigureEvent e;
		while ( XCheckTypedWindowEvent( dpy, client->winId(), ConfigureNotify, (XEvent*)&e ) )
		{
			geometry = QRect( e.x, e.y, e.width, e.height );
			borderWidth = e.border_width;
			above = e.above;
		}

		client->geometryChanged( geometry, borderWidth );
		restack( client, above );
	} else if ( event->window == rootId() ) {
		// Recreate the back buffer
		releasePicture( mBackbuffer );
		createBackbuffer();
	}
}


void Workspace::shapeNotifyEvent( const XShapeEvent *event )
{
	// Notify the client that its shape has changed
	Client *client = findClient( WindowIdMatchPredicate( event->window ) );
	if ( client )
		client->shapeChanged();
}


void Workspace::reparentNotifyEvent( const XReparentEvent *event )
{
	if ( event->parent == rootId() )
	{
		// When a window is reparented to the root window we treat it the
		// same as if it was just created (but we don't animate it).
		XWindowAttributes attr;
		if ( !XGetWindowAttributes( dpy, event->window, &attr ) )
			return;

		Client *client = new Client( event->window, attr );
		mClientList.prepend( client );
	} else
	{
		// If a client is reparented out of the root window, delete it.
		Client *client = findClient( WindowIdMatchPredicate( event->window ) );
		if ( client )
		{
			if ( Settings::fadeWindows() )
			{
				// If an animation is running for the client, abort it
				Animation *animation = findAnimation( client );
				if ( animation ) {
					animation->clearFlags( AllFlags );
					deleteAnimation( animation );
				}
			}

			deleteClient( client );
		}
	}
}


void Workspace::circulateNotifyEvent( const XCirculateEvent *event )
{
	Client *client = findClient( WindowIdMatchPredicate( event->window ) );
	if ( client ) {
		Window above = ( event->place == PlaceOnTop ) ? mClientList.first()->winId() : None;
		restack( client, above );
	}
}


// Handles expose events for the root window
// (We recieve these since the root window itself isn't redirected)
void Workspace::exposeEvent( const XExposeEvent *event )
{
	if ( event->window != rootId() )
		return;

	// Append the rectangle to the list of exposed rectangles
	XRectangle r = { event->x, event->y, event->width, event->height };
	mExposeRects.append( r );

	// If there are no additional expose events, create a damage region from the
	// exposed rects, and add it to the workspace's damage.
	if ( event->count == 0 ) {
		XserverRegion damage = XFixesCreateRegion( dpy, mExposeRects.data(),
		                                    mExposeRects.count() );
		addDamage( damage ); // Destroys damage
		mExposeRects.clear();
	}
}


void Workspace::damageNotifyEvent( const XDamageNotifyEvent *event )
{
	Client *client = findClient( WindowIdMatchPredicate( event->drawable ) );

	// ### should never happen
	if ( !client )
		return;

	XserverRegion parts;

	if ( !client->isPainted() ) {
		// If the window has been painted for the first time, mark the whole
		// window + shadow as damaged.
		parts = client->createRegion( WindowAndShadow );
		XDamageSubtract( dpy, client->damageHandle(), None, None );
		client->setPainted( true );
	} else {
		// Create a new empty region
		parts = XFixesCreateRegion( dpy, 0, 0 );

		// Copy the damage region to parts, subtracting it from the clients damage
		XDamageSubtract( dpy, client->damageHandle(), None, parts );

		// Offset parts with the clients position
		XFixesTranslateRegion( dpy, parts, client->x(), client->y() );
	}

	// Add the damaged region to the total damage for the workspace
	// (destroys parts)
	addDamage( parts );
}


struct PropertyDescription
{
	PropertyDescription( Window w, Atom *props, int nprops )
		: window( w ), atoms( props ), natoms( nprops ) {}
	Window window;
	Atom *atoms;
	int natoms;
};

static Bool propertyChangePredicate( Display *, XEvent *event, XPointer arg )
{
	PropertyDescription *desc = reinterpret_cast<PropertyDescription*>( arg );
	if ( event->type == PropertyNotify && event->xproperty.window == desc->window )
	{
		for ( int i = 0; i < desc->natoms; i++ )
			if ( event->xproperty.atom == desc->atoms[i] )
				return True;
	}

	return False;
}

// Discards property change events in the event queue, for the specified window
// and properties
void Workspace::discardPropertyEvents( Window window, Atom *properties, int nproperties )
{
	PropertyDescription desc( window, properties, nproperties );
	XPointer arg = reinterpret_cast< XPointer >( &desc );

	XEvent dummy;
	while ( XCheckIfEvent( dpy, &dummy, propertyChangePredicate, arg ) );
}


void Workspace::propertyNotifyEvent( const XPropertyEvent *event )
{
	//char *name = XGetAtomName( dpy, event->atom );
	//std::cout << "property \"" << name << "\" changed.\n";
	//XFree( name );


	// Handle property events for the root window
	// ------------------------------------------------------------------------
	if ( event->window == rootId() )
	{
		// If the active window has been changed
		if ( event->atom == net_active_window )
		{
			// Compress these events
			discardPropertyEvents( rootId(), &net_active_window );
			updateActiveClient();
		}

		// Desktop switching optimization; we'll block updates and just let damage
		// accumulate until all newly mapped windows have been painted.
		else if ( event->atom == net_current_desktop && !Settings::fadeWindows() )
		{
			mWaitForClients = true;
			XSync( dpy, false ); // Helps to accumulate more events
		}

		// If the root window background pixmap has been changed
		else if ( event->atom == xa_xrootpmap_id ||
		          event->atom == xa_xsetroot_id  ||
		          event->atom == xa_esetroot_pmap_id )
		{
			// Compress these events
			Atom properties[] = { xa_xrootpmap_id, xa_xsetroot_id, xa_esetroot_pmap_id };
			discardPropertyEvents( rootId(), properties, 3 );

			// Destroy the root tile picture.
			// (it will be recreated for the new pixmap in the next paint cycle)
			releasePicture( mRootTile );

			// Damage the whole display
			XserverRegion region = createRegionFromQRect( rect() );
			addDamage( region );
		}
	}

	// Handle property events for clients
	// ------------------------------------------------------------------------
	else
	{
		// If an opacity hint has been changed
		if ( event->atom == net_wm_window_opacity )
		{
			// Compress these events
			discardPropertyEvents( event->window, &net_wm_window_opacity );

			// We handle opacity change events for both the frame and the embedded window
			Client *client = findClient( ClientOrWindowIdMatchPredicate( event->window ) );
			if ( client ) {
				client->setOpacityHintValid( false );

				// Damage the region the window occupies
				XserverRegion region = client->createRegion( WindowAndShadow );
				addDamage( region );
			}
		}

		// If the window type for a client has been changed
		else if ( event->atom == net_wm_window_type )
		{
			// We won't bother to compress these events, since we only mark the
			// window type as dirty. The window type will be updated the next time
			// it's requested, which will be when the window is drawn if it's
			// visible, or when it's mapped if it isn't.
			Client *client = findClient( ClientIdMatchPredicate( event->window ) );
			if ( client )
				client->setWindowTypeValid( false );
		}

		// If the transient-for hint for a client has been changed
		else if ( event->atom == XA_WM_TRANSIENT_FOR )
		{
			Client *client = findClient( ClientIdMatchPredicate( event->window ) );
			if ( client ) {
				Window transient;
				XGetTransientForHint( dpy, client->clientId(), &transient );
				client->setTransient( transient );
			}
		}

		// If a window has been moved to another desktop
		else if ( event->atom == net_wm_desktop )
		{
			Client *client = findClient( ClientIdMatchPredicate( event->window ) );
			if ( client ) {
				int desktop;
				if ( readCardinalProperty( client->clientId(), net_wm_desktop, &desktop ) )
					client->setDesktop( desktop );
			}
		}
	}
}


bool Workspace::x11Event( XEvent *event )
{
	switch ( event->type )
	{
		case CreateNotify:
			createNotifyEvent( &event->xcreatewindow );
			break;

		case DestroyNotify:
			destroyNotifyEvent( &event->xdestroywindow );
			break;

		case ConfigureNotify:
			configureNotifyEvent( &event->xconfigure );
			break;

		case MapNotify:
			mapNotifyEvent( &event->xmap );
			break;

		case UnmapNotify:
			unmapNotifyEvent( &event->xunmap );
			break;

		case ReparentNotify:
			reparentNotifyEvent( &event->xreparent );
			break;

		case CirculateNotify:
			circulateNotifyEvent( &event->xcirculate );
			break;

		case Expose:
			exposeEvent( &event->xexpose );
			break;

		case PropertyNotify:
			propertyNotifyEvent( &event->xproperty );
			break;

		default:
			if ( event->type == Extensions::damageEvent( XDamageNotify ) )
				damageNotifyEvent( reinterpret_cast< XDamageNotifyEvent * >( event ) );

			else if ( event->type == Extensions::shapeEvent( ShapeNotify ) )
				shapeNotifyEvent( reinterpret_cast< XShapeEvent * >( event ) );
	}


	// Update the root window if there's damage and there are no more events pending
	if ( mDamage && QLength( dpy ) == 0 )
	{
		if ( mWaitForClients && !allVisibleClientsPainted() )
		{
			//std::cout << "delaying update.\n.";
		} else {
			mWaitForClients = false;
			repaint();
		}

		// Helps to accumulate more events
		XSync( dpy, false );
	}

	return false;
}


