/*
 * KDE Composite Manager, 
 * Copyright (C) 2003-2005 Fredrik Höglund <fredrik@kde.org>
 *
 * Based on xcompmgr,
 * Copyright (C) 2003 Keith Packard <keithp@keithp.com>
 *
 * Internal structure and naming style inspired by KWin,
 * Copyright (C) 1999, 2000 Matthias Ettrich <ettrich@kde.org>
 * Copyright (C) 2003 Lubos Lunak <l.lunak@kde.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Fredrik Höglund not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Fredrik Höglund makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * FREDRIK HÖGLUND DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL FREDRIK HÖGLUND BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __WORKSPACE_H
#define __WORKSPACE_H

#include <QLinkedList>
#include <QVector>
#include <QTimer>
#include <QObject>

#include "interface.h"
#include "globaldefs.h"

#include <X11/Xlib.h>
#include <X11/extensions/Xdamage.h>
#include <X11/extensions/Xfixes.h>
#include <X11/extensions/Xrender.h>
#include <X11/extensions/shape.h>

class Client;
class Animation;

extern Display *dpy;

class Workspace : public QObject, virtual public CMInterface
{
	Q_OBJECT	

	public:
		static inline Workspace *instance();
		~Workspace();

		void reconfigure();

		int width() const { return DisplayWidth( dpy, DefaultScreen(dpy) ); }
		int height() const { return DisplayHeight( dpy, DefaultScreen(dpy) ); }
		Picture frontBuffer() const { return mFrontbuffer; }
		Picture backBuffer() const { return mBackbuffer; }
		Visual *visual() const { return DefaultVisual( dpy, DefaultScreen(dpy) ); }
		XRenderPictFormat *format() const { return mFormat; }
		int depth() const { return DefaultDepth( dpy, DefaultScreen(dpy) ); }
		Window rootId() const { return RootWindow( dpy, DefaultScreen(dpy) ); }
		QRect rect() const { return QRect( 0, 0, width(), height() ); }
		Pixmap rootTile();
	
		template <typename T>
		Client *findClient( T predicate ) const;

		void deleteClient( Client *client );
		bool damaged() const { return mDamage != None; }
		void addDamage( XserverRegion & );
		bool allVisibleClientsPainted() const;
		void repaint();

		bool x11Event( XEvent * );

	private:
		Workspace();
		void restack( Client *, Window );
		void createRootTile();
		void createBackbuffer();
		void initClientList();
		void updateActiveClient();
		Animation *findAnimation( const Client * );
		void createAnimation( Client *, AnimationType, uint flags = NoFlags );
		void deleteAnimation( Animation * );
		void discardPropertyEvents( Window, Atom *, int natoms = 1 );

		void createNotifyEvent( const XCreateWindowEvent * );
		void destroyNotifyEvent( const XDestroyWindowEvent * );
		void mapNotifyEvent( const XMapEvent * );
		void unmapNotifyEvent( const XUnmapEvent * );
		void configureNotifyEvent( const XConfigureEvent * );
		void shapeNotifyEvent( const XShapeEvent * );
		void reparentNotifyEvent( const XReparentEvent * );
		void circulateNotifyEvent( const XCirculateEvent * );
		void exposeEvent( const XExposeEvent * );
		void damageNotifyEvent( const XDamageNotifyEvent * );
		void propertyNotifyEvent( const XPropertyEvent * );

	private slots:
		void updateAnimations();

	private:
		static Workspace        *sInst;

		XRenderPictFormat       *mFormat;
		Client                  *mActiveClient;
		QLinkedList<Client *>    mClientList;
		QLinkedList<Animation *> mAnimationList;
		int                      mCurrentDesktop;
		Damage                   mDamage;
		Picture                  mBackbuffer, mFrontbuffer;
		Picture                  mRootTile;
		QVector<XRectangle>      mExposeRects;
		QTimer                   mAnimationTimer;
		bool                     mUpdatePending:1;
		bool                     mWaitForClients:1;
		bool                     mInitialRepaint:1;
};


Workspace *Workspace::instance()
{
	if ( !sInst )
		new Workspace;

	return sInst;
}

inline Picture Workspace::rootTile()
{
	if ( !mRootTile )
		createRootTile();

	return mRootTile;
}

#endif

