/*
 *   Copyright © 2009 Dmitry Suzdalev (dimsuz@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#ifndef KLIPBOARD_H
#define KLIPBOARD_H

#include <QObject>

#include "klipper_export.h"

#include "clipboardtypes.h"

class KlipboardPrivate;
class KlipboardSigletonPrivate;

/**
 * This is a central class which polls for clipboard changes
 * and accumulates clipboard history within one session
 * (i.e. it doesn't store clipboard between runs - application
 * using this class should do this manually if it wishes
 */
class KLIPPER_EXPORT Klipboard : public QObject
{
    Q_OBJECT
public:
    static Klipboard* self();

    void setHistorySize(int size);
    int historySize() const;

    /**
     * Sets the name of the file to hold the clipboard history.
     * If the file exists, history will be loaded from it
     */
    void setHistoryFile(const QString& filename);

    int entryCount() const;

    bool isHistoryEmpty() const;

    void clearHistory();

    /**
     * Moves entry with specific uuid to the top of the history list
     */
    void setCurrentEntry(const QString& uuid);

    /**
     * Returns an entry from the top of the history list
     */
    ClipboardEntry currentEntry() const;

    /**
     * Returns an entry with the specified uuid
     */
    ClipboardEntry entry(const QString& uuid) const;

    /**
     * Returns an entry at the given index.
     * Index 0 - means least recently added entry
     * Index 'entryCount()-1' - will be the newest entry
     */
    ClipboardEntry entry(int index) const;


    /**
     * Fills clipboard history with entries from @p entryList.
     * Entries are added to top of the history, possibly moving out old ones.
     * (if historySize is too small to hold current entries + newly added ones)
     *
     * I.e. if list of "a", "b", "c" gets added, entry(0) will return "c" - the entry on top
     *
     * Uuids for new entries are regenerated during addition
     */
    void populateHistory(const ClipboardEntryList& entryList);

signals:
    void entryAdded(const QString& uuid);
    void entryRemoved(const QString& uuid);
    void currentEntryChanged(const QString& uuid);

private slots:
    void onClipboardChanged(bool selectionMode);

private:
    friend class KlipboardSingletonPrivate;
    Klipboard();
    Klipboard(const Klipboard&);
    ~Klipboard();

    KlipboardPrivate * const d;
    friend class KlipboardPrivate;
};

#endif
