/*
 *   Copyright © 2009 Dmitry Suzdalev (dimsuz@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "klipboardunittest.h"
#include "klipboard.h"
#include "historyfile.h"
#include "testutils.h"

#include <QClipboard>

static int signalWaitTimeout = 500;

void tst_Klipboard::initTestCase()
{
    Klipboard::self();
}

void tst_Klipboard::cleanupTestCase()
{

}

void tst_Klipboard::clipEntryTest()
{
    ClipboardEntry entry;

    QVERIFY(!entry.data.isValid());
    QVERIFY(entry.uuid.isNull());
    QVERIFY(entry.isText == true);
    QVERIFY(entry.isSelection == false);

    // copy & assignment test
    ClipboardEntry entry1(entry);
    QCOMPARE(entry1.data, entry.data);
    QCOMPARE(entry1.uuid, entry.uuid);
    QCOMPARE(entry1.isText, entry.isText);
    QCOMPARE(entry1.isSelection, entry.isSelection);

    ClipboardEntry entry2;
    entry2 = entry;
    QCOMPARE(entry2.data, entry.data);
    QCOMPARE(entry2.uuid, entry.uuid);
    QCOMPARE(entry2.isText, entry.isText);
    QCOMPARE(entry2.isSelection, entry.isSelection);
}

void tst_Klipboard::copyText()
{
    Klipboard::self()->clearHistory();
    kapp->clipboard()->setText("Test", QClipboard::Clipboard);
    QTest::qWait(signalWaitTimeout);
    QCOMPARE(Klipboard::self()->currentEntry().data.toString(), QString("Test"));
    QVERIFY(Klipboard::self()->currentEntry().isSelection == false );

    kapp->clipboard()->setText("Test", QClipboard::Selection);
    QTest::qWait(signalWaitTimeout);
    QCOMPARE(Klipboard::self()->currentEntry().data.toString(), QString("Test"));
    QVERIFY(Klipboard::self()->currentEntry().isSelection == true );
}

void tst_Klipboard::selectText()
{
}

void tst_Klipboard::populateHistory()
{
    ClipboardEntryList list = makeEntryList(100);
    Klipboard::self()->clearHistory();
    QVERIFY(Klipboard::self()->isHistoryEmpty());

    /*
     * 1. List size equals history size
     */
    Klipboard::self()->setHistorySize(100);
    Klipboard::self()->populateHistory(list);

    QCOMPARE(Klipboard::self()->entryCount(), list.count());
    QCOMPARE(Klipboard::self()->historySize(), 100);

    // verify results
    for (int i=0; i<Klipboard::self()->entryCount(); ++i)
    {
        ClipboardEntry entry = Klipboard::self()->entry(i);
        QCOMPARE(entry.data, list.at(i).data);
        QCOMPARE(entry.isText, list.at(i).isText);
        QCOMPARE(entry.isSelection, list.at(i).isSelection);
        QVERIFY(!entry.uuid.isEmpty());
        i++;
    }

    /*
     * 2. Adding when there are already some items
     */
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(110);
    QVERIFY(Klipboard::self()->isHistoryEmpty());

    // first put some entries
    ClipboardEntryList tmpList;
    ClipboardEntry entry1, entry2, entry3;
    entry1.data = "TestEntry1";
    entry2.data = "TestEntry2";
    entry3.data = "TestEntry3";
    tmpList << entry1 << entry2 << entry3;
    Klipboard::self()->populateHistory(tmpList);
    Klipboard::self()->populateHistory(list);

    QCOMPARE(Klipboard::self()->entryCount(), tmpList.count()+list.count());
    QCOMPARE(Klipboard::self()->entry(0).data, entry1.data);
    QCOMPARE(Klipboard::self()->entry(1).data, entry2.data);
    QCOMPARE(Klipboard::self()->entry(2).data, entry3.data);
    QCOMPARE(Klipboard::self()->entry(3).data, list.at(0).data);
    QCOMPARE(Klipboard::self()->entry(4).data, list.at(1).data);

    /*
     * 3. Adding when there are already some items, but shifting existing ones is needed
     */
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(101); // two items will need to be dropped from history
    QVERIFY(Klipboard::self()->isHistoryEmpty());

    // first put some entries
    Klipboard::self()->populateHistory(tmpList);
    // after this it should overflow and remove first two entries
    Klipboard::self()->populateHistory(list);

    int entryCount = Klipboard::self()->entryCount();
    // it shouldn't grow more than allowed
    QCOMPARE(entryCount, Klipboard::self()->historySize());
    // entry1 and entry2 should be popped out
    QCOMPARE(Klipboard::self()->entry(0).data, entry3.data);
    QCOMPARE(Klipboard::self()->entry(1).data, list.at(0).data);
    QCOMPARE(Klipboard::self()->entry(entryCount-1).data, list.last().data);

    /*
     * 4. Adding with complete popping out existing history and part of added one
     */
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(10);
    QVERIFY(Klipboard::self()->isHistoryEmpty());

    // first put some entries
    Klipboard::self()->populateHistory(tmpList);
    // after this it should overflow
    Klipboard::self()->populateHistory(list);

    entryCount = Klipboard::self()->entryCount();
    // it shouldn't grow more than allowed
    QCOMPARE(Klipboard::self()->entryCount(), Klipboard::self()->historySize());
    // only last 10 items from list should stay
    QCOMPARE(Klipboard::self()->entry(0).data, list.at(90).data );
    QCOMPARE(Klipboard::self()->entry(entryCount-1).data, list.at(99).data);

    /*
     * Appending an empty list
     */
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(10);
    Klipboard::self()->populateHistory(ClipboardEntryList());
    QVERIFY(Klipboard::self()->isHistoryEmpty());

    /*
     * Finally test entryCount function
     */
    list = makeEntryList(10);
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(100);
    QVERIFY(Klipboard::self()->isHistoryEmpty());
    QCOMPARE(Klipboard::self()->entryCount(), 0);
    Klipboard::self()->populateHistory(list);
    QCOMPARE(Klipboard::self()->entryCount(), 10);
    QCOMPARE(Klipboard::self()->historySize(), 100);
}

void tst_Klipboard::setCurrentEntryTest()
{
    Klipboard::self()->clearHistory();
    QVERIFY(Klipboard::self()->isHistoryEmpty());
    Klipboard::self()->setHistorySize(10);

    ClipboardEntryList list = makeEntryList(10);
    Klipboard::self()->populateHistory(list);
    QCOMPARE(Klipboard::self()->entryCount(), 10);

    ClipboardEntry entry = Klipboard::self()->entry(5);
    QCOMPARE(entry.data.toString(), QString("5"));
    QString uuid = entry.uuid;
    QVariant data = entry.data;
    QVERIFY(!uuid.isEmpty());

    Klipboard::self()->setCurrentEntry(uuid);
    // wait for clipboard to catch up
    QTest::qWait(signalWaitTimeout);
    QCOMPARE(Klipboard::self()->currentEntry().data.toString(), QString("5"));

    // TODO: also verify that "entry" item has been removed from middle of the list
}

void tst_Klipboard::entryGetterTest()
{
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(10);

    ClipboardEntryList list = makeEntryList(10);
    Klipboard::self()->populateHistory(list);

    // by value
    ClipboardEntry entry = Klipboard::self()->entry(0);
    QCOMPARE(Klipboard::self()->entry(9).data.toString(), QString("9"));
    QCOMPARE(Klipboard::self()->entry(0).data.toString(), QString("0"));

    // by uuid
    QCOMPARE(Klipboard::self()->entry(entry.uuid).data.toString(), QString("0"));
}

void tst_Klipboard::addEntryTest()
{
    {
        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistorySize(10);

        ClipboardEntryList list = makeEntryList(4);
        Klipboard::self()->populateHistory(list);

        // check that duplicates will be excluded during addition
        ClipboardEntry last = list.last();
        ClipboardEntryList dupsList;
        for (int i=0; i<5; i++) dupsList.append(last);

        Klipboard::self()->populateHistory(dupsList);
        QCOMPARE(Klipboard::self()->entryCount(), 4); // shouldn't change
    }

    {
        // if one item is copied from selection and other from clipboard,
        // but both have the same content, they shouldn't be treated as dups
        ClipboardEntry entry1;
        entry1.data = "data\ndata";
        entry1.isSelection = true;
        ClipboardEntry entry2;
        entry2.data = "data\ndata";
        entry2.isSelection = false;
        ClipboardEntryList list;
        list << entry1 << entry2;
        Klipboard::self()->clearHistory();
        Klipboard::self()->populateHistory(list);
        QCOMPARE(Klipboard::self()->entryCount(), 2);
    }

    {
        // check that items with empty data will be excluded
        Klipboard::self()->clearHistory();
        ClipboardEntryList list;
        list.append(ClipboardEntry());
        Klipboard::self()->populateHistory(list);
        QCOMPARE(Klipboard::self()->entryCount(), 0); // shouldn't change
    }

    {
        // but not spaces or tabs, they should be added
        Klipboard::self()->clearHistory();
        ClipboardEntry entry;
        entry.data = "   \t";
        ClipboardEntryList list;
        list.append(entry);
        Klipboard::self()->populateHistory(list);
        QCOMPARE(Klipboard::self()->entryCount(), 1);
    }

    removeTempFile();

    {
        // check saving to history file
        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistoryFile(tempFileName);
        ClipboardEntryList list = makeEntryList(10);
        Klipboard::self()->populateHistory(list);

        HistoryFile file(tempFileName);
        QCOMPARE(file.entryCount(), 10);
    }
}

void tst_Klipboard::signalsTest()
{
    Klipboard::self()->clearHistory();
    Klipboard::self()->setHistorySize(10);

    // test populate + signal emission
    QSignalSpy entryAddedSpy(Klipboard::self(), SIGNAL(entryAdded(QString)));
    QSignalSpy entryRemovedSpy(Klipboard::self(), SIGNAL(entryRemoved(QString)));
    QSignalSpy currentChangedSpy(Klipboard::self(), SIGNAL(currentEntryChanged(QString)));

    ClipboardEntryList list = makeEntryList(10);
    Klipboard::self()->populateHistory(list);

    QCOMPARE(entryAddedSpy.count(), 10);
    QCOMPARE(currentChangedSpy.count(), 10);
    QCOMPARE(entryRemovedSpy.count(), 0);

    // test emission when clearing
    entryAddedSpy.clear();
    entryRemovedSpy.clear();
    currentChangedSpy.clear();

    Klipboard::self()->clearHistory();

    QCOMPARE(entryAddedSpy.count(), 0);
    QCOMPARE(currentChangedSpy.count(), 0);
    QCOMPARE(entryRemovedSpy.count(), 10);

    // test emission when changing current
    Klipboard::self()->clearHistory();
    Klipboard::self()->populateHistory(list);

    entryAddedSpy.clear();
    entryRemovedSpy.clear();
    currentChangedSpy.clear();

    Klipboard::self()->setCurrentEntry(Klipboard::self()->entry(5).uuid);
    QTest::qWait(signalWaitTimeout);
    QCOMPARE(entryAddedSpy.count(), 1);
    QCOMPARE(entryRemovedSpy.count(), 1);
    QCOMPARE(currentChangedSpy.count(), 1);
}

void tst_Klipboard::setHistoryFileTest()
{
    removeTempFile();
    {
        // setting to empty file should not change num entries
        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistoryFile(tempFileName);
        QCOMPARE(Klipboard::self()->entryCount(), 0);
    }
    removeTempFile();
    {
        QSignalSpy entryAddedSpy(Klipboard::self(), SIGNAL(entryAdded(QString)));

        // reading from file
        HistoryFile file(tempFileName);
        ClipboardEntryList list = makeEntryList(10);
        file.addEntries(list);
        file.close();

        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistoryFile(tempFileName);
        QCOMPARE(Klipboard::self()->entryCount(), 10);
        for (int i=0; i<10; ++i)
        {
            QCOMPARE(Klipboard::self()->entry(i).data.toString(), list.at(i).data.toString());
            QCOMPARE(Klipboard::self()->entry(i).uuid, list.at(i).uuid);
            QCOMPARE(Klipboard::self()->entry(i).isText, list.at(i).isText);
            QCOMPARE(Klipboard::self()->entry(i).isSelection, list.at(i).isSelection);
        }

        QTest::qWait(signalWaitTimeout);

        QCOMPARE(entryAddedSpy.count(), 10);
    }

    removeTempFile();

    {
        // changing history file should erase all entries
        HistoryFile file(tempFileName);
        ClipboardEntryList list = makeEntryList(10);
        file.addEntries(list);
        file.close();

        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistoryFile(tempFileName);
        QCOMPARE(Klipboard::self()->entryCount(), 10);

        Klipboard::self()->setHistoryFile(tempFileName1);
        QCOMPARE(Klipboard::self()->entryCount(), 0);
    }

    removeTempFile();

    {
        // changing history file should re-read all entries from new file
        HistoryFile file1(tempFileName);
        ClipboardEntryList list = makeEntryList(10);
        file1.addEntries(list);
        file1.close();

        HistoryFile file2(tempFileName1);
        ClipboardEntryList list1 = makeEntryList(5);
        file2.addEntries(list1);
        file2.close();

        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistoryFile(tempFileName);
        QCOMPARE(Klipboard::self()->entryCount(), 10);

        Klipboard::self()->setHistoryFile(tempFileName1);
        QCOMPARE(Klipboard::self()->entryCount(), 5);
    }

    removeTempFile();

    {
        // Klipboard should not insert more than its historySize
        HistoryFile file(tempFileName);
        ClipboardEntryList list = makeEntryList(100);
        bool res = file.addEntries(list);
        QVERIFY(res == true);
        file.close();

        Klipboard::self()->clearHistory();
        Klipboard::self()->setHistorySize(58);
        Klipboard::self()->setHistoryFile(tempFileName);
        QCOMPARE(Klipboard::self()->entryCount(), 58);
        // it should leave only last 58 items from history file
        QCOMPARE(Klipboard::self()->entry(0).data.toString(), list.at(100-58).data.toString());
        QCOMPARE(Klipboard::self()->entry(57).data.toString(), list.at(99).data.toString());
    }

    removeTempFile();
}

QTEST_KDEMAIN_KLIPPER(tst_Klipboard, GUI)
