/*
 *   Copyright © 2009 Dmitry Suzdalev (dimsuz@gmail.com)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License version 2 as
 *   published by the Free Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
#include "klipboardmodel.h"
#include "commondefs.h"

#include <KIcon>

KlipboardModel::KlipboardModel(const Plasma::DataEngine* dataEngine, QObject* parent)
    : QAbstractItemModel(parent), m_dataEngine(dataEngine)
{
    // init existing data
    foreach (const QString& source, m_dataEngine->sources())
        onClipEntryAdded(source);

    connect(m_dataEngine, SIGNAL(sourceAdded(const QString&)),
            SLOT(onClipEntryAdded(const QString&)));
    connect(m_dataEngine, SIGNAL(sourceRemoved(const QString&)),
            SLOT(onClipEntryRemoved(const QString&)));
}

int KlipboardModel::rowCount(const QModelIndex& parent) const
{
    Q_UNUSED(parent)

    QStringList sourcesList = m_dataEngine->sources();
    int count = sourcesList.count();
    if (sourcesList.contains(CURRENT_ENTRY_SOURCE))
        count--;
    return count;
}

int KlipboardModel::columnCount(const QModelIndex& parent) const
{
    Q_UNUSED(parent)
    return 1;
}

QModelIndex KlipboardModel::index(int row, int column, const QModelIndex& parent) const
{
    Q_UNUSED(parent)
    return createIndex(row, column);
}

QModelIndex KlipboardModel::indexForUuid(const QString& uuid) const
{
    int row = m_uuidRowMap.value(uuid, -1);
    return (row != -1 ? index(row, 0) : QModelIndex());
}

QModelIndex KlipboardModel::parent(const QModelIndex& index) const
{
    Q_UNUSED(index)
    return QModelIndex();
}

QVariant KlipboardModel::data(const QModelIndex& index, int role) const
{
    if (!index.isValid())
    {
        kDebug() << "Got invalid index, no go";
        return QVariant();
    }

    // fetch data only for roles which need it
    bool needToQueryEngineForData = (role == Qt::DisplayRole ||
                                     role == KlipboardModel::ClipboardTypeRole );

    QString uuid;
    Plasma::DataEngine::Data clipData;

    if (needToQueryEngineForData)
    {
        uuid = m_uuidRowMap.key(index.row());
        clipData = m_dataEngine->query(uuid);
    }

    if (role == Qt::DisplayRole)
    {
        QString text;
        if (clipData.value("IsText").toBool() == true)
            text = clipData.value("Data").toString();
        else
            // TODO image support
            text = "Images not supported yet, sorry";

        // remove '\n,\t' etc
        return text.simplified();
    }
    else if (role == Qt::DecorationRole)
    {
        return KIcon("text-plain");
    }
    else if (role == KlipboardModel::UuidRole )
    {
        return m_uuidRowMap.key(index.row());
    }
    else if (role == KlipboardModel::ClipboardTypeRole )
    {
        // TODO i18n!
        return clipData.value("IsSelection").toBool() ? "Selection" : "Clipboard";
    }
    else if (role == AlwaysShowClipboardTypeRole)
        return true;

    return QVariant();
}

void KlipboardModel::onClipEntryAdded(const QString& sourceName)
{
    // ignore addition of currentEntry source,
    // model is interested in clipboard entry sources only
    if (sourceName == CURRENT_ENTRY_SOURCE)
        return;

    QString uuid = sourceName;

    // call this as required by QAbstractItemModel
    beginInsertRows(QModelIndex(), 0, 0);

    // increase row indexes for each entry - because we're inserting new row,
    // they need to shift
    QHash<QString,int>::iterator i = m_uuidRowMap.begin();
    while (i != m_uuidRowMap.end())
    {
        i.value() += 1;
        ++i;
    }

    // now insert a new entry with row = 0
    m_uuidRowMap[uuid] = 0;

    // call this as required by QAbstractItemModel
    endInsertRows();
}

void KlipboardModel::onClipEntryRemoved(const QString& sourceName)
{
    // ignore removal of currentEntry source,
    // model is interested in clipboard entry sources only
    if (sourceName == CURRENT_ENTRY_SOURCE)
        return;

    QString uuid = sourceName;

    int row = m_uuidRowMap.value(uuid, -1);
    if (row != -1)
    {
        // call this as required by QAbstractItemModel
        beginRemoveRows(QModelIndex(), row, row);

        // decrease row indexes of all rows below one being removed,
        // they need to shift up
        QHash<QString,int>::iterator i = m_uuidRowMap.begin();
        while (i != m_uuidRowMap.end())
        {
            if (i.value() > row)
                i.value() -= 1;
            ++i;
        }

        m_uuidRowMap.remove(uuid);

        // call this as required by QAbstractItemModel
        endRemoveRows();
    }
    else
        kDebug() << "failed to remove clipboard entry from model: can't find corresponding row";
}

#include "klipboardmodel.moc"
