/* This file is part of the KDE Project
   Copyright (c) 2009 Daniel Winter <dw@danielwinter.de>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QSignalMapper>
#include <QTimer>

#include "mountsignalfilter.h"
#include "filesystem.h"

namespace Nepomuk
{

MountSignalFilter::MountSignalFilter( QObject *parent, int delay, int maxPriority )
        : QObject( parent ), m_delay( delay ), m_maxPriority( maxPriority ), m_mountSignalMapper( 0 )
        , m_unmountSignalMapper( 0 )
{
    m_mountSignalMapper = new QSignalMapper( this );
    connect( m_mountSignalMapper, SIGNAL( mapped( const QString& ) ),
             this, SLOT( slotTimerForMountElapsed( const QString& ) ) );

    m_unmountSignalMapper = new QSignalMapper( this );
    connect( m_unmountSignalMapper, SIGNAL( mapped( const QString& ) ),
             this, SLOT( slotTimerForUnmountElapsed( const QString& ) ) );
}


void MountSignalFilter::slotFilesystemMounted( Filesystem fs )
{
    doSlotFilesystem( fs, m_mountSignalMapper, m_mountList, true );
}

void MountSignalFilter::slotFilesystemUnmounted( Filesystem fs )
{
    doSlotFilesystem( fs, m_unmountSignalMapper, m_unmountList, false );
}

void MountSignalFilter::doSlotFilesystem( const Filesystem& fs, QSignalMapper* mapper, QHash<QString, QPair< Filesystem, bool > >& mountList, bool mounted )
{
    QString mountpoint = fs.mountpoint();

    if ( mountList.contains( mountpoint ) ) {

        QPair< Filesystem, bool> fsData = mountList.value( mountpoint );

        if ( fsData.second ) {
            // already emmited a signal for this fs, block it by doing nothing
            return;
        } else {
            // if new fs has a higher priority replace it
            if ( fs.priority() > fsData.first.priority() ) {
                mountList.insert( mountpoint, QPair< Filesystem, bool>(fs,false) );
            }
        }
    } else {
        QTimer *timer = new QTimer( this );
        m_timerHash.insert( mountpoint, timer );
        timer->setSingleShot( true );
        mapper->setMapping( timer, mountpoint );
        connect( timer, SIGNAL( timeout() ),
            mapper, SLOT( map() ) );
        timer->start( m_delay );
        mountList.insert( mountpoint, QPair< Filesystem, bool>(fs, false) );
    }

    // if fs has the higest priority direct emit the signal and set it to emited
    if ( fs.priority() == m_maxPriority ) {
        if ( mounted ) {
            emit filesystemMounted( fs );
        } else {
            emit filesystemUnmounted( fs );
        }
        mountList.insert( mountpoint, QPair<Filesystem, bool>(fs, true) );
    }
}


void MountSignalFilter::slotTimerForMountElapsed( const QString& mountpoint )
{
    doSignalTimerForMount( mountpoint, m_mountSignalMapper, m_mountList, true );
}

void MountSignalFilter::slotTimerForUnmountElapsed( const QString& mountpoint )
{
    doSignalTimerForMount( mountpoint, m_unmountSignalMapper, m_unmountList, false );
}

void MountSignalFilter::doSignalTimerForMount( const QString& mountpoint, QSignalMapper* mapper, QHash<QString, QPair< Filesystem, bool > >& mountList, bool mounted )
{
    QPair<Filesystem, bool> fsData = mountList.take( mountpoint );
    QTimer* timer = m_timerHash.take( mountpoint );
    mapper->removeMappings( timer );
    delete timer;

    // if not already emmited (because of maxPriority signal emit
    if ( !fsData.second ) {
        if ( mounted ) {
            emit filesystemMounted( fsData.first );
        } else {
            emit filesystemUnmounted( fsData.first );
        }
    }
}

}

#include "mountsignalfilter.moc"
