/*
 * Copyright (c) 2010 Ivan Cukic <ivan.cukic(at)kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "activities.h"

#include <QCoreApplication>

#include <KDebug>
#include <KComponentData>
#include <KUser>
#include <kio/udsentry.h>

#include <kworkspace/kactivityconsumer.h>
#include <kworkspace/kactivityinfo.h>

class Activities::Private {
public:
    Private(Activities * parent)
        : q(parent)
    {
        client = new KActivityConsumer();
        initCategories();
    }

    ~Private()
    {
        delete client;
    }

    KIO::UDSEntry createUDSEntryActivity(const QString & activity)
    {
        kDebug(7102) << activity;
        QString activityID = activity;
        activityID.replace(QRegExp("[{}]"), QString());

        KIO::UDSEntry uds;

        uds.insert(KIO::UDSEntry::UDS_NAME,              activityID);
        uds.insert(KIO::UDSEntry::UDS_DISPLAY_NAME,      KActivityInfo::name(activity));
        uds.insert(KIO::UDSEntry::UDS_CREATION_TIME,     0);
        uds.insert(KIO::UDSEntry::UDS_MODIFICATION_TIME, 0);
        uds.insert(KIO::UDSEntry::UDS_ACCESS,            0700);
        uds.insert(KIO::UDSEntry::UDS_USER,              KUser().loginName());
        uds.insert(KIO::UDSEntry::UDS_FILE_TYPE,         S_IFDIR);
        uds.insert(KIO::UDSEntry::UDS_MIME_TYPE,         QString::fromLatin1("inode/directory"));
        uds.insert(KIO::UDSEntry::UDS_ICON_NAME,         QString::fromLatin1("preferences-activities"));

        return uds;
    }

    KIO::UDSEntry createUDSEntryActivitySubitem(const QString & activity, const QString & subitemID, const QString & subitemName, const QString & subitemIcon)
    {
        kDebug(7102) << activity << subitemID << subitemName << subitemIcon;

        KIO::UDSEntry uds;

        uds.insert(KIO::UDSEntry::UDS_NAME,              subitemID);
        uds.insert(KIO::UDSEntry::UDS_DISPLAY_NAME,      subitemName);
        uds.insert(KIO::UDSEntry::UDS_CREATION_TIME,     0);
        uds.insert(KIO::UDSEntry::UDS_MODIFICATION_TIME, 0);
        uds.insert(KIO::UDSEntry::UDS_ACCESS,            0700);
        uds.insert(KIO::UDSEntry::UDS_USER,              KUser().loginName());
        uds.insert(KIO::UDSEntry::UDS_FILE_TYPE,         S_IFDIR);
        uds.insert(KIO::UDSEntry::UDS_MIME_TYPE,         QString::fromLatin1("inode/directory"));
        uds.insert(KIO::UDSEntry::UDS_ICON_NAME,         subitemIcon);

        return uds;
    }

    void listActivities()
    {
        foreach (const QString & activity, client->availableActivities()) {
            kDebug(7102) << "Adding activity:" << activity;
            q->listEntry(createUDSEntryActivity(activity), false);
        }

        q->listEntry(KIO::UDSEntry(), true);
        q->finished();
    }

    void listActivity(const QString & activity)
    {
        foreach (const QString & category, categories.keys()) {
            q->listEntry(createUDSEntryActivitySubitem(activity,
                    category, category, categories[category].first), false);
        }

        // q->listEntry(createUDSEntryActivitySubitem(activity,
        //         "other", i18n("Other"), "folder-grey"), false);

        q->listEntry(KIO::UDSEntry(), true);
        q->finished();
    }

    /**
     * @returns whether path is handled by activities KIO
     */
    bool isPathVirtual(const KUrl & url)
    {
        QStringList pathSegments;

        if (!decodePath(url, pathSegments) || pathSegments.size() <= 1) {
            return true;
        }

        return false;
    }

    /**
     * Splits the path into segments.
     * @returns if the path is empty - root path
     */
    bool decodePath(const KUrl & url, QStringList & pathSegments)
    {
        QString path = url.path();

        if (path.isEmpty() || path == "/") {
            kDebug(7102) << "Path is empty" << path;
            return false;
        }

        pathSegments = path.split('/');
        pathSegments = pathSegments.filter(QRegExp(".+"));

        kDebug(7102) << "Path segments:" << path << pathSegments;

        return true;
    }

    void initCategories()
    {
        #define AddCategory(A, B, C)                      \
            categories[A] = QPair < QString, QString > (B, C)

        AddCategory(
                i18n("Applications"),
                "applications-other",
                "nfo:Application");

        AddCategory(
                i18n("Documents"),
                "document-multiple",
                "nfo:Document");

        AddCategory(
                i18n("Places"),
                "computer",
                "nfo:Folder");

        // AddCategory(
        //         i18n("Contacts"),
        //         "system-users",
        //         "nco:Contact");

        #undef AddCategory
    }

    QMap < QString, QPair < QString, QString > > categories;
    KActivityConsumer * client;
    Activities * q;
};

void Activities::listDir(const KUrl & url)
{
    QStringList pathSegments;

    if (!d->decodePath(url, pathSegments)) {
        d->listActivities();

    } else if (pathSegments.size() == 1) {
        d->listActivity(pathSegments[0]);

    } else {
        ForwardingSlaveBase::listDir(url);

    }
}

void Activities::mimetype(const KUrl & url)
{
    kDebug(7102) << "in" << url;

    if (!d->isPathVirtual(url)) {
        mimeType("inode/directory");
        finished();

    } else {
        ForwardingSlaveBase::mimetype(url);

    }

    kDebug(7102) << "out" << url;

}

bool Activities::rewriteUrl(const KUrl & url, KUrl & newURL)
{
    kDebug(7102) << url;

    QStringList pathSegments;

    if (!d->decodePath(url, pathSegments) || pathSegments.size() == 1) {
        return false;

    } else {
        kDebug(7102) << pathSegments;

        // TODO: hardcoded uri is bad
        // And this takes toooooo much memory
        QString query = QString("select ?r where { <%1> nao:isRelated ?r . ?r a %2 . }")
                .arg((new KActivityInfo(pathSegments[0], this))->resourceUri().url())
                .arg(d->categories[pathSegments[1]].second);
        kDebug(7102) << query;

        newURL = "nepomuksearch:/";
        newURL.addQueryItem( "sparql", query );

        for (int i = 2; i < pathSegments.size(); i++) {
            newURL.addPath(pathSegments[i]);
        }

        kDebug(7102) << newURL;
        kDebug() << pathSegments << query << newURL;

        return true;

    }

    // newURL = url;
    return false;
}

void Activities::get(const KUrl& url)
{
    if ( !d->isPathVirtual(url) ) {
        ForwardingSlaveBase::get(url);
    }
    else {
        error( KIO::ERR_DOES_NOT_EXIST, url.prettyUrl() );
    }
}

void Activities::stat(const KUrl & url)
{
    QString path = url.path();

    if (d->isPathVirtual(url)) {
        KIO::UDSEntry uds;
        if (!path.endsWith('/')) {
            path += '/';
        }

        uds.insert(KIO::UDSEntry::UDS_NAME, path);
        uds.insert(KIO::UDSEntry::UDS_FILE_TYPE, S_IFDIR);
        uds.insert(KIO::UDSEntry::UDS_MIME_TYPE, QString::fromLatin1("inode/directory"));

        statEntry( uds );
        finished();

    } else {
        ForwardingSlaveBase::stat(url);

    }
}


Activities::Activities(const QByteArray & pool, const QByteArray & app)
    : ForwardingSlaveBase("activities", pool, app), d(new Private(this))
{
}

Activities::~Activities()
{
    delete d;
}

extern "C" int KDE_EXPORT kdemain(int argc, char **argv)
{
    KComponentData instance("kio_activities");
    QCoreApplication app(argc, argv);

    kDebug(7102) << "Starting activities slave";

    if (argc != 4) {
        kError() << "Usage: kio_activities protocol domain-socket1 domain-socket2";
        exit(-1);
    }

    Activities slave(argv[2], argv[3]);

    slave.dispatchLoop();

    kDebug(7102) << "Activities slave finished";

    return EXIT_SUCCESS;
}

