/*
Copyright (c) 2010 Oszkar Ambrus <aoszkar@gmail.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ABSTRACTRESOURCEEDITWIDGET_H
#define ABSTRACTRESOURCEEDITWIDGET_H

#include <QWidget>
#include <QUrl>
#include <QList>

#include <Nepomuk/Resource>

#include "nepomuksearchgui_export.h"

class KPushButton;
class QAction;

namespace Nepomuk {
    /**
     * The abstract base class for widgets that provide displaying, editing and custom actions
     * on a Nepomuk::Resource.
     *
     * These widgets are provided by ResourceItemDelegate as editors in a view.
     *
     * Editing widgets are created through the ResourceEditPluginFactory that tries
     * to find an editing widget plugin for the given resource type.
     *
     * GenericResourceEditPlugin handles generic literal properties of resources,
     * such as  rdfs:label or rdfs:comment, and a standard editing of
     * resource properties.
     *
     * When subclassing, make sure to initGui() to set up the GUI (PropertyEditWidgets can be used for the properties),
     * and setupActions() to add the desired actions for the given resource
     *
     * \section plugins Plugin
     *
     * KDE Service type: Nepomuk/ResourceEditPlugin
     *
     * X-Nepomuk-ResourceTypes should list the supported resource types.
     *
     * @author Oszkar Ambrus aoszkar@gmail.com
     */

    class PropertyEditWidget;

class NEPOMUKSEARCHGUI_EXPORT AbstractResourceEditPlugin : public QWidget
{
    Q_OBJECT

public:
    /**
     * Creates a new AbstractResourceEditPlugin
     *
     * @param parent the parent of the widget being built
     *
     */
    AbstractResourceEditPlugin( QWidget *parent = 0 );
    virtual ~AbstractResourceEditPlugin();

    /**
     * Sets the Resource object to be edited to @p resource and reinitializes the GUI
     * @see resource()
     */
    void setResource(const Resource &resource);

    /**
     * Returns the Resource object associated to the widget
     *
     * @return the Nepomuk resource handled by the widget
     */
    Nepomuk::Resource resource() const;

    /**
     * \returns true if editing is enabled
     */
    bool editingEnabled();

Q_SIGNALS:
    void geometryChanged();

public Q_SLOTS:
    /**
     * Enables or disables editing on each PropertyEditWidget
     *
     * To customize, override it, but call the default method first to handle all widgets added through addPropertyEditWidget()
     */
    virtual void setEditingEnabled(bool enable);

    /**
     * Reverts the changes done on the resource
     * The default version restores all PropertyEditWidgets from the resource
     *
     * To customize, override it, but call the default method first to handle all widgets added through addPropertyEditWidget()
     */
    virtual void revertChanges();

    /**
     * Saves the changes done on the resource
     * The default version saves all properties on the given resource from the PropertyEditWidgets
     *
     * To customize, override it, but call the default method first to handle all widgets added through addPropertyEditWidget()
     */
    virtual void saveChanges();

protected:
    /**
     * Performs initialization of the GUI.
     * This function is called from setResource whenever a new resource is set
     *
     * To have widgets handled (enabled, saved, disabled) in a default way, use addPropertyEditWidget()
     */
    virtual void initGui() = 0;

    /**
      * Adds @widget to the list of PropertyEditWidgets
      * AbstractResourceEditPlugin handles the default behaviour (enabling, disabling, saving) of these widgets
      */
    void addPropertyEditWidget(PropertyEditWidget *widget);

    /**
      * Sets up the list of custom actions to be performed on the handled resource
      */
    virtual void setupActions();

private:
    class Private;
    Private * const d;

private Q_SLOTS:
    /**
      * Shows a list of related actions
      */
    void showRelatedResources();
};

}

#endif // ABSTRACTRESOURCEEDITWIDGET_H

