/*
Copyright (c) 2010 Oszkar Ambrus <aoszkar@gmail.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "resourcecreateandsearchwidget.h"

#include "searchwidget.h"
#include "resourceeditdialog.h"

#include <QHBoxLayout>
#include <QLabel>
#include <QMouseEvent>
#include <QEvent>
#include <QClipboard>
#include <QKeySequence>
#include <QToolButton>
#include <QDialog>
#include <QMessageBox>
#include <QTimer>

#include <KDebug>
#include <KPushButton>
#include <KMessageBox>
#include <KStandardGuiItem>
#include <KAction>
#include <KStandardAction>
#include <KApplication>
#include <KLineEdit>

#include <Nepomuk/Types/Class>
#include <Nepomuk/Query/ResourceTypeTerm>
#include <Nepomuk/Resource>
#include <Nepomuk/ResourceManager>
#include <nepomuk/newresourcedialog.h>
#include <nepomuk/resourcecompletion.h>

class Nepomuk::ResourceCreateAndSearchWidget::Private
{
public:
    KLineEdit *lineEdit;
    Nepomuk::ResourceCompletion *completer;
    Nepomuk::Resource resource;

    QToolButton *btnOpen;
    QToolButton *btnNew;
    QToolButton *btnSearch;
};

// trueg: why not a lineedit by default which acts as follows:
// 1. when no resource is selected it is empty and allows to do a simple search (advanced search can still be done the current way)
// 2. when one resource is selected show its label and allow to change that the same way as in 1. ESC cancels the change
// 3. when the property has cardinality >1 show the + sign which opens another line edit exibiting the same behaviour.

/******************* ResourceCreateAndSearchWidget ************************/

Nepomuk::ResourceCreateAndSearchWidget::ResourceCreateAndSearchWidget(QWidget *parent)
    : QWidget( parent ) ,
      d( new Private() )
{
    //LineEdit for labellabel
    d->lineEdit = new KLineEdit(this);
    d->completer = new Nepomuk::ResourceCompletion(this);
    d->lineEdit->setCompleter(d->completer);

    connect(d->completer, SIGNAL(resourceActivated(Nepomuk::Resource)), this, SLOT(setResource(Nepomuk::Resource)));
    connect(d->completer, SIGNAL(resourceActivated(Nepomuk::Resource)), this, SIGNAL(valueChanged()));
    connect(d->lineEdit, SIGNAL(textEdited(QString)), d->completer, SLOT(setCompletionText(QString)));

    //new button
    d->btnOpen = new QToolButton(this);
    d->btnOpen->setIcon(KIcon("document-open-data"));
    d->btnOpen->setToolTip(i18n("Open Resource"));
    d->btnOpen->setMaximumSize(KIconLoader::SizeSmallMedium, KIconLoader::SizeSmallMedium);
    connect(d->btnOpen, SIGNAL(clicked()), this, SLOT(showResource()));

    //new button
    d->btnNew = new QToolButton(this);
    d->btnNew->setIcon(KIcon("document-new"));
    d->btnNew->setToolTip(i18n("New"));
    d->btnNew->setMaximumSize(KIconLoader::SizeSmallMedium, KIconLoader::SizeSmallMedium);
    connect(d->btnNew, SIGNAL(clicked()), this, SLOT(createResource()));

    //search buttin
    d->btnSearch = new QToolButton(this);
    d->btnSearch->setIcon(KIcon("edit-find"));
    d->btnSearch->setToolTip(i18n("Search"));
    d->btnSearch->setMaximumSize(KIconLoader::SizeSmallMedium, KIconLoader::SizeSmallMedium);
    connect(d->btnSearch, SIGNAL(clicked()), this, SLOT(searchResource()));

    //mainlayout
    QHBoxLayout *hLayout = new QHBoxLayout(this);
    hLayout->addWidget(d->lineEdit);
    hLayout->addStretch();
    hLayout->addWidget(d->btnOpen);
    hLayout->addWidget(d->btnNew);
    hLayout->addWidget(d->btnSearch);

    hLayout->setContentsMargins(0, 0, 0, 0);
    //setResource(d->lineEdit->resource());//to initialize the label

    //copy actions
    KAction *copyLabel = KStandardAction::copy(this, SLOT(copyLabelToClipboard()), this);
    copyLabel->setText(i18n("Copy label"));
    copyLabel->setShortcut(QKeySequence());
    addAction(copyLabel);

    KAction *copyUri = KStandardAction::copy(this, SLOT(copyUriToClipboard()), this);
    copyUri->setText(i18n("Copy URI"));
    copyUri->setShortcut(QKeySequence());
    addAction(copyUri);

    this->setContextMenuPolicy(Qt::ActionsContextMenu);
    setFocusPolicy(Qt::StrongFocus);
}

Nepomuk::ResourceCreateAndSearchWidget::~ResourceCreateAndSearchWidget()
{
    delete d;
}

void Nepomuk::ResourceCreateAndSearchWidget::setType(const Types::Class &type)
{
    d->completer->setType(type);
}

void Nepomuk::ResourceCreateAndSearchWidget::setResource(const Nepomuk::Resource& resource)
{
    if(resource.hasType(d->completer->type().uri())) {
        //kDebug() << "*********** Setting resource:" << resource.genericLabel() << resource.uri();

        //FIXME: This is a hack to prevent the lineEdit from restoring its text (probably due to a KCompleter bug)
        //d->lineEdit->setText(resource.genericLabel());
        QTimer::singleShot(0, this, SLOT(updateLabel()));

        d->resource = resource;
    }
}

Nepomuk::Resource Nepomuk::ResourceCreateAndSearchWidget::resource() const
{
    return d->resource;
}

void Nepomuk::ResourceCreateAndSearchWidget::changeEvent ( QEvent * event )
{
    if(event->type() == QEvent::EnabledChange) {
        if(this->isEnabled()) {
            d->btnNew->setVisible(true);
            d->btnSearch->setVisible(true);
        }
        else {
            d->btnNew->setVisible(false);
            d->btnSearch->setVisible(false);
        }
    }
}

void Nepomuk::ResourceCreateAndSearchWidget::showResource()
{
    Nepomuk::Resource res = d->resource;
    if(res.isValid()) {
        Nepomuk::ResourceEditDialog::editResource(res);
        setResource(res);//refresh, in case the dialog changed it
    }
    else {
        KMessageBox::sorry(this, "Resource is not valid", "Error!");
    }
}

void Nepomuk::ResourceCreateAndSearchWidget::createResource()
{
    Nepomuk::Resource resource = Nepomuk::NewResourceDialog::createResource(d->completer->type(), this);
    if(resource.isValid()) {
        setResource(resource);
        emit valueChanged();
    }
}

void Nepomuk::ResourceCreateAndSearchWidget::searchResource()
{
    Nepomuk::Query::ResourceTypeTerm typeTerm(d->completer->type());
    Nepomuk::Resource resource = Nepomuk::Utils::SearchWidget::searchResource(this, Nepomuk::Query::Query(typeTerm));
    if(resource.isValid()) {
        setResource(resource);
        emit valueChanged();
    }
}

void Nepomuk::ResourceCreateAndSearchWidget::copyLabelToClipboard()
{
    KApplication::clipboard()->setText(d->resource.genericLabel());
}

void Nepomuk::ResourceCreateAndSearchWidget::copyUriToClipboard()
{
    KApplication::clipboard()->setText(d->resource.uri());
}

void Nepomuk::ResourceCreateAndSearchWidget::updateLabel()
{
    d->lineEdit->setText(d->resource.genericLabel());
}

#include "resourcecreateandsearchwidget.moc"
