/*
Copyright (c) 2010 Oszkar Ambrus <aoszkar@gmail.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef RESOURCEEDITWIDGET_H
#define RESOURCEEDITWIDGET_H

#include <QWidget>
#include <QModelIndex>

#include "abstractresourceeditplugin.h"

#include <Nepomuk/Resource>

#include "nepomuksearchgui_export.h"


class KPushButton;

namespace Nepomuk
{

    /**
     * The  base class for widgets that provide displaying and editing
     * of a Nepomuk::Resource.
     *
     * These widgets are provided by ResourceItemDelegate as editors in a view.
     *
     * Editing plugins (AbstractResourceEditPlugins) are created through the ResourceEditPluginFactory that tries
     * to find an editing widget plugin for the given resource type.
     *
     * GenericResourceEditWidget handles generic literal properties of resources,
     * such as  rdfs:label or rdfs:comment, and a standard editing of
     * resource properties.
     *
     * \section plugins Plugin
     *
     * KDE Service type: Nepomuk/ResourceEditWidget
     *
     * X-Nepomuk-ResourceTypes should list the supported resource types.
     *
     * @author Oszkar Ambrus aoszkar@gmail.com
     */

class NEPOMUKSEARCHGUI_EXPORT ResourceEditWidget : public QWidget
{
    Q_OBJECT

public:
    /**
     * Creates a new ResourceEditWidget
     *
     * @param parent the parent of the widget being built
     * @return the Nepomuk resource to be handled by the widget
     *
     */
    ResourceEditWidget( QWidget *parent = 0, const Nepomuk::Resource &resource = Nepomuk::Resource() );
    virtual ~ResourceEditWidget();

    /**
     * Sets the Resource object to be edited to @p resource and reinitializes the GUI
     * @see resource()
     * @warning THIS DOES NOT WORK (layout is not replaced)
     */
    void setResource(const Resource &resource);

    /**
     * Returns the Resource object associated to the widget
     *
     * @return the Nepomuk resource handled by the widget
     */
    Nepomuk::Resource resource() const;

    /**
      * Changes the editing mode of the editing plugin.
      * @param specifies the action to be done: StartEditing enables editing mode,
      * SaveChanges saves changes and disables editing, CancelChanges discards changes and disables editing
      */
    //virtual void changeEditingMode(AbstractResourceEditPlugin::EditingAction action = AbstractResourceEditPlugin::CancelChanges);

    /**
     * @reimp
     *
     * Updates geometry and emits the geometryChanged() signal
     */
public Q_SLOTS:
    /**
     * @reimp
     */
    void updateGeometry();

    /**
     * Enables or disables editing on of the resource
     */
    virtual void setEditingEnabled(bool enable);

    /**
     * Reverts the changes done on the resource
     */
    void revertChanges();

    /**
     * Saves the changes done on the resource
     */
    void saveChanges();

Q_SIGNALS:
    /**
     * The signal is emitted when the geometry is changed
     */
    void geometryChanged();

private:
    void initGui();

    class Private;
    Private * const d;

private Q_SLOTS:
    /**
     * Changes the interaction mode.
     *
     * When activating editing mode, the Edit button is changed to Save,
     * and Cancel is enabled.
     */
    //void btnEditSaveClicked();

    /**
     * Closes editing mode
     */
    //void btnCancelClicked();

    /**
     * Changes the rating of resource to @p rating
     */

    void setRating(int rating);

};

}

#endif // RESOURCEEDITWIDGET_H

