/*  This file is part of the KDE semantic clipboard
    Copyright (C) 2008 Tobias Wolf <twolf@access.unizh.ch>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include <memory>
#include <iostream>

#include <QSet>
#include <QString>
#include <QTextStream>
#include <QUrl>

#include <Soprano/Model>
#include <Soprano/PluginManager>
#include <Soprano/Serializer>
#include <Soprano/StatementIterator>

#include <KAboutData>
#include <KApplication>
#include <KCmdLineArgs>
#include <KUrl>

#include "graphcompletor.h"
#include "graphretriever.h"

void printRdfXmlSerialization( const Soprano::Model* model )
{
    const Soprano::Serializer* serializer =
        Soprano::PluginManager::instance()->discoverSerializerForSerialization( Soprano::SerializationRdfXml );
    QTextStream stream( stdout );
    serializer->serialize( model->listStatements(), stream, Soprano::SerializationRdfXml );
}

int main( int argc, char* argv[] )
{
    KAboutData aboutData( "complete",                                  // the program name used internally
                          0,                                           // the message catalog name
                          ki18n( "RDF Completor" ),                    // a displayable program name string
                          "1.0",                                       // the program version string.
                          ki18n( "Completes RDF graphs" ),             // short description of what the app does
                          KAboutData::License_LGPL,                    // the license this code is released under
                          ki18n( "Copyright (c) 2008 Tobias Wolf" ),   // copyright statement
                          ki18n( ""),                                  // optional text shown in the about box
                          "",                                          // the program homepage string
                          "twolf@access.unizh.ch" );                   // the bug report email address

    KCmdLineOptions options;
    options.add( "p" );
    options.add( "property <URI>", ki18n( "completion property URI" ) );
    options.add( "s" );
    options.add( "short <prefix>", ki18n( "target ontology prefix (foaf, vcard, swrc)" ) );
    options.add( "t" );
    options.add( "target <URI>", ki18n( "target ontology URI" ) );
    options.add( "+input" , ki18n( "HTTP address of the input RDF graph" ) );

    KCmdLineArgs::init( argc, argv, &aboutData );
    KCmdLineArgs::addCmdLineOptions( options );

    KApplication app;
    KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
    if ( !args->count() ) {
        return -1;
    }

    KUrl input = args->url( 0 );

    QSet< QUrl > properties;
    foreach ( const QString property, args->getOptionList( "property" ) ) {
        properties.insert( QUrl( property ) );
    }

    QSet< QUrl > targets;
    foreach ( const QString target, args->getOptionList( "short" ) ) {
        if ( target.compare( "foaf", Qt::CaseInsensitive ) == 0 ) {
            targets.insert( QUrl( "http://xmlns.com/foaf/0.1/" ) );
        } else if ( target.compare( "vcard", Qt::CaseInsensitive ) == 0 ) {
            targets.insert( QUrl( "http://www.w3.org/2006/vcard/ns#" ) );
        } else if ( target.compare( "swrc", Qt::CaseInsensitive ) == 0 ) {
            targets.insert( QUrl( "http://swrc.ontoware.org/ontology#" ) );
        } else if ( target.compare( "nco", Qt::CaseInsensitive ) == 0 ) {
            targets.insert( QUrl( "http://www.semanticdesktop.org/ontologies/2007/03/22/nco#" ) );
        }
    }
    foreach ( const QString target, args->getOptionList( "target" ) ) {
        targets.insert( QUrl( target ) );
    }

    std::auto_ptr< SemClip::GraphRetriever > graphRetriever( new SemClip::GraphRetriever() );
    graphRetriever->get( input );
    std::auto_ptr< Soprano::Model > model( graphRetriever->model() );
    if ( !model.get() || model->isEmpty() ) {
        return -1;
    }

    SemClip::GraphCompletor* graphCompletor = SemClip::GraphCompletor::instance();
    graphCompletor->complete( model.get(), properties, targets, SemClip::GraphCompletor::CompleteByProperty );
    printRdfXmlSerialization( model.get() );

    return 0;
}
