/*
 * This file is part of Search Project.
 *
 * Copyright (C) 2007 Daniele Galdi <daniele.galdi@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include <QDebug>
#include <QVariant>
#include <QPainter>
#include <QDateTime>
#include <QModelIndex>
#include <QAbstractItemView>
#include <QAbstractItemModel>

#include <KUrl>
#include <KGlobal>
#include <KLocale>
#include <KMimeType>
#include <KIconLoader>

#include "ItemDelegate.h"

const int LIST_HEIGHT               = 20; // 20px
const int BOTTOM_BORDER_SIZE        =  1; // 1px 
const int TEXT_POSITION             = 20;
const int PIXMAP_SIZE               = 16;

ItemDelegate::ItemDelegate( QObject *parent )
  :QItemDelegate(parent)
{
}
  
ItemDelegate::~ItemDelegate()
{
}

QSize ItemDelegate::sizeHint ( const QStyleOptionViewItem& option, const QModelIndex & index ) const
{
  QAbstractItemView* view = qobject_cast<QAbstractItemView*>( parent() );
  QWidget* widget = view->indexWidget( index );

  // Here we can discriminate about root/group/leaf node
  // thanks to the index.internalId() value,
  // and return the appropriate hint size.

  if ( index.internalId() == 2 && widget )
    return widget->sizeHint();

  return QSize( 200, LIST_HEIGHT + BOTTOM_BORDER_SIZE);
}

void ItemDelegate::paint( QPainter* painter, const QStyleOptionViewItem& option, const QModelIndex& index ) const
{
  if ( !index.isValid() )
    return;
  
  //--------------------
  
  if ( index.internalId() == 1 )
  {
    QRect rectangle( TEXT_POSITION, option.rect.y(), option.rect.width(), option.rect.height() );

    painter->fillRect( option.rect, QBrush(Qt::gray) );
    painter->drawText( rectangle, Qt::AlignLeft | Qt::AlignVCenter, index.data( Qt::DisplayRole ).toString() );

    return;
  }

  //------------------------------------
  QAbstractItemView* view = qobject_cast<QAbstractItemView*>( parent() );
  QWidget* widget = view->indexWidget( index );
  if ( widget ) 
    return;

  const QAbstractItemModel* model = index.model();

  QModelIndex uriIndex = model->index(index.row(), 0, index );
  QVariant uri         = uriIndex.data( Qt::DisplayRole );
  KUrl url( uri.toString() );

  QModelIndex mimeTypeIndex = model->index(index.row(), 1, index );
  QVariant mimeType         = mimeTypeIndex.data( Qt::DisplayRole );
  
  QModelIndex mtimeIndex = model->index(index.row(), 3, index );
  QVariant mtime         = mtimeIndex.data( Qt::DisplayRole );
  QDateTime datetime =  QDateTime::fromTime_t( mtime.toInt() );

  QString modified = KGlobal::locale()->formatDate( datetime.date() );

  qDebug() << "ItemDelegate::paint" << url.fileName() << mimeType << modified;

  //---------------------
  
  qDebug() << option.rect;

  if (option.state & QStyle::State_Selected)
  {
    QRect region(0, option.rect.y(), option.rect.width() + option.rect.x(), LIST_HEIGHT );
    painter->fillRect(region, option.palette.highlight());
  }
  
  // Draw bottom border
  painter->save();

  // Set bottom border properties
  QPen pen;
  pen.setWidth( BOTTOM_BORDER_SIZE );
  pen.setColor( Qt::gray );    
  painter->setPen( pen );

  QPoint bottomLeft( 0, option.rect.bottomLeft().y() );
  QPoint bottomRight( option.rect.bottomRight().x(), option.rect.bottomRight().y() );

  painter->drawLine( bottomLeft , option.rect.bottomRight() );

  painter->restore();

  // Draw icon

  KMimeType::Ptr mt = KMimeType::mimeType( mimeType.toString() );
  QString iconName = "unknow";
  if ( mt )
  {
    iconName = mt->iconName();
  }

  KIconLoader* iconLoader = new KIconLoader();
  QPixmap icon = iconLoader->loadMimeTypeIcon( iconName, KIconLoader::Desktop, KIconLoader::SizeSmall).scaled( PIXMAP_SIZE, PIXMAP_SIZE );

  painter->save();

  int topMargin = ( (LIST_HEIGHT - PIXMAP_SIZE ) / 2 );

  painter->translate( TEXT_POSITION, option.rect.topLeft().y() + topMargin );
  painter->drawPixmap( QRect(0,0,PIXMAP_SIZE,PIXMAP_SIZE), icon, QRect(0,0,PIXMAP_SIZE,PIXMAP_SIZE) );

  painter->restore();

  // Draw text
  QRect rectangle( TEXT_POSITION + 20, option.rect.y(), 300, option.rect.height() );
  
  //painter->fillRect(rectangle, option.palette.highlight());
  painter->drawText( rectangle, Qt::AlignLeft | Qt::AlignVCenter, url.fileName() );

  // Draw date
  QRect dateRectangle( TEXT_POSITION + 20 + 300, option.rect.y(), 300, option.rect.height() );
  
  //painter->fillRect(rectangle, option.palette.highlight());
  painter->drawText( dateRectangle, Qt::AlignLeft | Qt::AlignVCenter, modified );

  // Draw switch icon

  QPixmap switchIcon( "info.png" );

  painter->save();

  painter->translate( option.rect.x() + option.rect.width() - LIST_HEIGHT, option.rect.topLeft().y() + 3 );
  painter->drawPixmap( QRect(0,0,14,14), switchIcon.scaled(14,14), QRect(0,0,14,14) );

  painter->restore();
}

bool ItemDelegate::editorEvent ( QEvent * event, QAbstractItemModel * model, const QStyleOptionViewItem & option, const QModelIndex & index )
{
  qDebug() << "ItemDelegate::editorEvent " << event;
  return QItemDelegate::editorEvent( event, model, option, index );
}

#include "ItemDelegate.moc"
