/*
   Copyright (c) 2009 Sebastian Trueg <trueg@kde.org>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "daterangeselectionwidget.h"
#include "ui_daterangeselectionwidget.h"
#include "daterange.h"

#include <QtGui/QToolButton>
#include <QtGui/QCalendarWidget>
#include <QtGui/QButtonGroup>
#include <QtGui/QDateEdit>
#include <QtGui/QTextCharFormat>
#include <QtGui/QPalette>


class DateRangeSelectionWidget::Private : public Ui::DateRangeSelectionWidgetBase
{
public:
    void _k_rangeButtonClicked( QAbstractButton* button );
    void _k_rangeStartEdited( const QDate& date );
    void _k_rangeEndEdited( const QDate& date );
    void _k_calendarDateClicked( const QDate& date );

    void updateCalendar( const DateRange& range );
    void updateEditBoxes( const DateRange& range );

    void setRange( const DateRange& range );

    DateRange m_range;

    DateRangeSelectionWidget* q;
};


void DateRangeSelectionWidget::Private::_k_rangeButtonClicked( QAbstractButton* button )
{
    if ( button == m_todayButton )
        setRange( DateRange::today() );
    else if ( button == m_thisWeekButton )
        setRange( DateRange::thisWeek() );
    else
        setRange( DateRange() );

    emit q->rangeChanged( m_range );
}


void DateRangeSelectionWidget::Private::_k_rangeStartEdited( const QDate& date )
{
    if ( date > m_range.end() )
        setRange( DateRange( date, date ) );
    else
        setRange( DateRange( date, m_range.end() ) );

    emit q->rangeChanged( m_range );
}


void DateRangeSelectionWidget::Private::_k_rangeEndEdited( const QDate& date )
{
    if ( date < m_range.start() )
        setRange( DateRange( date, date ) );
    else
        setRange( DateRange( m_range.start(), date ) );

    emit q->rangeChanged( m_range );
}


void DateRangeSelectionWidget::Private::_k_calendarDateClicked( const QDate& date )
{
    if ( QApplication::keyboardModifiers() & Qt::ShiftModifier ) {
        if ( date < m_range.start() ) {
            setRange( DateRange( date, m_range.end() ) );
            emit q->rangeChanged( m_range );
        }
        else if ( date > m_range.end() ) {
            setRange( DateRange( m_range.start(), date ) );
            emit q->rangeChanged( m_range );
        }
    }
    else {
        setRange( DateRange( date, date ) );
        emit q->rangeChanged( m_range );
    }
}


void DateRangeSelectionWidget::Private::updateCalendar( const DateRange& range )
{
    m_calendar->setDateTextFormat( QDate(), QTextCharFormat() );
    m_calendar->setSelectedDate( range.isValid() ? range.start() : QDate::currentDate() );
    if ( range.isValid() ) {
        QTextCharFormat selectedFormat;
        selectedFormat.setBackground( q->palette().color( QPalette::Highlight ) );
        selectedFormat.setForeground( q->palette().color( QPalette::HighlightedText ) );
        for ( QDate date = range.start(); date <= range.end(); date = date.addDays( 1 ) ) {
            m_calendar->setDateTextFormat( date, selectedFormat );
        }
    }
}


void DateRangeSelectionWidget::Private::updateEditBoxes( const DateRange& range )
{
    m_rangeStartEdit->setDate( range.start() );
    m_rangeEndEdit->setDate( range.end() );
}


void DateRangeSelectionWidget::Private::setRange( const DateRange& range )
{
    m_range = range;

    m_dateSelectionRange->setExclusive( true );
    if ( !range.isValid() ) {
        m_anytimeButton->setChecked( true );
    }
    else if ( range == DateRange::today() ) {
        m_todayButton->setChecked( true );
    }
    else if ( range == DateRange::thisWeek() ) {
        m_thisWeekButton->setChecked( true );
    }
    else {
        m_dateSelectionRange->setExclusive( false );
        foreach( QAbstractButton* button, m_dateSelectionRange->buttons() )
            button->setChecked( false );
    }

    updateCalendar( range );
    updateEditBoxes( range );
}


DateRangeSelectionWidget::DateRangeSelectionWidget( QWidget* parent )
    : QWidget( parent ),
      d( new Private() )
{
    d->q = this;
    d->setupUi( this );
    d->m_calendar->setFirstDayOfWeek( Qt::DayOfWeek( KGlobal::locale()->weekStartDay() ) );
    connect( d->m_dateSelectionRange, SIGNAL( buttonClicked( QAbstractButton* ) ),
             this, SLOT( _k_rangeButtonClicked( QAbstractButton* ) ) );
    connect( d->m_rangeStartEdit, SIGNAL( dateChanged( QDate ) ),
             this, SLOT( _k_rangeStartEdited( QDate ) ) );
    connect( d->m_rangeEndEdit, SIGNAL( dateChanged( QDate ) ),
             this, SLOT( _k_rangeEndEdited( QDate ) ) );
    connect( d->m_calendar, SIGNAL( clicked( QDate ) ),
             this, SLOT( _k_calendarDateClicked( QDate ) ) );

    setRange( DateRange() );
}


DateRangeSelectionWidget::~DateRangeSelectionWidget()
{
}


DateRange DateRangeSelectionWidget::range() const
{
    return d->m_range;
}


void DateRangeSelectionWidget::setRange( const DateRange& range )
{
    d->setRange( range );
}

#include "daterangeselectionwidget.moc"
