/***************************************************************************
 *   Copyright 2008 by Thomas Schildknecht <danakil.kde@gmail.com>         *
 *   Copyright 2010 by Stefan Majewsky <majewsky@gmx.net>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

#include "button.h"
#include "windowmanager.h"

#include <KLocalizedString>
#include <Plasma/Svg>

//BEGIN Awc::ButtonMetadata

Awc::ButtonMetadata::ButtonMetadata(Awc::ButtonType buttonType)
    : m_buttonType(buttonType)
{
    switch (m_buttonType) {
        case Awc::CloseButton:
            m_elementIdentifier = QLatin1String("close");
            m_actionString = QLatin1String("close");
            m_actionText = i18n("Close");
            break;
        case Awc::MaximizeButton:
            m_elementIdentifier = QLatin1String("maximize");
            m_actionString = QLatin1String("toggleMaximized");
            m_actionText = i18n("Maximize");
            break;
        case Awc::MinimizeButton:
            m_elementIdentifier = QLatin1String("minimize");
            m_actionString = QLatin1String("toggleMinimized");
            m_actionText = i18n("Minimize");
            break;
        case Awc::ShadeButton:
            m_elementIdentifier = QLatin1String("shade");
            m_actionString = QLatin1String("toggleShaded");
            m_actionText = i18n("Shade"); //krazy:exclude=i18ncheckarg to use the standard translation
            break;
        case Awc::KeepBelowButton:
            m_elementIdentifier = QLatin1String("keepbelow");
            m_actionString = QLatin1String("toggleKeptBelowOthers");
            m_actionText = i18n("Keep below others");
            break;
        case Awc::KeepAboveButton:
            m_elementIdentifier = QLatin1String("keepabove");
            m_actionString = QLatin1String("toggleAlwaysOnTop");
            m_actionText = i18n("Keep above others");
            break;
        default:
            m_actionText = i18n("--- Separator ---");
            break;
    }
}

QIcon Awc::ButtonMetadata::renderedIcon() const
{
    //TODO: This could perhaps use some optimization.
    Plasma::Svg svg;
    svg.setImagePath("widgets/activewindowcontrol");
    return QIcon(svg.pixmap(m_elementIdentifier + QLatin1String("-normalstate")));
}

//END Awc::ButtonMetadata
//BEGIN Awc::Button

static const int SeparatorExtent = 8;
static const int MinimumExtent = 16;

Awc::Button::Button(Awc::ButtonType buttonType, Awc::WindowManager* windowManager, QGraphicsItem* parent)
    : Plasma::IconWidget(parent)
    , m_metadata(buttonType)
    , m_pushed(false)
{
    if (m_metadata.m_buttonType == Awc::SeparatorButton) {
        setMaximumSize(SeparatorExtent, SeparatorExtent);
    } else {
        setMinimumSize(MinimumExtent, MinimumExtent);
    }
    switch ((int) m_metadata.m_buttonType) {
        case MaximizeButton:
            connect(windowManager, SIGNAL(maximizedChanged(bool)), this, SLOT(setPushed(bool)));
            break;
        case ShadeButton:
            connect(windowManager, SIGNAL(shadedChanged(bool)), this, SLOT(setPushed(bool)));
            break;
        case KeepBelowButton:
            connect(windowManager, SIGNAL(keptBelowChanged(bool)), this, SLOT(setPushed(bool)));
            break;
        case KeepAboveButton:
            connect(windowManager, SIGNAL(keptAboveChanged(bool)), this, SLOT(setPushed(bool)));
            break;
    }
    connect(windowManager, SIGNAL(activeWindowChanged(bool)), this, SLOT(setEnabled(bool)));
    connect(this, SIGNAL(actionTriggered(const QString&)), windowManager, SLOT(triggerActiveWindowAction(const QString&)));
    setDrawBackground(false);
    updateAppearance();
    connect(this, SIGNAL(clicked()), this, SLOT(slotClicked()));
}

void Awc::Button::updateAppearance()
{
    static const QString svgPath = QLatin1String("widgets/activewindowcontrol");
    if (m_pushed) {
        setSvg(svgPath, m_metadata.m_elementIdentifier + QLatin1String("-pushedstate"));
    } else {
        setSvg(svgPath, m_metadata.m_elementIdentifier + QLatin1String("-normalstate"));
    }
}

Awc::ButtonType Awc::Button::buttonType() const
{
    return m_metadata.m_buttonType;
}

void Awc::Button::setEnabled(bool enabled)
{
    //TODO: visual indication
    //FIXME: The following does not work because events won't propagate properly.
    //Plasma::IconWidget::setEnabled(enabled);
    setPushed(false);
}

bool Awc::Button::isPushed() const
{
    return m_pushed;
}

void Awc::Button::setPushed(bool pushed)
{
    m_pushed = pushed;
    updateAppearance();
}

void Awc::Button::slotClicked()
{
    if (!m_metadata.m_actionString.isEmpty())
        emit actionTriggered(m_metadata.m_actionString);
}

//END Awc::Button

#include "button.moc"
