/***************************************************************************
 *   Copyright 2008 by Thomas Schildknecht <danakil.kde@gmail.com>         *
 *   Copyright 2010 by Stefan Majewsky <majewsky@gmx.net>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

#include "windowmanager.h"

#include <KWindowSystem>

Awc::WindowManager::WindowManager(Plasma::DataEngine* tasksEngine, QObject* parent)
    : QObject(parent)
    , m_cyclingStrategy(Awc::CyclingEnabled | Awc::CycleOnAllDesktops | Awc::IncludeMinimizedWindows)
    , m_reactivateActiveWindowAfterOperations(true)
    , m_tasksEngine(tasksEngine)
    , m_activeWindow(0)
    , m_activeDesktop(0) //because there is no active window
{
    connect(m_tasksEngine, SIGNAL(sourceAdded(const QString&)), this, SLOT(sourceAdded(const QString&)));
    connect(m_tasksEngine, SIGNAL(sourceRemoved(const QString&)), this, SLOT(sourceRemoved(const QString&)));
    m_tasksEngine->connectAllSources(this);
}

Awc::CyclingStrategy Awc::WindowManager::cyclingStrategy() const
{
    return m_cyclingStrategy;
}

void Awc::WindowManager::setCyclingStrategy(Awc::CyclingStrategy strategy)
{
    m_cyclingStrategy = strategy;
}

void Awc::WindowManager::setReactivateActiveWindowAfterOperations(bool reactivate)
{
    m_reactivateActiveWindowAfterOperations = reactivate;
}

void Awc::WindowManager::dataUpdated(const QString& name, const Plasma::DataEngine::Data& data)
{
    const WId wid = name.toLongLong();
    //check desktop association
    const int desktop = data["desktop"].toInt();
    QMutableMapIterator<int, QList<WId> > windowOrderIterator(m_windowOrder);
    bool correctDesktopAssociation = false;
    while (windowOrderIterator.hasNext()) {
        QList<WId>& desktopWindowList = windowOrderIterator.next().value();
        if (desktopWindowList.contains(wid)) {
            const int oldDesktop = windowOrderIterator.key();
            if (oldDesktop == desktop) {
                correctDesktopAssociation = true;
                break; //correct association
            }
            //start to fix wrong association
            desktopWindowList.removeAll(wid);
            if (desktopWindowList.isEmpty()) { //this desktop is empty now
                windowOrderIterator.remove();
            }
            break;
        }
    }
    if (!correctDesktopAssociation) {
        m_windowOrder[desktop] << wid;
    }
    //update window state
    Awc::WindowState state;
    if (data["shaded"].toBool()) {
        state |= Awc::IsShaded;
    }
    if (data["alwaysOnTop"].toBool()) {
        state |= Awc::IsKeptAbove;
    }
    if (data["keptBelowOthers"].toBool()) {
        state |= Awc::IsKeptBelow;
    }
    if (data["maximized"].toBool()) {
        state |= Awc::IsMaximized;
    }
    if (data["minimized"].toBool()) {
        state |= Awc::IsMinimized;
    }
    if (m_windowStates.value(wid) & Awc::IsTemporarilyRestored) {
        state |= Awc::IsTemporarilyRestored;
    }
    m_windowStates[wid] = state;
    //check for changes in "active" property
    const int isActive = data["active"].toBool();
    if (!isActive && m_activeWindow == wid) {
        m_activeWindow = m_activeDesktop = 0;
        publishActiveWindowState();
    } else if (isActive) {
        m_activeWindow = wid;
        if (desktop != -1) {
            m_activeDesktop = desktop;
        }
        publishActiveWindowState();
    }
}

void Awc::WindowManager::publishActiveWindowState()
{
    if (!m_activeWindow) {
        emit activeWindowChanged(false);
    } else {
        emit activeWindowChanged(true);
        const Awc::WindowState state = m_windowStates[m_activeWindow];
        emit shadedChanged(state & Awc::IsShaded);
        emit keptAboveChanged(state & Awc::IsKeptAbove);
        emit keptBelowChanged(state & Awc::IsKeptBelow);
        emit maximizedChanged(state & Awc::IsMaximized);
    }
}

void Awc::WindowManager::sourceAdded(const QString& source)
{
    m_tasksEngine->connectSource(source, this);
}

void Awc::WindowManager::sourceRemoved(const QString& source)
{
    m_tasksEngine->disconnectSource(source, this);
    const WId wid = source.toLongLong();
    //remove deleted window from internal state storage
    m_windowStates.remove(wid);
    //remove deleted window from desktop association map
    QMutableMapIterator<int, QList<WId> > windowOrderIterator(m_windowOrder);
    while (windowOrderIterator.hasNext()) {
        QList<WId>& desktopWindowList = windowOrderIterator.next().value();
        if (desktopWindowList.contains(wid)) {
            desktopWindowList.removeAll(wid);
            if (desktopWindowList.isEmpty()) { //this desktop is empty now
                windowOrderIterator.remove();
            }
            break;
        }
    }
}

void Awc::WindowManager::triggerActiveWindowAction(const QString& actionName)
{
    if (!m_activeWindow) {
        return;
    }
    executeWindowAction(m_activeWindow, actionName);
    if (m_reactivateActiveWindowAfterOperations) {
        if (actionName != QLatin1String("close") && actionName != QLatin1String("toggleMinimized")) {
            executeWindowAction(m_activeWindow, "activate");
        }
    }
}

void Awc::WindowManager::executeWindowAction(WId wid, const QString& actionName)
{
    //TODO: check for memleaks
    Plasma::Service* service = m_tasksEngine->serviceForSource(QString::number(wid));
    if (service) {
        service->startOperationCall(service->operationDescription(actionName));
    }
}

void Awc::WindowManager::cycleToAdjacentWindow(int direction)
{
    if (!(m_cyclingStrategy & Awc::CyclingEnabled)) {
        return;
    }
    //NOTE: desktop == -1 is perfectly possible, refers to windows on all desktops.
    //FIXME: When cycling reaches a window with desktop == -1, the cyclingOrder generation may/will break.
    //normalize direction
    if (direction > 0) {
        direction = 1;
    } else {
        direction = -1;
    }
    //build cycling order from strategy
    const int activeDesktop = (m_activeDesktop == 0) ? KWindowSystem::currentDesktop() : m_activeDesktop;
    QList<WId> cyclingOrder;
    if (m_cyclingStrategy & Awc::CycleOnAllDesktops) {
        //cycle on all desktops
        QMapIterator<int, QList<WId> > iterator(m_windowOrder);
        while (iterator.hasNext()) {
            cyclingOrder << iterator.next().value();
        }
    } else {
        cyclingOrder = m_windowOrder.value(activeDesktop);
        cyclingOrder << m_windowOrder.value(-1);
    }
    if (!(m_cyclingStrategy & Awc::IncludeMinimizedWindows)) {
        //remove minimized windows from list
        QMutableListIterator<WId> iterator(cyclingOrder);
        while (iterator.hasNext()) {
            if (m_windowStates.value(iterator.next()) & Awc::IsMinimized) {
                iterator.remove();
            }
        }
    }
    kDebug() << m_windowOrder << cyclingOrder;
    if (cyclingOrder.isEmpty()) {
        return;
    }
    //find next window
    const int cyclingOrderOldId = cyclingOrder.indexOf(m_activeWindow);
    int cyclingOrderId = (cyclingOrderOldId + 1) % cyclingOrder.size();
    //exception: when no window is active (i.e. cyclingOrderId == -1), try windows from the current desktop first
    //NOTE: This is pointless if CycleOnAllDesktops is disabled, because cyclingOrder contains only windows from the current desktop.
    kDebug() << "Direct:" << cyclingOrderOldId << cyclingOrderId;
    if (cyclingOrderOldId < 0 && (m_cyclingStrategy & Awc::CycleOnAllDesktops)) {
        const QList<WId> currentDesktopWindows = m_windowOrder.value(activeDesktop) + m_windowOrder.value(-1);
        kDebug() << "Current desktop:" << activeDesktop << currentDesktopWindows;
        if (!currentDesktopWindows.isEmpty()) {
            while (!currentDesktopWindows.contains(cyclingOrder[cyclingOrderId])) {
                ++cyclingOrderId;
            }
        }
    }
    kDebug() << "Using:" << cyclingOrderId;
    //find new active window
    WId newActiveWindow = cyclingOrder[cyclingOrderId];
    if (newActiveWindow == m_activeWindow) {
        return;
    }
    //minimize current window if it was restored temporarily only
    if (m_windowStates.value(m_activeWindow) & Awc::IsTemporarilyRestored) {
        m_windowStates[m_activeWindow] &= ~Awc::IsTemporarilyRestored;
        if (!(m_windowStates.value(m_activeWindow) & Awc::IsMinimized)) {
            executeWindowAction(m_activeWindow, "toggleMinimized");
        }
    }
    //activate next window
    if (m_windowStates.value(newActiveWindow) & Awc::IsMinimized) {
        m_windowStates[newActiveWindow] |= Awc::IsTemporarilyRestored;
    }
    executeWindowAction(newActiveWindow, "activate");
}

#include "windowmanager.moc"
