/*****************************************************************************
 *   Copyright (C) 2009-2010 by Matthieu Gallien <matthieu_gallien@yahoo.fr> *
 *                                                                           *
 *   This program is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by    *
 *   the Free Software Foundation; either version 2 of the License, or       *
 *   (at your option) any later version.                                     *
 *                                                                           *
 *   This program is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *   GNU General Public License for more details.                            *
 *                                                                           *
 *   You should have received a copy of the GNU General Public License       *
 *   along with this program; if not, write to the                           *
 *   Free Software Foundation, Inc.,                                         *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .          *
 *****************************************************************************/

#include "tasks_model.h"
#include "tasks_roles.h"

#include <KIcon>
#include <Plasma/Service>

#include <QtGlobal>

FlatTasksModel::FlatTasksModel(Plasma::DataEngine *tasksEngine,
                               Plasma::DataEngine *statusNotifierItemEngine,
                               QObject *parent) :
    QAbstractListModel(parent),
    m_tasksEngine(tasksEngine),
    m_statusNotifierItemEngine(statusNotifierItemEngine),
    m_tasksData(),
    m_statusNotifierItemsData(),
    m_indexFromTaskName(),
    m_indexFromStatusNotifierItemName()
{
    connect(m_tasksEngine, SIGNAL(sourceAdded(const QString&)),
            this, SLOT(taskAdded(const QString&)));
    connect(m_tasksEngine, SIGNAL(sourceRemoved(const QString&)),
            this, SLOT(taskRemoved(const QString&)));
    const QStringList &allTaskSources(m_tasksEngine->sources());
    for (int i = 0; i < allTaskSources.size(); ++i) {
        kDebug() << "add initial task:" << allTaskSources.at(i);
        m_tasksData[allTaskSources.at(i)]["isValid"] = false;
    }
    m_tasksEngine->connectAllSources(this);
    m_tasksEngine->connectSource("virtualDesktops", this);
    m_numberOfDesktops = m_tasksEngine->query("virtualDesktops")["number"].toInt();
    m_desktopNames = m_tasksEngine->query("virtualDesktops")["names"].toList();

    connect(m_statusNotifierItemEngine, SIGNAL(sourceAdded(const QString&)),
            this, SLOT(statusNotifierItemAdded(const QString&)));
    connect(m_statusNotifierItemEngine, SIGNAL(sourceRemoved(const QString&)),
            this, SLOT(statusNotifierItemRemoved(const QString&)));
    const QStringList &allStatusNotifierItemSources(m_statusNotifierItemEngine->sources());
    for (int i = 0; i < allStatusNotifierItemSources.size(); ++i) {
        kDebug() << "add initial status notifier item:" << allStatusNotifierItemSources.at(i);
        m_statusNotifierItemsData[allStatusNotifierItemSources.at(i)]["isValid"] = false;
    }
    m_statusNotifierItemEngine->connectAllSources(this);
}

FlatTasksModel::~FlatTasksModel()
{
}

QVariant FlatTasksModel::data(const QModelIndex &index, int role) const
{
    const QString &taskName = m_allItems[index.row()].first;

    switch (role)
    {
        case Qt::DisplayRole:
            if (m_allItems[index.row()].first.isEmpty()) {
                return m_statusNotifierItemsData[m_allItems[index.row()].second.first()]["Title"];
            } else {
                if (m_tasksData[taskName]["startup"].toBool()) {
                    return m_tasksData[taskName]["text"];
                } else {
                    return m_tasksData[taskName]["visibleName"];
                }
            }
            break;
        case Qt::DecorationRole:
            if (m_allItems[index.row()].second.empty()) {
                return m_tasksData[taskName]["icon"];
            } else {
                return m_statusNotifierItemsData[m_allItems[index.row()].second.first()]["Icon"];
            }
            break;
        case TasksModel::IsMinimizedRole:
            return m_tasksData[taskName]["minimized"];
            break;
        case TasksModel::IsMaximizedRole:
            return m_tasksData[taskName]["maximized"];
            break;
        case TasksModel::IsShadedRole:
            return m_tasksData[taskName]["shaded"];
            break;
        case TasksModel::DesktopNumberRole:
            return m_tasksData[taskName]["desktop"];
            break;
        case TasksModel::IsOnCurrentDesktopRole:
            if (!m_allItems[index.row()].second.empty()) {
                return (m_statusNotifierItemsData[m_allItems[index.row()].second.first()]["Status"].toString() == "Active") || (!m_allItems[index.row()].first.isEmpty());
            } else {
                if (m_tasksData[taskName]["startup"].toBool()) {
                    return true;
                } else {
                    return m_tasksData[taskName]["onCurrentDesktop"];
                }
            }
            break;
        case TasksModel::IsActive:
            return m_tasksData[taskName]["active"];
            break;
        case TasksModel::IsDemandingAttention:
            return m_tasksData[taskName]["demandsAttention"];
            break;
        case TasksModel::BackgroundPrefix:
            return m_tasksData[taskName]["backgroundPrefix"];
            break;
        case TasksModel::ClassName:
            return m_tasksData[taskName]["className"];
            break;
        case TasksModel::ClassClass:
            if (m_tasksData[taskName]["startup"].toBool()) {
                return m_tasksData[taskName]["text"];
            } else {
                return m_tasksData[taskName]["classClass"];
            }
            break;
        case TasksModel::IsFullScreen:
            return m_tasksData[taskName]["fullScreen"];
            break;
        case TasksModel::IsAlwaysOnTop:
            return m_tasksData[taskName]["alwaysOnTop"];
            break;
        case TasksModel::IsKeptBelowOthers:
            return m_tasksData[taskName]["keptBelowOthers"];
            break;
        case TasksModel::IsStartup:
            return m_tasksData[taskName]["startup"];
            break;
        case TasksModel::TaskSourceName:
            return taskName;
            break;
        case TasksModel::StatusItemSourceNames:
            return m_allItems[index.row()].second;
            break;
        case TasksModel::ContextMenuRole:
            if (m_allItems[index.row()].first.isEmpty()) {
                return createContextMenuForStatusItem(m_allItems[index.row()].second.first());
            } else {
                return createContextMenuForTask(taskName);
            }
            break;
        case TasksModel::ActivateActionRole:
            if (m_allItems[index.row()].first.isEmpty()) {
                return createActivateActionForStatusItem(m_allItems[index.row()].second.first());
            } else {
                return createActivateActionForTask(taskName);
            }
            break;
        case TasksModel::SecondaryActivateActionRole:
            if (m_allItems[index.row()].first.isEmpty()) {
                return createSecondaryActivateActionForStatusItem(m_allItems[index.row()].second.first());
            } else {
                return QVariant();
            }
            break;
    }

    return QVariant();
}

int FlatTasksModel::rowCount(const QModelIndex &index) const
{
    Q_UNUSED(index);
    if (index.isValid()) {
        return 0;
    } else {
        return m_allItems.size();
    }
}

void FlatTasksModel::taskAdded(const QString &taskName)
{
    if (taskName != "numberOfDesktops") {
        kDebug() << taskName;
        m_tasksData[taskName]["isValid"] = false;
        m_tasksEngine->connectSource(taskName, this);
    }
}

void FlatTasksModel::taskRemoved(const QString &taskName)
{
    removeItemFromWindowID(taskName);
}

void FlatTasksModel::dataUpdated(const QString &sourceName, const Plasma::DataEngine::Data &sourceData)
{
    if (m_tasksData.contains(sourceName)) {
        if (m_tasksData[sourceName]["isValid"].toBool()) {
            m_tasksData[sourceName] = sourceData;
            m_tasksData[sourceName]["isValid"] = true;
            emit dataChanged(index(m_indexFromTaskName[sourceName], 0), index(m_indexFromTaskName[sourceName], 0));
            updateBackgroundPrefix(sourceName, sourceData);
        } else {
            if (sourceData.contains("className") || sourceData.contains("bin")) {
                kDebug() << "add a new task" << sourceName;
                addItemFromWindowID(sourceName, sourceData);
            }
        }
    } else if (m_statusNotifierItemsData.contains(sourceName)) {
        if (m_statusNotifierItemsData[sourceName]["isValid"].toBool()) {
            m_statusNotifierItemsData[sourceName] = sourceData;
            emit dataChanged(index(m_indexFromStatusNotifierItemName[sourceName], 0), index(m_indexFromStatusNotifierItemName[sourceName], 0));
            m_statusNotifierItemsData[sourceName]["isValid"] = true;
        } else {
            if (sourceData["Category"].toString() == "ApplicationStatus") {
                addItemFromStatusNotifierItem(sourceName, sourceData);
            }
        }
    } else if (sourceName == "virtualDesktops") {
        m_numberOfDesktops = sourceData["number"].toInt();
        m_desktopNames = sourceData["names"].toList();
    } else {
        kDebug() << "unknown source name:" << sourceName;
    }
}

void FlatTasksModel::updateBackgroundPrefix(const QString &taskName, const Plasma::DataEngine::Data &taskData)
{
    if (taskData.contains("demandsAttention")) {
        if (taskData["demandsAttention"].toBool()) {
            if (m_tasksData[taskName]["backgroundPrefix"] != "attention") {
                m_tasksData[taskName]["backgroundPrefix"] = "attention";
                emit dataChanged(index(m_indexFromTaskName[taskName], 0), index(m_indexFromTaskName[taskName], 0));
            }
        } else {
            if (taskData["active"].toBool()) {
                if (m_tasksData[taskName]["backgroundPrefix"] != "focus") {
                    m_tasksData[taskName]["backgroundPrefix"] = "focus";
                    emit dataChanged(index(m_indexFromTaskName[taskName], 0), index(m_indexFromTaskName[taskName], 0));
                }
            } else {
                if (m_tasksData[taskName]["backgroundPrefix"] != "normal") {
                    m_tasksData[taskName]["backgroundPrefix"] = "normal";
                    emit dataChanged(index(m_indexFromTaskName[taskName], 0), index(m_indexFromTaskName[taskName], 0));
                }
            }
        }
    } else {
        if (m_tasksData[taskName]["backgroundPrefix"] != "normal") {
            m_tasksData[taskName]["backgroundPrefix"] = "normal";
            emit dataChanged(index(m_indexFromTaskName[taskName], 0), index(m_indexFromTaskName[taskName], 0));
        }
    }
}

void FlatTasksModel::updateCachedIndex(int removedIndex)
{
    for (int index = removedIndex; index < m_allItems.size() - 1; ++index) {
        if (!m_allItems.at(index + 1).first.isEmpty()) {
            m_indexFromTaskName[m_allItems[index + 1].first] = index;
        }
        if (!m_allItems.at(index + 1).second.empty()) {
            const QStringList &allStatusItems(m_allItems[index + 1].second);
            for (int j = 0; j < allStatusItems.size(); ++j) {
                m_indexFromStatusNotifierItemName[m_allItems[index + 1].second.at(j)] = index;
            }
        }
        m_allItems[index] = m_allItems[index + 1];
    }
}

void FlatTasksModel::statusNotifierItemAdded(const QString &itemName)
{
    m_statusNotifierItemsData[itemName]["isValid"] = false;
    m_statusNotifierItemEngine->connectSource(itemName, this);
}

void FlatTasksModel::statusNotifierItemRemoved(const QString &itemName)
{
    removeItemFromStatusNotifierItem(itemName);
}

bool FlatTasksModel::statusNotifierItemHaveMatchingTasks(const QString &itemName) const
{
    const QString &windowId(m_statusNotifierItemsData[itemName]["WindowId"].toString());
    return m_tasksData.contains(windowId);
}

bool FlatTasksModel::statusNotifierItemIsApplication(const QString &itemName) const
{
    return m_statusNotifierItemsData[itemName]["Category"].toString() == "ApplicationStatus";
}

bool FlatTasksModel::statusNotifierItemIsActive(const QString &itemName) const
{
    return m_statusNotifierItemsData[itemName]["Status"].toString() == "Active";
}

void FlatTasksModel::addItemFromWindowID(const QString &windowId,
                                         const Plasma::DataEngine::Data &taskData)
{
    bool alreadyExists = false;
    for (int i = 0; i < m_allItems.size() && !alreadyExists; ++i) {
        const QStringList &allStatusNotifierItem(m_allItems.at(i).second);
        for (int j = 0; j < allStatusNotifierItem.size() && !alreadyExists; ++j) {
            if (m_statusNotifierItemsData[allStatusNotifierItem.at(j)]["WindowId"].toString() == windowId) {
                m_allItems[i].first = windowId;
                m_indexFromTaskName[windowId] = i;
                m_tasksData[windowId] = taskData;
                m_tasksData[windowId]["isValid"] = true;
                emit dataChanged(index(i, 0), index(i, 0));
                kDebug() << "an item with windowId:" << windowId << "already exists";
                alreadyExists = true;
            }
        }
    }

    if (!alreadyExists) {
        const int newIndex = m_allItems.size();
        kDebug() << "insert a new item at index:" << newIndex;
        beginInsertRows(QModelIndex(), newIndex, newIndex);
        m_indexFromTaskName[windowId] = newIndex;
        m_tasksData[windowId] = taskData;
        m_tasksData[windowId]["isValid"] = true;
        itemInfo_type newItem;
        newItem.first = windowId;
        m_allItems.append(newItem);
        endInsertRows();
        updateBackgroundPrefix(windowId, taskData);
    }
}

void FlatTasksModel::removeItemFromWindowID(const QString &windowId)
{
    int oldIndex = m_indexFromTaskName[windowId];
    kDebug() << "the old index of the removed window is:" << oldIndex;
    if (m_allItems.at(oldIndex).second.empty()) {
        kDebug() << "remove an item because it is a window not associated to a status item";
        beginRemoveRows(QModelIndex(), oldIndex, oldIndex);
        m_tasksEngine->disconnectSource(windowId, this);
        updateCachedIndex(oldIndex);
        m_tasksData.remove(windowId);
        m_indexFromTaskName.remove(windowId);
        m_allItems.removeLast();
        endRemoveRows();
    } else {
        kDebug() << "remove window associated to a status item";
        m_tasksEngine->disconnectSource(windowId, this);
        m_tasksData.remove(windowId);
        m_indexFromTaskName.remove(windowId);
        m_allItems[oldIndex].first.clear();
        emit dataChanged(index(oldIndex, 0), index(oldIndex, 0));
    }
}

void FlatTasksModel::addItemFromStatusNotifierItem(const QString &itemName,
                                                   const Plasma::DataEngine::Data &itemData)
{
    bool alreadyExists = false;
    for (int i = 0; i < m_allItems.size() && !alreadyExists; ++i) {
        const QString &windowId(m_allItems.at(i).first);
        if (itemData["WindowId"] == windowId) {
            m_allItems[i].second.append(itemName);
            m_statusNotifierItemsData[itemName] = itemData;
            m_statusNotifierItemsData[itemName]["isValid"] = true;
            m_indexFromStatusNotifierItemName[itemName] = i;
            emit dataChanged(index(i, 0), index(i, 0));
            kDebug() << "an item with windowId:" << windowId << "already exists";
            alreadyExists = true;
        }
    }

    if (!alreadyExists) {
        kDebug() << "insert a new item";
        const int newIndex = m_allItems.size();
        beginInsertRows(QModelIndex(), newIndex, newIndex);
        m_indexFromStatusNotifierItemName[itemName] = newIndex;
        m_statusNotifierItemsData[itemName] = itemData;
        m_statusNotifierItemsData[itemName]["isValid"] = true;
        itemInfo_type newItem;
        newItem.second.append(itemName);
        m_allItems.append(newItem);
        endInsertRows();
    }
}

void FlatTasksModel::removeItemFromStatusNotifierItem(const QString &itemName)
{
    int oldIndex = m_indexFromStatusNotifierItemName[itemName];
    kDebug() << "the old index of the removed status item is:" << oldIndex;
    if (m_allItems.at(oldIndex).first.isEmpty()) {
        kDebug() << "remove an item because it is a status item without associated window";
        debugPrint();
        beginRemoveRows(QModelIndex(), oldIndex, oldIndex);
        m_statusNotifierItemEngine->disconnectSource(itemName, this);
        updateCachedIndex(oldIndex);
        m_statusNotifierItemsData.remove(itemName);
        m_indexFromStatusNotifierItemName.remove(itemName);
        m_allItems.removeLast();
        endRemoveRows();
        debugPrint();
    } else {
        kDebug() << "remove status item associated to a window";
        m_statusNotifierItemEngine->disconnectSource(itemName, this);
        m_statusNotifierItemsData.remove(itemName);
        m_indexFromStatusNotifierItemName.remove(itemName);
        m_allItems[oldIndex].second.removeOne(itemName);
        emit dataChanged(index(oldIndex, 0), index(oldIndex, 0));
    }
}

void FlatTasksModel::debugPrint() const
{
    for (int i = 0; i < m_allItems.size(); ++i) {
        kDebug() << "item at place" << i << "is:" << m_allItems.at(i);
    }
}

QVariant FlatTasksModel::createContextMenuForStatusItem(const QString &statusItemName) const
{
    Q_UNUSED(statusItemName);
    
    QVariantMap theAction;
    theAction["operationName"] = "ContextMenu";
    theAction["useTaskService"] = false;
    return theAction;
}

QList<QVariant> FlatTasksModel::createContextMenuForTask(const QString &taskName) const
{
    QList<QVariant> res;
    const Plasma::DataEngine::Data &theTaskInfo(m_tasksData[taskName]);

    if (m_numberOfDesktops > 1) {
        {
            QList<QVariant> subMenuItems;

            {
                /** Send a Task to all Desktops*/
                QVariantMap theAction;
                theAction["menuName"] = i18n("&All Desktops");
                theAction["operationName"] = "toDesktop";
                theAction["enabled"] = theTaskInfo["actionChangeDesktop"];
                theAction["checked"] = theTaskInfo["onAllDesktops"];
                theAction["useTaskService"] = true;
                QVariantMap extraParams;
                extraParams["desktop"] = 0;
                theAction["extraOperationParameters"] = extraParams;
                subMenuItems.append(theAction);
            }

            {
                QVariantMap theAction;
                theAction["isSeparator"] = true;
                subMenuItems.append(theAction);
            }

            for (int desktop = 1; desktop <= m_numberOfDesktops; ++desktop) {
                /** Send a Task to all Desktops*/
                QVariantMap theAction;
                theAction["menuName"] = QString("&%1 %2").arg(desktop).arg(m_desktopNames.at(desktop - 1).toString().replace('&', "&&"));
                theAction["operationName"] = "toDesktop";
                theAction["enabled"] = theTaskInfo["actionChangeDesktop"];
                theAction["checked"] = !theTaskInfo["onAllDesktops"].toBool() && theTaskInfo["desktop"].toInt() == desktop;
                theAction["useTaskService"] = true;
                QVariantMap extraParams;
                extraParams["desktop"] = desktop;
                theAction["extraOperationParameters"] = extraParams;
                subMenuItems.append(theAction);
            }

            QVariantMap theAction;
            theAction["menuName"] = i18n("To &Desktop");
            theAction["subMenuItems"] = subMenuItems;
            res.append(theAction);
        }

        {
            /** Send a Task to the current Desktop*/
            QVariantMap theAction;
            theAction["menuName"] = i18n("&To Current Desktop");
            theAction["operationName"] = "toCurrentDesktop";
            theAction["enabled"] = !theTaskInfo["onCurrentDesktop"].toBool() && theTaskInfo["actionChangeDesktop"].toBool();
            theAction["useTaskService"] = true;
            res.append(theAction);
        }
    }

    {
        /** Keep a Window or all windows in a group above the rest */
        QVariantMap theAction;
        theAction["menuName"] = i18n("&Move");
        theAction["icon"] = KIcon("transform-move");
        theAction["operationName"] = "move";
        theAction["enabled"] = theTaskInfo["actionMove"];
        theAction["useTaskService"] = true;
        res.append(theAction);
    }

    {
        /** Keep a Window or all windows in a group above the rest */
        QVariantMap theAction;
        theAction["menuName"] = i18n("Re&size");
        theAction["operationName"] = "resize";
        theAction["enabled"] = theTaskInfo["actionResize"];
        theAction["useTaskService"] = true;
        res.append(theAction);
    }

    {
        /** Keep a Window or all windows in a group above the rest */
        QVariantMap theAction;
        theAction["menuName"] = i18n("Mi&nimize");
        theAction["checked"] = theTaskInfo["minimized"];
        theAction["enabled"] = theTaskInfo["actionMinimize"];
        theAction["operationName"] = "toggleMinimized";
        theAction["useTaskService"] = true;
        res.append(theAction);
    }

    {
        /** Keep a Window or all windows in a group above the rest */
        QVariantMap theAction;
        theAction["menuName"] = i18n("Ma&ximize");
        theAction["operationName"] = "toggleMaximized";
        theAction["checked"] = theTaskInfo["maximized"];
        theAction["enabled"] = theTaskInfo["actionMaximize"];
        theAction["useTaskService"] = true;
        res.append(theAction);
    }

    {
        /** Keep a Window or all windows in a group above the rest */
        QVariantMap theAction;
        theAction["menuName"] = i18n("&Shade");
        theAction["operationName"] = "toggleShaded";
        theAction["checked"] = theTaskInfo["shaded"];
        theAction["enabled"] = theTaskInfo["actionShade"];
        theAction["useTaskService"] = true;
        res.append(theAction);
    }

    {
        QList<QVariant> subMenuItems;

        {
            /** Keep a Window or all windows in a group above the rest */
            QVariantMap theAction;
            theAction["menuName"] = i18n("Keep &Above Others");
            theAction["icon"] = KIcon("go-up");
            theAction["checked"] = theTaskInfo["alwaysOnTop"];
            theAction["operationName"] = "toggleAlwaysOnTop";
            theAction["useTaskService"] = true;
            subMenuItems.append(theAction);
        }

        {
            /** Keep a Window or all windows in a group below the rest*/
            QVariantMap theAction;
            theAction["menuName"] = i18n("Keep &Above Others");
            theAction["icon"] = KIcon("go-down");
            theAction["checked"] = theTaskInfo["keptBelowOthers"];
            theAction["operationName"] = "toggleKeptBelowOthers";
            theAction["useTaskService"] = true;
            subMenuItems.append(theAction);
        }

        {
            /** Set a window or all windows in a group to FullScreen*/
            QVariantMap theAction;
            theAction["menuName"] = i18n("&Fullscreen");
            theAction["icon"] = KIcon("view-fullscreen");
            theAction["checked"] = theTaskInfo["fullScreen"];
            theAction["enabled"] = theTaskInfo["actionFullScreen"];
            theAction["operationName"] = "toggleFullScreen";
            theAction["useTaskService"] = true;
            subMenuItems.append(theAction);
        }

        QVariantMap theAction;
        theAction["menuName"] = i18n("Ad&vanced");
        theAction["subMenuItems"] = subMenuItems;
        res.append(theAction);
    }

    {
        QVariantMap theAction;
        theAction["isSeparator"] = true;
        res.append(theAction);
    }

    {
        /** Keep a Window or all windows in a group above the rest */
        QVariantMap theAction;
        theAction["menuName"] = i18n("&Close");
        theAction["icon"] = KIcon("window-close");
        theAction["operationName"] = "close";
        theAction["useTaskService"] = true;
        res.append(theAction);
    }

    return res;
}

QVariantMap FlatTasksModel::createActivateActionForStatusItem(const QString &statusItemName) const
{
    Q_UNUSED(statusItemName);

    QVariantMap theAction;
    theAction["operationName"] = "Activate";
    theAction["useTaskService"] = false;
    return theAction;
}

QVariantMap FlatTasksModel::createActivateActionForTask(const QString &taskName) const
{
    Q_UNUSED(taskName);

    QVariantMap theAction;
    theAction["operationName"] = "activateRaiseOrIconify";
    theAction["useTaskService"] = true;
    return theAction;
}

QVariantMap FlatTasksModel::createSecondaryActivateActionForStatusItem(const QString &statusItemName) const
{
    Q_UNUSED(statusItemName);

    QVariantMap theAction;
    theAction["operationName"] = "SecondaryActivate";
    theAction["useTaskService"] = false;
    return theAction;
}
