/*
 *   Copyright (C) 2007 Ivan Cukic <ivan.cukic+kde@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser/Library General Public License version 2,
 *   or (at your option) any later version, as published by the Free
 *   Software Foundation
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Lesser/Library General Public License for more details
 *
 *   You should have received a copy of the GNU Lesser/Library General Public
 *   License along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <statesicon.h>
#include <KDebug>

StatesIcon::StatesIcon(QGraphicsItem * parent, Qt::WindowFlags wFlags)
  : QGraphicsWidget(parent, wFlags),
    m_state(Offline), m_buttonState(Normal)
{
    setAcceptsHoverEvents(true);
}

StatesIcon::~StatesIcon()
{
}

void StatesIcon::registerState(State state, KIcon icon)
{
    m_states[state] = icon;
    if (m_states.size() == 1) {
        setState(state);
    }
}

void StatesIcon::setState(State state)
{
    m_state = state;
    update();
}

StatesIcon::State StatesIcon::state() const
{
    return m_state;
}

void StatesIcon::setBackgroundSvg(const QString & svg)
{
    m_svg.setImagePath(svg);
    kDebug() << m_svg.isValid();
    setSvgPrefix();
    update();
}

void StatesIcon::setIconSize(const QSize & iconSize)
{
    m_iconSize = iconSize;
    update();
}

QSize StatesIcon::iconSize() const
{
    return m_iconSize;
}

void StatesIcon::setSvgPrefix()
{
    QString prefix1, prefix2;

    switch (m_state) {
        case Offline:
            prefix1 = "offline";
            break;
        case Starting:
            prefix1 = "starting";
            break;
        case Running:
            prefix1 = "running";
            break;
        case Active:
            prefix1 = "active";
            break;
    }

    switch (m_buttonState) {
        case Normal:
            prefix2 += "normal";
            break;
        case Disabled:
            prefix2 += "disabled";
            break;
        case Pressed:
            prefix2 += "pressed";
            break;
        case Hovered:
            prefix2 += "hovered";
            break;
    }

    if (m_svg.hasElementPrefix(prefix1 + "-" + prefix2)) {
        m_svg.setElementPrefix(prefix1 + "-" + prefix2);
    } else if (m_svg.hasElementPrefix(prefix1)) {
        m_svg.setElementPrefix(prefix1);
    } else {
        m_svg.setElementPrefix(prefix2);
    }
}

void StatesIcon::paint(QPainter * painter, const QStyleOptionGraphicsItem * option, QWidget * widget)
{
    // paints background
    if (m_svg.isValid()) {
        m_svg.resizePanel(size());
        m_svg.paintPanel(painter);
    }

    // paints icons
    KIcon icon;

    if (m_states.contains(m_state)) {
        icon = m_states[m_state];
    } else {
        icon = m_states[Running];
    }
    if (!icon.isNull()) {
        QIcon::Mode mode = QIcon::Normal;

        switch (m_buttonState) {
            case Normal:
                mode = QIcon::Normal;
                break;
            case Disabled:
                mode = QIcon::Disabled;
                break;
            case Pressed:
                mode = QIcon::Selected;
                break;
            case Hovered:
                mode = QIcon::Active;
                break;
        }

        icon.paint(
            painter,
            QRect(QPoint(), size().toSize()),
            Qt::AlignCenter,
            mode,
            QIcon::Off
        );
    }
}

void StatesIcon::hoverEnterEvent(QGraphicsSceneHoverEvent * event)
{
    m_buttonState = Hovered;
    QGraphicsWidget::hoverEnterEvent(event);
    setSvgPrefix();
    update();
}

void StatesIcon::hoverLeaveEvent(QGraphicsSceneHoverEvent * event)
{
    m_buttonState = Normal;
    QGraphicsWidget::hoverLeaveEvent(event);
    setSvgPrefix();
    update();
}

void StatesIcon::mousePressEvent(QGraphicsSceneMouseEvent * event)
{
    kDebug() << "Icon is pressed";

    m_buttonState = Pressed;
    setSvgPrefix();
}

void StatesIcon::mouseReleaseEvent(QGraphicsSceneMouseEvent * event)
{
    kDebug() << "Icon is released";

    if (m_buttonState == Pressed) {
        kDebug() << "Sending click event";
        emit clicked();
    }
    m_buttonState = Hovered;
    setSvgPrefix();
    update();
}
