/***************************************************************************
 *   Copyright (C) 2008 by Sebastian Kuegler <sebas@kde.org>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include <limits.h>

// Qt
#include <QPainter>
#include <QColor>
#include <QGraphicsGridLayout>
#include <QGraphicsProxyWidget>
#include <QGraphicsSceneDragDropEvent>
#include <QLabel>
//#include <QRectF>

// KDE
#include <KIcon>
#include <KColorScheme>
#include <KMimeType>
#include <KDebug>
#include <KConfigDialog>

// Plasma
#include <plasma/corona.h>
#include <plasma/theme.h>

// Own
#include "grid.h"


Grid::Grid(QObject *parent, const QVariantList &args)
    : Plasma::Containment(parent, args)
{
    setHasConfigurationInterface(true);
    //setAspectRatioMode(Plasma::Square);
    setSizePolicy(QSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored));
    //setMinimumSize(20, 80);
    setMaximumSize(INT_MAX, INT_MAX);
    resize(120, 120);
    //setAcceptsHoverEvents(true);
    setAcceptDrops(true);
    int margin = 4;
    m_layout = new QGraphicsGridLayout(this);
    m_layout->setSpacing(margin);
    m_layout->setContentsMargins(margin, margin, margin, margin);
    setSizePolicy(QSizePolicy::Minimum,QSizePolicy::Minimum);
    m_appletNames << "clock" << "launcher" << "battery" << "" << "icon";
    kDebug() << "Loading applets: " << m_appletNames;
}

Grid::~Grid()
{
    for (int n = 0; n < m_applets.size(); ++n) {
        if (m_applets[n]) {
            delete m_applets[n];
        }
    }
    clearApplets();
}

void Grid::init()
{
    kDebug() << "------------------------------------------------";
    kDebug() << "Init Grid::CustomContainment -- Yeah!";
    kDebug() << "------------------------------------------------";

    KConfigGroup cg = config();
    m_rows = cg.readEntry("rows", 2);
    m_columns = cg.readEntry("columns", 2);
    int fields = m_rows * m_columns;
    int applets = m_appletNames.size();
    if (m_rows * m_columns > applets) {
        for (int i = applets; i < fields; ++i) {
            kDebug() << "Appended CELL Applet";
            m_appletNames << "cell";
        }
    }
    kDebug() << "Rows::Columns" << m_rows << "::" << m_columns;

    //setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    setContainmentType(Containment::CustomContainment);
    setBackgroundHints(NoBackground);
    initLayout();
    initApplets();
    updateGeometry();
    connect(this, SIGNAL(appletAdded(Plasma::Applet*,QPointF)),
            this, SLOT(layoutApplet(Plasma::Applet*,QPointF)));
    //resize(contentSizeHint());

}

void Grid::initLayout ()
{
    int a = m_appletNames.size();
    int rows = qMax(1, m_rows);
    int columns = qMax(1, m_columns);

    if (formFactor() == Plasma::Vertical) {
        // number of columns has precedence, expands vertically
        rows = a / m_columns;
        if (a % m_columns > 0) {
            rows++;
        }
    } else {
        // number of rows has precedence, expands horizontally
        columns = a / m_rows;
        if (a % columns > 0) {
            columns++;
        }
    }
    int r = 0;
    int c = 0;
    for (int n = 0; n < a; ++n) {
        m_fields[n] = QPair<int, int>(r,c);
        kDebug() << "m_fields[" << n << "]" << m_fields[n];
        if (formFactor() == Plasma::Vertical) {
            c++;
            if (c == columns) {
                r++; // row full, next one
                c = 0;
            }
        } else {
            r++;
            if (r == rows) {
                c++; // row full, next one
                r = 0;
            }
        }
    }

    // Are there any cells left that need filling?
    kDebug() << m_fields.size() << " = " << rows << " * " << columns;
    if (m_fields.size() < (rows * columns)) {
        for (int c = m_fields.size(); c < columns; ++c) {
            kDebug() << "Adding one cell applet";
            m_fields[m_fields.size()+1] = QPair<int, int>(r,c);
        }
    }
    kDebug() << "----initLayout() done -----------";
    kDebug() << "Rows: " << m_rows << rows << "Columns:" << m_columns << columns;
    kDebug() << "Fields: " << m_fields.size();

    m_columns = m_layout->columnCount();
    m_rows = m_layout->rowCount();
    setLayout(m_layout);
    kDebug() << "----END::initLayout() done -----------";
    kDebug() << "Rows: " << m_rows << rows << "Columns:" << m_columns << columns;
    kDebug() << "Fields: " << m_fields.size();
}

void Grid::initApplets ()
{
    // internal counters
    for (int n = 0; n < m_fields.size(); ++n) {
        QString appletName = "cell";
        if (m_appletNames.size() > n) {
            appletName = m_appletNames.at(n);
        }
        loadApplet(appletName, n);
    }
    kDebug() << "Applets: " << m_applets.size() << "AppletNames:" << m_appletNames.size();

    m_columns = m_layout->columnCount();
    m_rows = m_layout->rowCount();
}

void Grid::loadApplet (QString appletName, int n)
{
    if (appletName.isEmpty()) {
        appletName = "cell";
    }
    Plasma::Applet *applet = Plasma::Applet::load(appletName);
    if (applet) {
        applet->setBackgroundHints(NoBackground);
        loadApplet(applet, n);
    } else {
        kDebug() << "Could not load applet" << appletName;
    }
}

void Grid::loadApplet (Plasma::Applet* applet, int n)
{
    // TODO: cast icon to Plasma::Applet ... ?
    /*
    QGraphicsWidget* applet;
    if (name == "clock" || name == "battery") {
        applet = dynamic_cast<QGraphicsWidget*>(Applet::load(name));
        //Plasma::Applet* applet = Applet::load(name);
    } else {
        applet = dynamic_cast<QGraphicsWidget*>(new Plasma::Icon(KIcon(name), QString(), this));
        //Plasma::Applet* applet = Applet::load(name);
    }
    */
    if (applet) {
        // Mapping applet number to m_fields in grid
        kDebug() << applet->name() << " Applet loaded";
        applet->setParent(this);
        applet->setParentItem(this);
        applet->setFlag(QGraphicsItem::ItemIsMovable, false);
        m_layout->addItem(applet, m_fields[n].first, m_fields[n].second);
        connect(applet, SIGNAL(configNeedsSaving()), this, SIGNAL(configNeedsSaving()));
        connect(applet, SIGNAL(releaseVisualFocus()), this, SIGNAL(releaseVisualFocus()));
        connect(applet, SIGNAL(destroyed(QObject*)), this, SLOT(appletDestroyed(QObject*)));
        applet->init();

        if (m_applets.size() > n) {
            m_applets[n]->destroy();
        }
        m_applets[n] = applet;
    } else {
        kDebug() << "Could not load applet. :/";
    }
}

void Grid::layoutApplet(Plasma::Applet* applet, const QPointF &pos) {
    kDebug() << "------------------------------------------------";
    kDebug() << "                                    Applet Added";
    kDebug() << "------------------------------------------------";
    kDebug() << "Applet Added:" << applet->name() << pos;
    loadApplet(applet, getRectNumber(pos));

    // TODO: Put new applet somewhere in the grid, relayout ...

    m_columns = m_layout->columnCount();
    m_rows = m_layout->rowCount();

}

// from containment.cpp
void Grid::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    kDebug() << "------------------------------------------------";
    kDebug() << "                                    DropEvent";
    kDebug() << "------------------------------------------------";
    kDebug() << "Text :" << event->mimeData()->text();
    QString mimetype(static_cast<Plasma::Corona*>(scene())->appletMimeType());
    kDebug() << "Mimetype :" << mimetype << "::" << event->mimeData();
    kDebug() << "POS :" <<  mapFromScene(event->scenePos());
    if (event->mimeData()->hasUrls()) {
        foreach (const QUrl& url, event->mimeData()->urls()) {
            kDebug() << "URL: " << url;
        }
    }

    //QString mimetype(static_cast<Plasma::Corona*>(scene())->appletMimeType());
    if (event->mimeData()->hasFormat(mimetype) && scene()) {
        QString plasmoidName;
        plasmoidName = event->mimeData()->data(mimetype);
        QRectF geom(mapFromScene(event->scenePos()), QSize(0, 0));
        addApplet(plasmoidName, QVariantList(), geom);
        kDebug() << "|||==>" << plasmoidName << "dropped at " << geom << getRectNumber(geom.topLeft());
        loadApplet(plasmoidName, getRectNumber(geom.topLeft()));
        event->acceptProposedAction();
    } else if (KUrl::List::canDecode(event->mimeData())) {
        //TODO: collect the mimetypes of available script engines and offer
        //      to create widgets out of the matching URLs, if any
        KUrl::List urls = KUrl::List::fromMimeData(event->mimeData());
        foreach (const KUrl& url, urls) {
            KMimeType::Ptr mime = KMimeType::findByUrl(url);
            QString mimeName = mime->name();
            QRectF geom(event->scenePos(), QSize(0, 0));
            QVariantList args;
            args << url.url();
            //             kDebug() << mimeName;
            KPluginInfo::List appletList = Applet::listAppletInfoForMimetype(mimeName);

            if (appletList.isEmpty()) {
                // no special applet associated with this mimetype, let's
                //addApplet("icon", args, geom);
                kDebug() << "NO special applets";
            } else {
                //TODO: should we show a dialog here to choose which plasmoid load if
                //!appletList.isEmpty()
                //addApplet(appletList.first().pluginName(), args, geom);
                kDebug() << "Special applets" << appletList.first().pluginName();
            }
        }
        event->acceptProposedAction();
    }
}

int Grid::getRectNumber(const QPointF pos)
{
    for (int n = 0; n < m_rects.size(); ++n) {
        if (m_rects[n].contains(pos)) {
            kDebug() << "FOUND:" << pos << "is in rect" << n;
            return n;
        }
    }
    return 0;
}

QSizeF Grid::contentSizeHint() const
{
    QSizeF sizeHint = contentsRect().size();
    int r = qMax(1, m_rows);
    int c = qMax(1, m_columns);

    qreal aspect = (qreal)(c) / (qreal)(r);  // FIXME: account for margins
    switch (formFactor()) {
        case Plasma::Vertical:
            sizeHint.setHeight(sizeHint.width() * aspect); // FIXME: account for margins
            break;

        default:
            sizeHint.setWidth(sizeHint.height() * aspect); // FIXME: account for margins
            break;
    }
    kDebug() << "Aspect" << aspect;
    kDebug() << "SizeHint" << sizeHint;
    kDebug() << "Rows, Columns" << m_columns << m_rows << c << r;

    return sizeHint;
}

void Grid::constraintsEvent(Plasma::Constraints constraints)
{
    if (constraints & Plasma::SizeConstraint) {
        //initLayout();
        setMaximumSize(INT_MAX, INT_MAX); // reset maximum size
        int r = qMax(1, m_rows);
        int c = qMax(1, m_columns);

        qreal aspect = (qreal)(c) / (qreal)(r);  // FIXME: account for margins

        //qreal aspect = (qreal)(m_columns) / (qreal)(m_rows);
        int height = contentsRect().height();
        int width = contentsRect().width();
        kDebug() << "Aspect" << aspect;
        if (formFactor() == Plasma::Horizontal) {
            // We have a fixed height, set some sensible width
            width = (int)(height * aspect);
            setMaximumWidth(width);
        } else if (formFactor() == Plasma::Vertical) {
            // We have a fixed width, set some sensible height
            height = (int)(width / aspect);
            setMaximumHeight(height);
        } else {
            width = contentsRect().height() * aspect;
            //setMaximumWidth(width);
        }
        for (int n = 0; n < (m_fields.size()); ++n) {
            if (n < m_applets.size()) {
                QGraphicsLayoutItem* applet =m_layout->itemAt(m_fields[n].first, m_fields[n].second);
                if (applet) {
                    kDebug() << "### Applet " << n << ":: " << m_layout->itemAt(m_fields[n].first, m_fields[n].second)->geometry();
                    m_rects[n] = applet->geometry();
                }
            }
        }
    }
}

void Grid::createConfigurationInterface(KConfigDialog *parent)
{
    QWidget *widget = new QWidget();
    ui.setupUi(widget);
    parent->setButtons( KDialog::Ok | KDialog::Cancel | KDialog::Apply );
    parent->addPage(widget, parent->windowTitle(), icon());
    connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));

    if (formFactor() == Plasma::Vertical) {
        ui.note->setText(i18n("Note: The number of columns takes precedence."));
    } else {
        ui.note->setText(i18n("Note: The number of rows takes precedence."));
    }
    ui.rowSpin->setValue(m_rows);
    ui.colSpin->setValue(m_columns);
}

void Grid::configAccepted()
{
    KConfigGroup cg = config();
    kDebug() << " = = = ==  = = = = === =";
    kDebug() << "R, C:" << m_rows << m_columns;

    if (m_rows != ui.rowSpin->value()) {
        m_rows = ui.rowSpin->value();
        cg.writeEntry("rows", m_rows);
    }

    if (m_columns != ui.colSpin->value()) {
        m_columns = ui.colSpin->value();
        cg.writeEntry("columns", m_columns);
    }

    kDebug() << "R, C:" << m_rows << m_columns;
    emit configNeedsSaving();
}


void Grid::paintInterface(QPainter *p, const QStyleOptionGraphicsItem *option, const QRect &contentsRect)
{
    //kDebug() << " ---- " << geometry() << m_layout->geometry() << contentsRect;
    Q_UNUSED( option );

    p->setRenderHint(QPainter::SmoothPixmapTransform);
    p->setRenderHint(QPainter::Antialiasing);

    KColorScheme plasmaColorTheme = KColorScheme(QPalette::Active, KColorScheme::View, Plasma::Theme::defaultTheme()->colorScheme());
    QColor defaultTextColor = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);

    //color1.setAlphaF(0.5);
    defaultTextColor.setAlphaF(0.7);
    p->setPen(defaultTextColor);

    //for (int n = 0; n < m_rects.size(); ++n) {
    //    p->drawRoundedRect(m_rects[n], .2, .2, Qt::RelativeSize);
    //}
}
#include "grid.moc"
