/*
    This file is part of the Program Menu plasmoid, part of the KDE project.

    Copyright (c) 2002 Siegfried Nijssen <snijssen@liacs.nl>
    Copyright (c) 2003 Lubos Lunak <l.lunak@suse.cz>
    Copyright 2008 Friedrich W. H. Kossebau <kossebau@kde.org>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
    AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
    AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef MENUPLASMOID_H
#define MENUPLASMOID_H

// plasmoid
// #include "ui_tasksConfig.h"
// Plasma
#include <plasma/applet.h>
// Qt
#include <QtCore/QList>

class MenuBarEmbedContainer;

class KSelectionWatcher;
class KSelectionOwner;

class QGraphicsProxyWidget;
class QLabel;


/**
 * @short A central menu bar
 *
 * @description All status change, such as when an window is activated,
 * a new window popped up, etc, is received via @ref KWin::WindowInfo and @ref
 * NETWinInfo. Status changes for X selections are done via KSelectionWatcher.
 *
 * How it works in broad terms:  The MenuPlasmoid gets notified as soon a window
 * changes (a new pops up etc.) and accordingly updates the list @ref mMenuContainers,
 * which contains all known menus. When a new window gains focus, it looks up the
 * correct MenuBarEmbedContainer in @ref mMenuContainers, and then switches to that one.
 *
 * The documentation is a bit rusty -- read it with a critical eye.
 *
 * @author Siegfried Nijssen <snijssen@liacs.nl>
 * @author Lubos Lunak <l.lunak@suse.cz>
 */
class MenuPlasmoid : public Plasma::Applet
{
    Q_OBJECT

  public:
    explicit MenuPlasmoid( QObject *parent, const QVariantList &args = QVariantList() );
    virtual ~MenuPlasmoid();

  public: // ? API
//     virtual int widthForHeight( int height ) const;
//     virtual int heightForWidth( int width ) const;

  protected: // ? API
//     virtual void paletteChange(const QPalette& );
//     virtual void positionChange( Position p );

  protected:
    /**
     * Returns true if the selection is Not owned. That is, the menu applet
     * isn't "running" and is listening for the selection to be released.
     */
    bool isDisabled() const;

    /**
     * Does some sanity checks, and then sets mActiveMenuContainer to @param menuContainer.
     */
    void activateMenu( MenuBarEmbedContainer *menuContainer );

    void updateTopEdgeOffset();

  protected Q_SLOTS: // ? API
    /**
     * Reads whether a central menu bar should be used or not, basically.
     */
    virtual void readSettings();

  protected Q_SLOTS:
    /**
     * Called each time a window is added. When the selection is
     * initially claimed, it is called for every window. It does the big
     * work, and does the embedding with MenuBarEmbedContainer.
     */
    void registerPotentialMenu( WId windowId );

    /**
     * Finds @p w's menubar in @see mMenuContainers, and then activates it.
     *
     * @param w the activated window.
     */
    void onActiveWindowChanged( WId windowId );

    /**
     * Looks up @param menuContainer in @ref mMenuContainers, and removes it.
     */
    void onMenuLost( MenuBarEmbedContainer *menuContainer );

    void onSizeChanged( MenuBarEmbedContainer *menuContainer );

    /**
     * Called when the selection(selection_atom) is lost. Deletes the
     * embedded mMenuContainers, and starts listening for the selection again.
     *
     */
    void onLostSelection();

    void claimSelection();

  protected:
    QLabel *mLabel;
    QGraphicsProxyWidget *mMenuContainerProxyWidget;
    /**
     * List of all known mMenuContainers.
     */
    QList< MenuBarEmbedContainer* > mMenuContainers;

    /**
     * A pointer to the current active menu, which is member
     * of @ref mMenuContainers.
     */
    MenuBarEmbedContainer *mActiveMenuContainer;

    /**
     * A pointer to the desktop menu, which is member
     * of @ref mMenuContainers.
     */
    MenuBarEmbedContainer *mDesktopMenuContainer;

    KSelectionOwner *mSelectionOwner;

    /**
     * Only the messenger. Dispatches signals to  claimSelection().
     */
    KSelectionWatcher *mSelectionWatcher;

    /**
     * Whether the Desktop menu should be used, when a window
     * with no menu is used.
     */
    bool mUseDesktopMenu;

    /**
     * The distance to the top of the screen.
     */
    int mTopEdgeOffset;
};

#endif
