/***********************************************************************************
* Mini Player: Advanced media player for Plasma.
* Copyright (C) 2008 - 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "MiniPlayerPlayListImporter.h"
#include "MiniPlayerApplet.h"

#include <QDir>
#include <QFile>
#include <QRegExp>
#include <QFileInfo>
#include <QXmlStreamReader>
#include <QXmlStreamAttributes>

#include <KLocale>
#include <KMimeType>
#include <KMessageBox>

MiniPlayerPlayListImporter::MiniPlayerPlayListImporter(QString playList, KUrl::List urls, bool play, int index, MiniPlayerApplet *parent) : QObject(parent)
{
    m_play = play;

    m_index = index;

    m_playList = playList;

    m_imports = 0;

    connect(this, SIGNAL(processedTracks(QString, KUrl::List, QHash<KUrl, QString>, bool, int)), parent, SLOT(importPlayList(QString, KUrl::List, QHash<KUrl, QString>, bool, int)));

    addUrls(urls);
}

void MiniPlayerPlayListImporter::addUrls(KUrl::List items, int level)
{
    ++m_imports;

    foreach (const KUrl &url, items)
    {
        PlayListType type = None;

        if (url.isLocalFile())
        {
            KMimeType::Ptr mimeType = KMimeType::findByUrl(url);

            if (mimeType->is("inode/directory"))
            {
                readDirectory(url, level);

                continue;
            }
            else if (mimeType->is("audio/x-scpls"))
            {
                type = PLS;
            }
            else if (mimeType->is("audio/x-mpegurl"))
            {
                type = M3U;
            }
            else if (mimeType->is("application/xspf+xml"))
            {
                type = XSPF;
            }
            else if (mimeType->is("audio/x-ms-asx"))
            {
                type = ASX;
            }

            if (type != None)
            {
                importPlayList(url, type);

                continue;
            }

            QString mimeTypeName = mimeType.data()->name();

            if (mimeTypeName.indexOf("video/") == -1 && mimeTypeName.indexOf("audio/") == -1 && mimeTypeName != "application/ogg")
            {
                continue;
            }
        }
        else
        {
            if (url.pathOrUrl().endsWith(QString(".pls"), Qt::CaseInsensitive))
            {
                type = PLS;
            }
            else if (url.pathOrUrl().endsWith(QString(".m3u"), Qt::CaseInsensitive))
            {
                type = M3U;
            }
            else if (url.pathOrUrl().endsWith(QString(".xpsf"), Qt::CaseInsensitive))
            {
                type = XSPF;
            }
            else if (url.pathOrUrl().endsWith(QString(".asx"), Qt::CaseInsensitive))
            {
                type = ASX;
            }

            if (type != None)
            {
                KIO::Job *job = KIO::get(url, KIO::NoReload, KIO::HideProgressInfo);

                connect(job, SIGNAL(data(KIO::Job*, const QByteArray&)), this, SLOT(importPlayListData(KIO::Job*, const QByteArray&)));
                connect(job, SIGNAL(result(KJob*)), this, SLOT(importPlayListResult(KJob*)));

                job->start();

                m_remotePlayListsData[job] = QByteArray();
                m_remotePlayListsType[job] = type;

                continue;
            }
        }

        m_tracks.append(url);
    }

    --m_imports;

    if (!m_imports)
    {
        emit processedTracks(m_playList, m_tracks, m_titles, m_play, m_index);

        deleteLater();
    }
}

void MiniPlayerPlayListImporter::importPlayList(KUrl url, PlayListType type)
{
    KUrl::List items;

    QFileInfo *currentLocation = new QFileInfo(url.pathOrUrl());

    QDir::setCurrent(currentLocation->absolutePath());

    delete currentLocation;

    QFile data(url.pathOrUrl());

    if (data.open(QFile::ReadOnly))
    {
        if (type == XSPF)
        {
            QByteArray byteArray = data.readAll();

            readXspf(byteArray);
        }
        else if (type == ASX)
        {
            QByteArray byteArray = data.readAll();

            readAsx(byteArray);
        }
        else
        {
            QTextStream stream(&data);

            if (type == PLS)
            {
                readPls(stream);
            }
            else
            {
                readM3u(stream);
            }
        }

        data.close();
    }
    else
    {
        KMessageBox::error(NULL, i18n("Cannot open file for reading."));
    }

    addUrls(items);
}

void MiniPlayerPlayListImporter::importData(KIO::Job* job, const QByteArray &data)
{
    m_remotePlayListsData[job].append(data);
}

void MiniPlayerPlayListImporter::importResult(KJob* job)
{
    if (m_remotePlayListsType[job] == XSPF)
    {
        readXspf(m_remotePlayListsData[job]);
    }
    else if (m_remotePlayListsType[job] == ASX)
    {
        readAsx(m_remotePlayListsData[job]);
    }
    else
    {
        QTextStream stream(m_remotePlayListsData[job]);

        if (m_remotePlayListsType[job] == PLS)
        {
            readPls(stream);
        }
        else
        {
            readM3u(stream);
        }
    }

    m_remotePlayListsData.remove(job);
    m_remotePlayListsType.remove(job);
}

void MiniPlayerPlayListImporter::readM3u(QTextStream &stream)
{
    KUrl::List items;
    QRegExp m3uTitle("^#EXTINF:\\d*,");
    QString line;
    QString title;
    bool addUrl;

    do
    {
        line = stream.readLine();

        if (line.isEmpty())
        {
            continue;
        }

        addUrl = false;

        if (!line.startsWith('#'))
        {
            addUrl = true;
        }

        if (line.contains(m3uTitle))
        {
            title = line.remove(m3uTitle);
        }

        if (addUrl)
        {
            KUrl url(line);

            if (!url.isValid())
            {
                continue;
            }

            if (!title.isEmpty())
            {
                m_titles[url] = title;

                title = "";
            }

            if (!url.isLocalFile())
            {
                items.append(url);

                continue;
            }

            QFileInfo *location = new QFileInfo(line);
            location->makeAbsolute();

            if (location->exists())
            {
                items.append(KUrl(location->filePath()));
            }

            delete location;
        }
    }
    while (!line.isNull());

    addUrls(items);
}

void MiniPlayerPlayListImporter::readPls(QTextStream &stream)
{
    KUrl::List items;
    QRegExp plsUrl("^File\\d+=");
    QRegExp plsTitle("^Title\\d+=");
    QString line;
    QString title;
    bool addUrl;

    do
    {
        line = stream.readLine();

        if (line.isEmpty())
        {
            continue;
        }

        addUrl = false;

        if (line.contains(plsUrl))
        {
            addUrl = true;

            line = line.remove(plsUrl);
        }
        else if (line.contains(plsTitle))
        {
            m_titles[items.last()] = line.remove(plsTitle);
        }

        if (addUrl)
        {
            KUrl url(line);

            if (!url.isValid())
            {
                continue;
            }

            if (!url.isLocalFile())
            {
                items.append(url);

                continue;
            }

            QFileInfo *location = new QFileInfo(line);
            location->makeAbsolute();

            if (location->exists())
            {
                items.append(KUrl(location->filePath()));
            }

            delete location;
        }
    }
    while (!line.isNull());

    addUrls(items);
}

void MiniPlayerPlayListImporter::readXspf(QByteArray &data)
{
    KUrl::List items;
    QXmlStreamReader reader(data);
    QString title;

    while (!reader.atEnd())
    {
        reader.readNext();

        if (reader.name().toString() == "track")
        {
            title = "";
        }

        if (reader.name().toString() == "title")
        {
            title = reader.text().toString();
        }

        if (reader.name().toString() == "location")
        {
            KUrl url(reader.text().toString());

            if (!url.isValid())
            {
                continue;
            }

            if (!url.isLocalFile())
            {
                items.append(url);

                continue;
            }

            QFileInfo *location = new QFileInfo(url.toLocalFile());
            location->makeAbsolute();

            if (location->exists())
            {
                items.append(KUrl(location->filePath()));
            }

            delete location;
        }
    }

    addUrls(items);
}

void MiniPlayerPlayListImporter::readAsx(QByteArray &data)
{
    KUrl::List items;
    QXmlStreamReader reader(data);
    QString title;

    while (!reader.atEnd())
    {
        reader.readNext();

        if (reader.name().toString() == "entry")
        {
            title = "";
        }

        if (reader.name().toString() == "title")
        {
            title = reader.text().toString();
        }

        if (reader.name().toString() == "ref")
        {
            QXmlStreamAttributes attributes = reader.attributes();

            KUrl url(attributes.value("href").toString());

            if (!url.isValid())
            {
                continue;
            }

            if (!url.isLocalFile())
            {
                items.append(url);

                continue;
            }

            QFileInfo *location = new QFileInfo(url.toLocalFile());
            location->makeAbsolute();

            if (location->exists())
            {
                items.append(KUrl(location->filePath()));
            }

            delete location;
        }
    }

    addUrls(items);
}

void MiniPlayerPlayListImporter::readDirectory(KUrl url, int level)
{
    if (level > 9)
    {
        return;
    }

    QStringList entries = QDir(url.toLocalFile()).entryList(QDir::Readable | QDir::Dirs | QDir::Files | QDir::NoDotAndDotDot);
    KUrl::List urls;

    for (int i = 0; i < entries.count(); ++i)
    {
        KUrl newUrl = url;
        newUrl.addPath(entries.at(i));

        urls.append(newUrl);
    }

    addUrls(urls, ++level);
}

#include "MiniPlayerPlayListImporter.moc"
