/*
*   Copyright 2007 by Aaron Seigo <aseigo@kde.org>
*   Copyright 2008 by Alexis Ménard <darktears31@gmail.com>
*
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU Library General Public License version 2,
*   or (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details
*
*   You should have received a copy of the GNU Library General Public
*   License along with this program; if not, write to the
*   Free Software Foundation, Inc.,
*   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "desktop.h"

#include <limits>

#include <QAction>
#include <QGraphicsProxyWidget>
#include <QFile>
#include <QFileInfo>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QPainter>
#include <QTimeLine>

#include <KAuthorized>
#include <KComboBox>
#include <KDebug>
#include <KFileDialog>
#include <KImageFilePreview>
#include <KRun>
#include <KStandardDirs>
#include <KWindowSystem>

#include <Plasma/Corona>
#include <Plasma/Animator>
#include <Plasma/Theme>
#include "kworkspace/kworkspace.h"
#include "knewstuff2/engine.h"

#include "krunner_interface.h"
#include "screensaver_interface.h"

#ifdef Q_OS_WIN
#define _WIN32_WINNT 0x0500 // require NT 5.0 (win 2k pro)
#include <windows.h>
#endif // Q_OS_WIN

using namespace Plasma;

ClutteredDesktop::ClutteredDesktop(QObject *parent, const QVariantList &args)
    : Containment(parent, args),
      m_lockDesktopAction(0),
      m_appletBrowserAction(0),
      m_addPanelAction(0),
      m_runCommandAction(0),
      m_lockScreenAction(0),
      m_logoutAction(0),
      dropping(false)
{
    qRegisterMetaType<QImage>("QImage");
    qRegisterMetaType<QPersistentModelIndex>("QPersistentModelIndex");

    m_layout = new DesktopLayout;
    m_layout->setAlignment(Qt::AlignTop|Qt::AlignLeft);
    m_layout->setPlacementSpacing(20);
    m_layout->setScreenSpacing(5);
    m_layout->setShiftingSpacing(0);
    m_layout->setTemporaryPlacement(true);
    m_layout->setVisibilityTolerance(0.5);

    resize(800, 600);

    setHasConfigurationInterface(true);
    //kDebug() << "!!! loading desktop";
}

void ClutteredDesktop::constraintsEvent(Plasma::Constraints constraints)
{
    if (constraints & Plasma::ImmutableConstraint && m_appletBrowserAction) {
        // we need to update the menu items that have already been created
        bool locked = immutability() != Mutable;
        m_addPanelAction->setVisible(!locked);
    }

    if (constraints & Plasma::SizeConstraint) {
        refreshWorkingArea();
    }

    //FIXME: are refres on working are change and refresh on resize mutually exclusive?
    if (constraints & Plasma::StartupCompletedConstraint) {
        connect(corona(), SIGNAL(availableScreenRegionChanged()),
                this, SLOT(refreshWorkingArea()));
        refreshWorkingArea();

        connect(this, SIGNAL(appletAdded(Plasma::Applet *, const QPointF &)),
                this, SLOT(onAppletAdded(Plasma::Applet *, const QPointF &)));
        connect(this, SIGNAL(appletRemoved(Plasma::Applet *)),
                this, SLOT(onAppletRemoved(Plasma::Applet *)));

        foreach (Applet *applet, applets()) {
            m_layout->addItem(applet, true, false);
            connect(applet, SIGNAL(appletTransformedByUser()), this, SLOT(onAppletTransformedByUser()));
            connect(applet, SIGNAL(appletTransformedItself()), this, SLOT(onAppletTransformedItself()));
        }

        m_layout->adjustPhysicalPositions();
    }
}

void ClutteredDesktop::addPanel()
{
    if (corona()) {
        // make a panel at the top
        Containment* panel = corona()->addContainment("panel");
        panel->showConfigurationInterface();

        panel->setScreen(screen());

        QList<Plasma::Location> freeEdges = corona()->freeEdges(screen());
        kDebug() << freeEdges;
        Plasma::Location destination;
        if (freeEdges.contains(Plasma::TopEdge)) {
            destination = Plasma::TopEdge;
        } else if (freeEdges.contains(Plasma::BottomEdge)) {
            destination = Plasma::BottomEdge;
        } else if (freeEdges.contains(Plasma::LeftEdge)) {
            destination = Plasma::LeftEdge;
        } else if (freeEdges.contains(Plasma::RightEdge)) {
            destination = Plasma::RightEdge;
        } else destination = Plasma::TopEdge;

        panel->setLocation(destination);

        // trigger an instant layout so we immediately have a proper geometry
        // rather than waiting around for the event loop
        panel->updateConstraints(Plasma::StartupCompletedConstraint);
        panel->flushPendingConstraintsEvents();

        if (destination == Plasma::LeftEdge || destination == Plasma::RightEdge) {
            panel->setMinimumSize(10, 35);
            panel->setMaximumSize(35, corona()->screenGeometry(screen()).height());
            panel->resize(QSize(35, corona()->screenGeometry(screen()).height()));
        }

    }
}

void ClutteredDesktop::runCommand()
{
    if (!KAuthorized::authorizeKAction("run_command")) {
        return;
    }

    QString interface("org.kde.krunner");
    org::kde::krunner::App krunner(interface, "/App", QDBusConnection::sessionBus());
    krunner.display();
}

void ClutteredDesktop::lockScreen()
{
    if (!KAuthorized::authorizeKAction("lock_screen")) {
        return;
    }

#ifndef Q_OS_WIN
    QString interface("org.freedesktop.ScreenSaver");
    org::freedesktop::ScreenSaver screensaver(interface, "/ScreenSaver",
                                              QDBusConnection::sessionBus());
    if (screensaver.isValid()) {
        screensaver.Lock();
    }
#else
    LockWorkStation();
#endif // !Q_OS_WIN
}

QList<QAction*> ClutteredDesktop::contextualActions()
{
    //TODO: should we offer "Switch User" here?

    if (!m_appletBrowserAction) {
        m_appletBrowserAction = action("add widgets");

        m_addPanelAction = new QAction(i18n("Add Panel"), this);
        connect(m_addPanelAction, SIGNAL(triggered(bool)), this, SLOT(addPanel()));
        m_addPanelAction->setIcon(KIcon("list-add"));

        m_runCommandAction = new QAction(i18n("Run Command..."), this);
        connect(m_runCommandAction, SIGNAL(triggered(bool)), this, SLOT(runCommand()));
        m_runCommandAction->setIcon(KIcon("system-run"));

        m_setupDesktopAction = action("activity settings");
        m_lockDesktopAction = action("lock widgets");

        m_lockScreenAction = new QAction(i18n("Lock Screen"), this);
        m_lockScreenAction->setIcon(KIcon("system-lock-screen"));
        connect(m_lockScreenAction, SIGNAL(triggered(bool)), this, SLOT(lockScreen()));

        m_logoutAction = new QAction(i18n("Leave..."), this);
        m_logoutAction->setIcon(KIcon("system-shutdown"));
        connect(m_logoutAction, SIGNAL(triggered(bool)), this, SLOT(logout()));
        constraintsEvent(Plasma::ImmutableConstraint);

        m_separator = new QAction(this);
        m_separator->setSeparator(true);
    }

    QList<QAction*> actions;

    if (KAuthorized::authorizeKAction("run_command")) {
        actions.append(m_runCommandAction);
    }

    actions.append(m_appletBrowserAction);
    actions.append(m_addPanelAction);
    actions.append(m_setupDesktopAction);
    if (screen() == -1) {
        actions.append(action("remove"));
    }

    actions.append(m_lockDesktopAction);

    actions.append(m_separator);

    if (KAuthorized::authorizeKAction("lock_screen")) {
        actions.append(m_lockScreenAction);
    }

    if (KAuthorized::authorizeKAction("logout")) {
        actions.append(m_logoutAction);
    }

    return actions;
}

void ClutteredDesktop::logout()
{
    if (!KAuthorized::authorizeKAction("logout")) {
        return;
    }
#ifndef Q_WS_WIN
    KWorkSpace::requestShutDown(KWorkSpace::ShutdownConfirmDefault,
                                KWorkSpace::ShutdownTypeDefault,
                                KWorkSpace::ShutdownModeDefault);
#endif
}

void ClutteredDesktop::onAppletAdded(Plasma::Applet *applet, const QPointF &pos)
{
    if (dropping || pos != QPointF(-1,-1) || applet->geometry().topLeft() != QPointF(0,0)) {
        // add item to the layout using the current position
        m_layout->addItem(applet, true, false);
    } else {
        // auto-position
        m_layout->addItem(applet, true, true);
    }

    m_layout->adjustPhysicalPositions();

    connect(applet, SIGNAL(appletTransformedByUser()), this, SLOT(onAppletTransformedByUser()));
    connect(applet, SIGNAL(appletTransformedItself()), this, SLOT(onAppletTransformedItself()));
}

void ClutteredDesktop::onAppletRemoved(Plasma::Applet *applet)
{
    for (int i=0; i < m_layout->count(); i++) {
        if (applet == m_layout->itemAt(i)) {
            m_layout->removeAt(i);
            m_layout->adjustPhysicalPositions();
            return;
        }
    }
}

void ClutteredDesktop::onAppletTransformedByUser()
{
    m_layout->itemTransformed((Applet *)sender(), DesktopLayout::ItemTransformUser);
    m_layout->adjustPhysicalPositions();
}

void ClutteredDesktop::onAppletTransformedItself()
{
    m_layout->itemTransformed((Applet *)sender(), DesktopLayout::ItemTransformSelf);
    m_layout->adjustPhysicalPositions();
}

void ClutteredDesktop::refreshWorkingArea()
{
    Corona *c = corona();
    if (!c) {
        //kDebug() << "no corona?!";
        return;
    }

    QRectF workingGeom;
    if (screen() != -1) {
        // we are associated with a screen, make sure not to overlap panels
        workingGeom = c->availableScreenRegion(screen()).boundingRect();
        //kDebug() << "got" << workingGeom;
        // From screen coordinates to containment coordinates
        workingGeom.translate(-c->screenGeometry(screen()).topLeft());
    } else {
        workingGeom = geometry();
        workingGeom = mapFromScene(workingGeom).boundingRect();
        //kDebug() << "defaults due to no screen; got:" << workingGeom;
    }

    if (workingGeom != QRectF()) {
        //kDebug() << "!!!!!!!!!!!!! workingGeom is" << workingGeom;
        m_layout->setWorkingArea(workingGeom);
        m_layout->adjustPhysicalPositions();
    }
}

void ClutteredDesktop::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    dropping = true;
    Containment::dropEvent(event);
    dropping = false;
}

// Cluttered Desktop code starts here
// Edit cluttereddesktop.cpp (not desktop.cpp) and run update.sh

#include <typeinfo>
#include <QTextDocument>
#include <QAbstractTextDocumentLayout>
#include <QGraphicsLinearLayout>
#include <KConfigDialog>
#include <Plasma/Wallpaper>
#include <Plasma/WebView>
#include "ui_config.h"

class ClutteredDesktopPrivate
{
public:
    QGraphicsItem* toolbox;
    QTextDocument doc;
    QGraphicsLinearLayout* layout;
    Plasma::WebView* html;
    bool cashew;
    bool debug;
    Ui::Config ui;
};

void ClutteredDesktop::destroyed()
{
    delete d;
}

void ClutteredDesktop::init()
{
    Containment::init();
    d = new ClutteredDesktopPrivate;

    KConfigGroup cfg = config();
    d->cashew = cfg.readEntry("cashew", true);
    d->debug = cfg.readEntry("debug", false);

    findToolbox();
    if (d->toolbox && !d->cashew) {
        d->toolbox->hide();
    }
    connect(this, SIGNAL(appletAdded(Plasma::Applet*, const QPointF)),
            this, SLOT(appletAddedSlot(Plasma::Applet*, const QPointF)));
    connect(this, SIGNAL(appletRemoved(Plasma::Applet*)), this, SLOT(updateText()));
    connect(this, SIGNAL(screenChanged(int, int, Plasma::Containment*)), this, SLOT(updateText()));
    connect(this, SIGNAL(geometryChanged()), this, SLOT(updateText()));
    /*
    m_settingsAction = new QAction(i18n("Cluttered Desktop Extra Settings"), this);
    m_settingsAction->setIcon(KIcon("configure"));
    connect(m_settingsAction, SIGNAL(triggered(bool)),
            this, SLOT(showAppletConfigurationInterface()));
    */
    QTimer::singleShot(250, this, SLOT(updateText()));
}

void ClutteredDesktop::appletAddedSlot(Plasma::Applet* applet, const QPointF pos)
{
    Q_UNUSED(pos)
    connect(applet, SIGNAL(geometryChanged()), this, SLOT(updateText()));
    updateText();
}

void ClutteredDesktop::updateText()
{
    if (d->debug) {
        QString toolbox = (d->toolbox) ? (d->cashew) ? i18n("Visible") : i18n("Hidden") : i18n("Not Found");
        QString row("<tr><td>%1:</td><td>%2</td></tr>");
        QString text =
                "<h2>Cluttered Desktop</h2><table>" +
                QString(row).arg(i18n("ID")).arg(id()) +
                QString(row).arg(i18n("Screen")).arg(screen()) +
                QString(row).arg(i18n("Geometry")).arg(QString("%1,%2 %3,%4")
                        .arg(geometry().left()).arg(geometry().top())
                        .arg(geometry().width()).arg(geometry().height())) +
                QString(row).arg(i18n("Wallpaper"))
                        .arg((wallpaper()) ? wallpaper()->pluginName() : QString()) +
                QString(row).arg(i18n("Wallpaper mode")).arg((wallpaper()) ?
                        wallpaper()->renderingMode().name() : QString()) +
                QString(row).arg(i18n("Toolbox")).arg(toolbox) +
                "</table><h3>Applets</h3><table>";
        foreach (Applet* applet, applets()) {
            text += QString("<tr><td>%1:</td><td>%2</td><td>%3</td></tr>")
                    .arg(applet->name()).arg(applet->id())
                    .arg(QString("%2,%3 %4,%5")
                        .arg(applet->geometry().left()).arg(applet->geometry().top())
                        .arg(applet->geometry().width()).arg(applet->geometry().height()));
        }
        text += "</table>";
        d->doc.setHtml(text);
        update();
    }
}

void ClutteredDesktop::findToolbox()
{
    d->toolbox = 0;
    foreach (QGraphicsItem* item, childItems()) {
        QString s = typeid(*item).name();
        if (s.contains("DesktopToolBox")) {
            d->toolbox = item;
            break;
        }
    }
}

void ClutteredDesktop::paintInterface(QPainter* painter,
                                     const QStyleOptionGraphicsItem* option,
                                     const QRect& contentsRect)
{
    Q_UNUSED(option)
    Q_UNUSED(contentsRect)
    if (d->debug) {
        QAbstractTextDocumentLayout* layout = d->doc.documentLayout();
        QAbstractTextDocumentLayout::PaintContext context;

        painter->save();
        painter->translate(52, 52);
        context.palette.setColor(QPalette::Text, QColor(Qt::black));
        layout->draw(painter, context);
        painter->translate(-2, -2);
        context.palette.setColor(QPalette::Text, QColor(Qt::white));
        layout->draw(painter, context);
        painter->restore();
    }
}

void ClutteredDesktop::createConfigurationInterface(KConfigDialog* parent)
{
    QWidget* widget = new QWidget;
    d->ui.setupUi(widget);

    d->ui.cashewCheckBox->setCheckState((Qt::CheckState)((int)d->cashew * 2));
    d->ui.debugCheckBox->setCheckState((Qt::CheckState)((int)d->debug * 2));

    parent->setButtons(KDialog::Ok | KDialog::Cancel | KDialog::Apply);
    parent->addPage(widget, i18n("Desktop"), "user-desktop");
    connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
}

void ClutteredDesktop::configAccepted()
{
    KConfigGroup cfg = config();

    d->cashew = (d->ui.cashewCheckBox->checkState() == Qt::Checked);
    d->debug = (d->ui.debugCheckBox->checkState() == Qt::Checked);
    cfg.writeEntry("cashew", d->cashew);
    cfg.writeEntry("debug", d->debug);

    emit configNeedsSaving();

    if (d->toolbox) {
        if (d->cashew) {
            d->toolbox->show();
        } else {
            d->toolbox->hide();
        }
    }
    if (d->debug) {
        updateText();
    } else {
        update();
    }
}

void ClutteredDesktop::showAppletConfigurationInterface()
{
    Applet::showConfigurationInterface();
}

// Cluttered Desktop code ends here

K_EXPORT_PLASMA_APPLET(cluttereddesktop, ClutteredDesktop)

#include "desktop.moc"
