
#include <QtSql/QSqlDatabase>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlError>
#include <QtCore/QDir>
#include <QStack>

#include <kstandarddirs.h>
#include <KBookmarkManager>

#include <browserbookmarks.h>

BrowserBookmarksEngine::BrowserBookmarksEngine(QObject* parent, const QVariantList& args)
{
    Q_UNUSED(parent)
    Q_UNUSED(args)
    m_firefoxDbCopyPath = KStandardDirs::locateLocal("cache", "") + "bookmarkrunnerfirefoxdbfile.sqlite";
    setMinimumPollingInterval(30000); // half a minute
    m_bookmarkManager = KBookmarkManager::userBookmarksManager();
}


bool BrowserBookmarksEngine::sourceRequestEvent(const QString& name)
{
    if (name.compare("firefox") == 0) {
        if (QSqlDatabase::isDriverAvailable("QSQLITE")) {
            KConfig firefoxProfile(QDir::homePath() + "/.mozilla/firefox/profiles.ini",
                               KConfig::SimpleConfig);
            QStringList profilesList = firefoxProfile.groupList();
            profilesList = profilesList.filter(QRegExp("^Profile\\d+$"));
            int size = profilesList.size();

            QString profilePath;
            if (size == 1) {
                // There is only 1 profile so we select it
                KConfigGroup fGrp = firefoxProfile.group(profilesList.first());
                profilePath = fGrp.readEntry("Path", "");
            } else {
                // There are multiple profiles, find the default one
                foreach(const QString & profileName, profilesList) {
                    KConfigGroup fGrp = firefoxProfile.group(profileName);
                    if (fGrp.readEntry<int>("Default", 0)) {
                        profilePath = fGrp.readEntry("Path", "");
                        break;
                    }
                }
            }

            if (profilePath.isEmpty()) {
                kDebug() << "No default firefox profile found";
                m_firefoxDb = QSqlDatabase();
            } else {
                profilePath.prepend(QString("%1/.mozilla/firefox/").arg(QDir::homePath()));
                m_firefoxDbFilePath = profilePath + "/places.sqlite";
                m_firefoxDb = QSqlDatabase::addDatabase("QSQLITE");
                m_firefoxDb.setHostName("localhost");
            }
        } else {
            kDebug() << "SQLITE driver isn't available";
            m_firefoxDb = QSqlDatabase();
        }
    }

    return updateSourceEvent(name);
}


bool BrowserBookmarksEngine::updateSourceEvent(const QString& source)
{
    if (source.compare("firefox") == 0) {
        if (m_firefoxDb.isValid()) {
            kDebug() << "Cache file was removed: " << QFile(m_firefoxDbCopyPath).remove();
            kDebug() << "Database was copyed: " <<
            QFile(m_firefoxDbFilePath).copy(m_firefoxDbCopyPath);
            m_firefoxDb.setDatabaseName(m_firefoxDbCopyPath);
        }
        if (m_firefoxDb.open()) {
            kDebug() << "Database was opened";
            QSqlQuery query("SELECT moz_bookmarks.title, moz_places.url" \
                        " FROM moz_bookmarks, moz_places WHERE " \
                        "moz_bookmarks.type = 1 AND moz_bookmarks.fk = moz_places.id");

            QList<QVariant> titles;
            QList<QVariant> urls;
            for (int i = 0; query.next(); i++) {
                const QString title = query.value(0).toString();
                const QUrl url = query.value(1).toString();
                if (url.isValid()) { //Try to ignore Firefox's own bookmarks
                    titles << title;
                    urls << url;
                }
            }
            setData(source, QString("titles"), titles);
            setData(source, QString("urls"), urls);
        } else {
            kDebug() << "Was not able to open the database";
            kDebug() << m_firefoxDb.lastError().text();
        }
        m_firefoxDb.close();
    } else if (source.compare("konqueror") == 0) {
        KBookmarkGroup bookmarkGroup = m_bookmarkManager->root();
        QStack<KBookmarkGroup> groups;
        KBookmark bookmark = bookmarkGroup.first();
        QList<QVariant> titles;
        QList<QVariant> urls;
        while (!bookmark.isNull()) {
            if (bookmark.isGroup()) { // descend
                //kDebug () << "descending into" << bookmark.text();
                groups.push(bookmarkGroup);
                bookmarkGroup = bookmark.toGroup();
                bookmark = bookmarkGroup.first();

                while (bookmark.isNull() && !groups.isEmpty()) {
                    bookmark = bookmarkGroup;
                    bookmarkGroup = groups.pop();
                    bookmark = bookmarkGroup.next(bookmark);
                }

                continue;
            }
            const QString text = bookmark.text();
            const QString url = bookmark.url().prettyUrl();
            titles << text;
            urls << url;

            bookmark = bookmarkGroup.next(bookmark);
            while (bookmark.isNull() && !groups.isEmpty()) {
                bookmark = bookmarkGroup;
                bookmarkGroup = groups.pop();
                //kDebug() << "ascending from" << bookmark.text() << "to" << bookmarkGroup.text();
                bookmark = bookmarkGroup.next(bookmark);
            }
        }
        setData(source, QString("titles"), titles);
        setData(source, QString("urls"), urls);
    }
    return true;
}

K_EXPORT_PLASMA_DATAENGINE(browserbookmarks, BrowserBookmarksEngine)
 
#include "browserbookmarks.moc"