/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   Copyright (C) 2008 by Dennis Nienhüser <earthwings@gentoo.org>        *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 ***************************************************************************/

#include <QtCore/QSharedData>
#include <QtCore/QString>
#include <QtCore/QPointF>
#include <QtCore/QStringList>
#include <QtCore/QDateTime>

#include <ksharedptr.h>

#ifndef CONTACT_H
#define CONTACT_H

/**
 * Represents a contact in terms of the contacts engine. This class is only used
 * for data exchange between the contacts engine and contact sources
 * (kopete, nepomuk, decibel/akonadi, ...)
 */
class Contact: public QSharedData
{
public:
	typedef KSharedPtr<Contact> Ptr;

	/**
	 * Constructor
	 * @param id A unique, persistent identifier for this contact.
	 */
	Contact(const QString &id);

	/** Destructor */
	virtual ~Contact();

	/**
	 * @return A unique, persistent identifier for this contact
	 */
	QString id() const;

	/**
	 * @return Best known way to display a contact's name to the user. May be
	 * the full name, a nick name or anything else, but should not be empty.
	 */
	virtual QString displayName() = 0;

	/**
	 * @return A path to an image containing a photo/avatar image. An empty
	 * string indicates that no photo is set or known, which is what the default
	 * implementation always returns.
	 */
	virtual QString photo();

	/**
	 * @return Instant messaging online status. This may be an arbitrary string
	 * An empty string indicates that the online status cannot be determined.
	 * Implementing this is optional, the default implementation returns an empty
	 * string.
	 */
	virtual QString onlineStatus();

	/**
	 * @return Instant messaging online status message. This may be an arbitrary
	 * string. An empty string indicates that the online status cannot be
	 * determined. Implementing this is optional, the default implementation
	 * returns an empty string.
	 */
	virtual QString onlineStatusMessage();

	/**
	 * @return Email address. This may be an arbitrary string. An empty string
	 * indicates that the email address is not known. Implementing this is
	 * optional, the default implementation returns an empty string.
	 */
	virtual QString email();

    /**
     * @return Instant messaging id (icq, jabber, ...). This may be an arbitrary 
     * string. An empty string indicates that the im id is not known. Implementing 
     * this is optional, the default implementation returns an empty string.
     */
    virtual QString instantMessagingId();

    /**
	 * @return True iff the geographical location is known
	 * @see #location
	 */
	virtual bool hasLocation();

	/**
	 * @return Geographical contact location (longitude and latitude)
	 * @see #hasLocation
	 */
	virtual QPointF location();
	
	/**
	 * @return A list of unread instant messages
	 */
	virtual QStringList unreadInstantMessages();
	
	/**
	 * @return Contact's birthday
	 */    
	virtual QDateTime birthday();

private:
	/** Contact identifier */
	const QString m_id;
};

#endif // CONTACT_H
