/*
 * Copyright (C) 2010 Eduardo Robles Elvira <edulix@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB. If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */
#include "globalshortcuts_runner.h"
#include "kglobalaccel_interface.h"
#include "kglobalaccel_component_interface.h"

#include <QtDBus>

#include <KDebug>
#include <KActionCollection>
#include <KGlobalAccel>
#include <KShortcut>
#include <KIcon>
#include <QtGui/QKeySequence>

struct GlobalShortcutsRunner::Private
{
    Private() {}

    void strComparator(const QString& str, const QString& term, qreal& relevance);

    QList<KGlobalShortcutInfo> shortcuts;
};

GlobalShortcutsRunner::GlobalShortcutsRunner(QObject *parent, const QVariantList& args)
    : Plasma::AbstractRunner(parent, args), d(new Private)
{
    Q_UNUSED(args);
    setSpeed(NormalSpeed);
    setPriority(NormalPriority);
    setHasRunOptions(false);
//     setIgnoredTypes(Plasma::RunnerContext::FileSystem | Plasma::RunnerContext::Help);
    setObjectName("globalshortcuts_runner");
}

GlobalShortcutsRunner::~GlobalShortcutsRunner()
{
}

void GlobalShortcutsRunner::Private::strComparator(const QString& str, const QString& term,
    qreal& globalRelevance)
{
    qreal localRelevance = 0;
    if (str.toLower().contains(term)) {
      localRelevance = 0.7;
    }
    if (str.toLower() == term) {
      localRelevance = 1;
    }
    globalRelevance = qMax(localRelevance, globalRelevance);
}

void GlobalShortcutsRunner::match(Plasma::RunnerContext &context)
{
    const QString term = context.query();
    kDebug() << "term" << term;
    if (term.length() < 3) {
        return;
    }
    QList<Plasma::QueryMatch> matches;

    // Match names
    Q_FOREACH (const KGlobalShortcutInfo& shortcut, d->shortcuts) {
        kDebug() << "comparing with shortcut = " << shortcut.friendlyName();
        qreal relevance = 0;
        d->strComparator(shortcut.friendlyName(), term, relevance);
        d->strComparator(shortcut.uniqueName(), term, relevance);
        Q_FOREACH (const QKeySequence& keySequence, shortcut.keys()) {
            d->strComparator(keySequence.toString(), term, relevance);
        }

        if (!relevance) {
            continue;
        }

        kDebug() << "matches!";

        Plasma::QueryMatch match(this);
        match.setText(shortcut.friendlyName());
        match.setId(shortcut.uniqueName());
        
        QString keys;
        Q_FOREACH (const QKeySequence& keySequence, shortcut.keys()) {
          keys.append(keySequence.toString());
        }
        match.setSubtext(keys + "(" + shortcut.componentFriendlyName() + ")");

        match.setData(qVariantFromValue(shortcut));
        match.setIcon(KIcon("preferences-desktop-keyboard-shortcut"));
        match.setRelevance(relevance);
        matches.append(match);
    }
    context.addMatches(term, matches);
}

void GlobalShortcutsRunner::run(const Plasma::RunnerContext &context, const Plasma::QueryMatch &match)
{
    Q_UNUSED(context)

    KGlobalShortcutInfo shortcut = match.data().value<KGlobalShortcutInfo>();
    
    org::kde::kglobalaccel::Component component("org.kde.kglobalaccel",
        "/component/" + shortcut.componentUniqueName(), QDBusConnection::sessionBus());

    if (!component.isValid()) {
        kDebug() << "Component " << shortcut.componentUniqueName() << "not valid! Skipping!";
        return;
    }
    
    component.invokeShortcut(shortcut.uniqueName());
}

void GlobalShortcutsRunner::init()
{
    kDebug();
    // Connect to kglobalaccel. If that fails there is no need to continue.
    qRegisterMetaType<QList<int> >();
    qDBusRegisterMetaType<QList<int> >();
    qDBusRegisterMetaType<QList<KGlobalShortcutInfo> >();
    qDBusRegisterMetaType<KGlobalShortcutInfo>();

    org::kde::KGlobalAccel kglobalaccel("org.kde.kglobalaccel", "/kglobalaccel",
        QDBusConnection::sessionBus());

    if (!kglobalaccel.isValid()) {
        return;
    }

    QDBusReply< QList<QDBusObjectPath> > componentsRc = kglobalaccel.allComponents();
    if (!componentsRc.isValid()) {
        return;
    }
    QList<QDBusObjectPath> components = componentsRc;

    Q_FOREACH (const QDBusObjectPath &componentPath, components) {
        // Get the component
        org::kde::kglobalaccel::Component component("org.kde.kglobalaccel", componentPath.path(),
            QDBusConnection::sessionBus());

        if (!component.isValid()) {
            kDebug() << "Component " << componentPath.path() << "not valid! Skipping!";
            continue;
        }

        // Get the shortcut contexts.
        QDBusReply<QStringList> shortcutContextsRc = component.getShortcutContexts();
        if (!shortcutContextsRc.isValid()) {
            kDebug() << "Failed to get contexts for component "
                    << componentPath.path() <<"! Skipping!";
            kDebug() << shortcutContextsRc.error();
            continue;
        }
        QStringList shortcutContexts = shortcutContextsRc;

        // We add the shortcuts for all shortcut contexts. This way the user keeps full control
        // of it's shortcuts.
        Q_FOREACH (const QString &shortcutContext, shortcutContexts) {
            QDBusReply< QList<KGlobalShortcutInfo> > shortcutsRc =
                component.allShortcutInfos(shortcutContext);
            if (!shortcutsRc.isValid()) {
                kDebug() << "allShortcutInfos() failed for " << componentPath.path() << shortcutContext;
                continue;
            }
            QList< KGlobalShortcutInfo > shortcuts = shortcutsRc.value();
            d->shortcuts.append(shortcuts);
        }
    }
}


#include "globalshortcuts_runner.moc"
