/*
 * Copyright 2007 Frerich Raabe <raabe@kde.org>
 * Copyright 2007-2008 Aaron Seigo <aseigo@kde.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "fullview.h"

#include <iostream>

#include <QPixmapCache>

#include <KApplication>
#include <KAboutData>
#include <KAction>
#include <KCmdLineArgs>
#include <KLocale>
#include <KStandardAction>

// for --list
#include <Plasma/Applet>
#include <Plasma/Theme>
#include <Plasma/AccessManager>
#include <Plasma/AuthorizationManager>
#include <Plasma/Wallpaper>

using namespace Plasma;

static const char description[] = I18N_NOOP("Run Plasma widgets in their own window");

class RemotePlasmoidWatcher : public QObject
{
Q_OBJECT

public:
    RemotePlasmoidWatcher(AccessManager *manager)
        : QObject(manager)
    {
        kDebug();
        connect(manager, SIGNAL(remotePlasmoidAdded(Plasma::PackageMetadata)),
                this, SLOT(slotServiceAdded(Plasma::PackageMetadata)));
        connect(manager, SIGNAL(remotePlasmoidRemoved(Plasma::PackageMetadata)),
                this, SLOT(slotServiceRemoved(Plasma::PackageMetadata)));
    }

    ~RemotePlasmoidWatcher() {}

public Q_SLOTS:
    void slotServiceAdded(Plasma::PackageMetadata metadata) {
        std::cout << "New service published:" << std::endl;
        std::cout << metadata.remoteLocation().prettyUrl().toLocal8Bit().data() << std::endl;
        std::cout << metadata.name().toLocal8Bit().data() << " - "
                  << metadata.description().toLocal8Bit().data() << std::endl;
    }

    void slotServiceRemoved(Plasma::PackageMetadata metadata) {
        std::cout << "Service removed:" << std::endl;
        std::cout << metadata.remoteLocation().prettyUrl().toLocal8Bit().data() << std::endl;
    }
};

void listPlugins(const KPluginInfo::List & plugins)
{
    int maxLen = 0;
    QMap<QString, QString> applets;
    foreach (const KPluginInfo &info, plugins) {
        if (info.property("NoDisplay").toBool()) {
            continue;
        }

        int len = info.pluginName().length();
        if (len > maxLen) {
            maxLen = len;
        }

        QString name = info.pluginName();
        QString comment = info.comment();

        if (comment.isEmpty()) {
            comment = i18n("No description available");
        }

        applets.insert(name, comment);
    }

    QMap<QString, QString>::const_iterator it;
    for (it = applets.constBegin(); it != applets.constEnd(); it++) {
        QString applet("%1 - %2");

        applet = applet.arg(it.key().leftJustified(maxLen, ' ')).arg(it.value());
        std::cout << applet.toLocal8Bit().data() << std::endl;
    }
}

int main(int argc, char **argv)
{
    KAboutData aboutData("plasma-server", 0, ki18n("Plasma Widget Viewer"),
                         "1.0", ki18n(description), KAboutData::License_BSD,
                         ki18n("2007-2008, Frerich Raabe"));
    aboutData.setProgramIconName("plasma");
    aboutData.addAuthor(ki18n("Frerich Raabe"),
                         ki18n("Original author"),
                        "raabe@kde.org");

    KCmdLineArgs::init(argc, argv, &aboutData);

    KCmdLineOptions options;
    options.add("list", ki18n("Displays a list of known applets"));
    options.add("+applet", ki18n("Name of applet to view; may refer to the plugin name or be a path "
                                "(absolute or relative) to a package. If not provided, then an "
                                "attempt is made to load a package from the current directory."));
    options.add("+[args]", ki18n("Optional arguments of the applet to share"));
    options.add("list-remote", ki18n("List zeroconf announced remote widgets"));
    KCmdLineArgs::addCmdLineOptions(options);

    KApplication app;

    KCmdLineArgs *args = KCmdLineArgs::parsedArgs() ;

    if (args->isSet("list")) {
        listPlugins(Plasma::Applet::listAppletInfo());
        return 0;
    }

    QString pluginName;
    if (args->count() > 0) {
        pluginName = args->arg(0);
    }

    QVariantList appletArgs;
    for (int i = 1; i < args->count(); ++i) {
        appletArgs << args->arg(i);
    }
    kDebug() << "setting auth policy";
    Plasma::AuthorizationManager::self()->setAuthorizationPolicy(
            Plasma::AuthorizationManager::PinPairing);

//     FullView view(formfactor, location);

    if (args->isSet("list-remote")) {
        kDebug() << "list remote...";
        /**
        QList<KUrl> list = AccessManager::self()->remotePlasmoids();
        foreach (const KUrl &url, list) {
            std::cout << url.prettyUrl().toLocal8Bit().data() << std::endl;
        }
        */
        new RemotePlasmoidWatcher(AccessManager::self());
    } else {
        kDebug() << "just load applet";
//         view.addApplet(pluginName, containment, wallpaper, appletArgs);
//         view.show();

//         QAction *action = KStandardAction::quit(&app, SLOT(quit()), &view);
//         view.addAction(action);
    }

    args->clear();

    return app.exec();
}

#include "main.moc"

