#include <QPainter>

#include "edge.h"
#include "nodeitem.h"

#include <QDebug>
#include <math.h>

Edge::Edge()
{
    setAcceptedMouseButtons(0);
    source = 0;
    dest = 0;
    setZValue(5);
    //adjust();
}

Edge::~Edge()
{
    if (source)
        source->removeEdge(this);
    if (dest)
        dest->removeEdge(this);
    hide();
}

void Edge::adjust()
{
    if (!source || !dest)
        return;
    QRectF rect = source->propertyRect(sourceProp);
    sourceRect = source->mapToScene(rect).boundingRect();

    rect = dest->propertyRect(destProp);
    destRect = dest->mapToScene(rect).boundingRect();
    adjust(start(), end());
}


NodeItem *Edge::destNode() const
{
    return dest;
}

NodeItem *Edge::sourceNode() const
{
    return source;
}

static inline QPointF midPoint(const QLineF &l)
{
    return (l.p2()+l.p1())/2;
}

void Edge::adjust(const QPointF &source, const QPointF &dest)
{
    prepareGeometryChange();
    sourcePoint = source;
    destPoint = dest;

    QLineF line(sourcePoint, destPoint);
    QPainterPath path;
    path.moveTo(sourcePoint);
    qreal halfMid = (midPoint(line).x() - sourcePoint.x())/2;
    path.cubicTo(QPointF(destPoint.x() - halfMid, sourcePoint.y()),
                 QPointF(sourcePoint.x() + halfMid, destPoint.y()),
                 destPoint);
    QPainterPathStroker stroker;
    stroker.setWidth(5);
    stroker.setCapStyle(Qt::FlatCap);
    m_path = stroker.createStroke(path);
    //qDebug()<<"source = "<<source<<sourcePoint
    //        <<", dest = "<<dest<<destPoint;
}

QRectF Edge::boundingRect() const
{
    QRectF base = m_path.boundingRect();
    if (!sourceRect.isNull())
        base |= sourceRect;
    if (!destRect.isNull())
        base |= destRect;
    return base;
}

void Edge::paint(QPainter *painter, const QStyleOptionGraphicsItem *, QWidget *)
{
    painter->setRenderHint(QPainter::Antialiasing);
    painter->setPen(QPen(Qt::black, 0, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin));
    painter->setBrush(Qt::yellow);

    painter->drawPath(m_path);

    painter->drawEllipse(sourceRect);
    if (!destPoint.isNull())
        painter->drawEllipse(destRect);
}

void Edge::setSource(NodeItem *node,
                     const QString &property)
{
    if (source)
        source->removeEdge(this);
    source = node;
    sourceProp = property;
    QRectF rect = source->propertyRect(sourceProp);
    sourceRect = source->mapToScene(rect).boundingRect();
    if (source)
        source->addEdge(this);
    adjust();
}

void Edge::setDest(NodeItem *node,
                   const QString &property)
{
    if (dest)
        dest->removeEdge(this);
    dest = node;
    destProp = property;
    QRectF rect = dest->propertyRect(destProp);
    destRect = dest->mapToScene(rect).boundingRect();
    if (dest)
        dest->addEdge(this);
    adjust();
}

QPointF Edge::start() const
{
    if (source) {
        QRectF rect = source->propertyRect(sourceProp);
        return source->mapToScene(rect.center());
    }
    return QPointF();
}

QPointF Edge::end() const
{
    if (dest) {
        QRectF rect = dest->propertyRect(destProp);
        return dest->mapToScene(rect.center());
    }
    return QPointF();
}

bool Edge::updateLink()
{
    if (!source || !dest)
        return false;
    Quasar::Node *srcNode = source->node();
    Quasar::Node *dstNode = dest->node();
    QString sProp = sourceProp;
    QString dProp = destProp;

    if (!srcNode || !dstNode ||
        sourceProp.isEmpty() || destProp.isEmpty())
        return false;

    if (srcNode->property(sourceProp).kind == Quasar::Property::InputProperty &&
        dstNode->property(destProp).kind == Quasar::Property::InputProperty) {
        qWarning("Can't connect two input properties!");
        return false;
    }
    if (srcNode->property(sourceProp).kind == Quasar::Property::OutputProperty &&
        dstNode->property(destProp).kind == Quasar::Property::OutputProperty) {
        qWarning("Can't connect two output properties!");
        return false;
    }

    if (srcNode->property(sourceProp).kind == Quasar::Property::InputProperty &&
        dstNode->property(destProp).kind == Quasar::Property::OutputProperty) {
        qSwap(srcNode, dstNode);
        qSwap(sProp, dProp);
    }

    return Quasar::link(srcNode, sProp,
                        dstNode, dProp);
}
