#include "window.h"

#include "scene.h"
#include "nodeitem.h"
#include "shadowitem.h"
#include "edge.h"
#include "compositionview.h"

#include "composition.h"
#include "node.h"

#include <QFileDialog>
#include <QGraphicsView>
#include <QGraphicsPixmapItem>
#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QPainter>
#include <QPixmap>
#include <QBrush>
#include <QMessageBox>
#include <QDebug>


Window::Window()
{
    init();
}

Window::~Window()
{
    delete m_view;
}

void Window::init()
{
    ui.setupUi(this);
    m_composition = new Quasar::Composition(this);
    m_view = NULL;

    QStringList groupNames = Quasar::Composition::groups();

    QMap<QString, QTreeWidgetItem*> groups;
    foreach(QString flt, groupNames) {
        QStringList nodesInGroup = Quasar::Composition::nodesInGroup(flt);
        foreach(QString nodeName, nodesInGroup) {
            QTreeWidgetItem *group = 0;
            if (groups.contains(flt))
                group = groups[flt];
            else {
                group = new QTreeWidgetItem(ui.filtersTree,
                                            QStringList(flt));
                groups.insert(flt, group);

                QColor color(193, 193, 193, 127);

                group->setData(1, Qt::DisplayRole, color);
				group->setFlags(Qt::ItemIsEnabled);
                group->setExpanded(true);
            }
            new QTreeWidgetItem(group, QStringList(nodeName));
        }
    }

    ui.graphicsView->setScene(new Scene(m_composition, QRectF(0, 0, 600, 400), this));

    QPixmap tile(100, 100);
    tile.fill(Qt::white);
    QPainter pt(&tile);
    QColor color(220, 220, 220);
    pt.setPen(color);
    pt.drawRect(0, 0, 100, 100);
    pt.drawLine(0, 50, 100, 50);
    pt.drawLine(50, 0, 50, 100);
    pt.end();
    QBrush b(tile);
    ui.graphicsView->setBackgroundBrush(b);

    connect(ui.graphicsView->scene(), SIGNAL(selectionChanged()), SLOT(updatePropertyEditor()));
    connect(ui.viewButton, SIGNAL(clicked()),
            SLOT(viewImage()));
    connect(ui.filtersTree, SIGNAL(itemDoubleClicked(QTreeWidgetItem *, int)),
            SLOT(filterClicked(QTreeWidgetItem*, int)));

    QAction* action = new QAction(tr("&Quit"), this);
    action->setShortcut(Qt::CTRL+Qt::Key_Q);
    connect(action, SIGNAL(triggered(bool)), qApp, SLOT(quit()));
    addAction(action);
}

void Window::filterClicked(QTreeWidgetItem * item, int)
{
    QString name = item->text(0);
    QColor color;

    if (!item->parent())
        return;

    QString parentName = item->parent()->text(0);
    color = qvariant_cast<QColor>(item->parent()->data(1, Qt::DisplayRole));
    name = QString("/%1/%2").arg(parentName).arg(name);


    qDebug()<<"filterClicked name = "<<name;
    NodeItem *fitem = new NodeItem();
    connect(fitem, SIGNAL(destroyed()),
            this, SLOT(applyFiltersOnScene()));
    fitem->setNode(m_composition->createNode(name));
    fitem->setBackgroundColor(color);
    ui.graphicsView->scene()->addItem(fitem);
    ShadowItem *sitem = new ShadowItem();
    sitem->setShadowParent(fitem);
    fitem->setShadow(sitem);
    ui.graphicsView->scene()->addItem(sitem);

    sitem->show();
    fitem->show();

    applyFiltersOnScene();
}

void Window::viewImage()
{
    if (!m_view) {
        m_view = new CompositionView(m_composition);
        connect(ui.propertyEditor, SIGNAL(propertyChanged()), m_view, SLOT(update()));
    }
    m_view->show();
}

void Window::applyFiltersOnScene()
{
    if (m_view)
        m_view->update();
}

void Window::updatePropertyEditor()
{
    QList<QGraphicsItem*> items = ui.graphicsView->scene()->selectedItems();
    if (items.size() != 1) {
        ui.propertyEditor->setNode(NULL);
        return;
    }
    NodeItem* item = qgraphicsitem_cast<NodeItem*>(items.at(0));
    ui.propertyEditor->setNode(item ? item->node() : NULL);
}

#include "moc_window.cpp"
