#include "editors.h"

#include "node.h"

#include <QSlider>
#include <QDoubleSpinBox>
#include <QLineEdit>
#include <QVBoxLayout>

class DoubleEditor : public WidgetEditor
{
    Q_OBJECT
public:
    DoubleEditor(qreal min, qreal max, qreal def,
                 const QString &propName, QWidget *parent)
        : WidgetEditor(propName, parent),
          m_min(min),
          m_max(max)
    {
        QVBoxLayout *layout = new QVBoxLayout;
        QSlider *slider = new QSlider(Qt::Horizontal, this);
        connect(slider, SIGNAL(valueChanged(int)),
                this, SLOT(changed(int)));
        slider->setSizePolicy(QSizePolicy::Minimum,
                              QSizePolicy::Fixed);
        slider->setMinimum(0);
        slider->setMaximum(100);
        slider->setValue((def-min)/(max-min) * 100);
        layout->addWidget(slider);
        setLayout(layout);
    }
private slots:
    void changed(int percentInt)
    {
        qreal percent = percentInt / 100.;
        qreal range = m_max - m_min;
        qreal val = range * percent;
        qreal result = val + m_min;
        emitChanged(result);
    }

private:
    qreal m_min, m_max;
};


class TransformEditor : public WidgetEditor
{
    Q_OBJECT
public:
    TransformEditor(const QString &propName, QWidget *parent)
        : WidgetEditor(propName, parent)
    {
        QVBoxLayout *layout = new QVBoxLayout;
        QSlider *slider = new QSlider(Qt::Horizontal, this);
        connect(slider, SIGNAL(valueChanged(int)),
                this, SLOT(changed(int)));
        slider->setSizePolicy(QSizePolicy::Minimum,
                              QSizePolicy::Fixed);
        slider->setMinimum(0);
        slider->setMaximum(100);
        slider->setValue(50);
        layout->addWidget(slider);
        setLayout(layout);
    }
private slots:
    void changed(int percentInt)
    {
        qreal percent = percentInt / 100.;
        qreal range = 2;
        qreal val = range * percent;
        QTransform trans;
        trans.scale(val, val);
        emitChanged(trans);
    }

};


class LocationEditor : public WidgetEditor
{
    Q_OBJECT
public:
    LocationEditor(const QString &propName, QWidget *parent)
        : WidgetEditor(propName, parent)
    {
        QVBoxLayout *layout = new QVBoxLayout;
        QLineEdit *lineEdit = new QLineEdit(this);
        layout->addWidget(lineEdit);
        setLayout(layout);
    }
};


WidgetEditor * WidgetEditor::editorFor(const Quasar::Property &prop,
                                       QWidget *parent)
{
    switch(prop.type) {
    case QVariant::Double:
    case QVariant::Int:
        return new DoubleEditor(prop.min.toDouble(), prop.max.toDouble(),
                                prop.value.toDouble(),
                                prop.name, parent);
    case QVariant::Transform:
        return new TransformEditor(prop.name, parent);
    case QVariant::String:
    case QVariant::Url:
        return new LocationEditor(prop.name, parent);
    default:
        return 0;
    }
}

WidgetEditor::WidgetEditor(const QString &propName, QWidget *parent)
    : QWidget(parent),
      m_prop(propName)
{
    setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
}

void WidgetEditor::emitChanged(const QVariant &property)
{
    emit changed(m_prop, property);
}


#include "editors.moc"
#include "moc_editors.cpp"
