#include "node.h"

#include "node_p.h"

#include <QDebug>

namespace Quasar {

static inline bool operator<(const QVariant &varA,
                             const QVariant &varB)
{
    switch(varA.type()) {
    case QVariant::Int:
        return varA.value<int>() < varB.value<int>();
        break;
    case QVariant::UInt:
        return varA.value<uint>() < varB.value<uint>();
        break;
    case QVariant::LongLong:
        return varA.value<qlonglong>() < varB.value<qlonglong>();
        break;
    case QVariant::ULongLong:
        return varA.value<qulonglong>() < varB.value<qulonglong>();
        break;
    case QVariant::Double:
        return varA.value<double>() < varB.value<double>();
        break;
    case QVariant::Char:
        return varA.value<char>() < varB.value<char>();
        break;
    default:
        return false;
    }
}

static inline bool operator==(const Node::Connection &one,
                              const Node::Connection &two)
{
    return (one.node == two.node &&
            one.key == two.key);
}

static inline bool valueWithinBounds(QVariant &var,
                                     const QVariant &min,
                                     const QVariant &max)
{
    if (min.isValid()) {
        if (var < min) {
            var = min;
            return false;
        }
    }
    if (max.isValid()) {
        if (max < var) {
            var = max;
            return false;
        }
    }
    return true;
}

bool NodePrivate::setPropertyPrivate(Property &prop, const QVariant &value)
{
    Q_Q(Node);
    QVariant converter = value;
    if (converter.type() != prop.type) {
        if (!converter.canConvert(prop.type)) {
            emit q->warning(q->tr("Value '%1' doesn't accept '%2'")
                            .arg(prop.name)
                            .arg(converter.typeName()));
            return false;
        }
        converter.convert(prop.type);
    }
    valueWithinBounds(converter, prop.min, prop.max);
    if (prop.value == converter) {
        return false;
    }
    prop.value = converter;
    return true;
}

Node::Node(const QString &name)
    : d_ptr(new NodePrivate)
{
    Q_D(Node);
    d_ptr->q_ptr = this;
    d->name = name;
}

Node::Node(const QString &name, NodePrivate &dd)
    : d_ptr(&dd)
{
    Q_D(Node);
    d_ptr->q_ptr = this;
    d->name = name;
}

Node::~Node()
{
    Q_D(Node);
    unlink();
    delete d;
    d_ptr = 0;
}

QList<Property> Node::properties() const
{
    Q_D(const Node);
    return d->properties.values();
}

QList<QString> Node::inputKeys() const
{
    QList<QString> result;
    foreach (Property prop, properties())
    {
        if (prop.kind == Property::InputProperty)
            result << prop.name;
    }
    return result;
}

QList<QString> Node::outputKeys() const
{
    QList<QString> result;
    foreach (Property prop, properties())
    {
        if (prop.kind == Property::OutputProperty)
            result << prop.name;
    }
    return result;
}

bool Node::setPropertyValue(const QString &key, const QVariant &value)
{
    Q_D(Node);
    if (!d->properties.contains(key)) {
        emit warning(tr("The property '%1' doesn't exist.")
                     .arg(qPrintable(key)));
        return false;
    }

    Property &prop = d->properties[key];
    if (prop.kind == Property::OutputProperty) {
        emit warning(tr("Can't set an output property."));
        return false;
    }
    bool ret = d->setPropertyPrivate(prop, value);
    qDebug()<< d->name << "setting property" << key << value;
    if (ret) {
        propertyChanged(key, value);
        d->dirty = true;
    }
    return ret;
}

QVariant Node::propertyValue(const QString &key) const
{
    Q_D(const Node);
    return d->properties[key].value;
}

void Node::addProperty(const QString &key, const Property &prop)
{
    Q_D(Node);
    d->properties.insert(key, prop);
}

bool Node::hasProperty(const QString &key, Property::PropertyKind k) const
{
    Q_D(const Node);
    if (!d->properties.contains(key))
        return false;
    const Property &prop = d->properties[key];
    return prop.kind == k;
}

const Property Node::property(const QString &key) const
{
    Q_D(const Node);
    return d->properties[key];
}

QVariant::Type Node::propertyType(const QString &key) const
{
    Q_D(const Node);
    return d->properties[key].type;
}

bool link(Node *src, const QString &outKey,
          Node *dst, const QString &inKey)
{
    if (!src)
        return false;

    return src->link(outKey, dst, inKey);
}

bool unlink(Node *src, const QString &outKey,
            Node *dst, const QString &inKey)
{
    if (!src)
        return false;

    return src->unlink(outKey, dst, inKey);
}

bool Node::process()
{
    Q_D(Node);
    bool ret = true;
    //qDebug()<<"--- process "<<d->dirty;
    //if (d->dirty) {
        ret = execute();
        d->dirty = false;
        informReceivers();
        //}
    return ret;
}

bool Node::link(const QString &outKey,
                Node *dst, const QString &inKey)
{
    Q_D(Node);
    if (!dst) {
        qWarning("Node::link destination node is null");
        return false;
    }

    if (!hasProperty(outKey, Property::OutputProperty)) {
        qWarning("Node '%s' doesn't have output property named '%s'",
                 qPrintable(d->name), qPrintable(outKey));
        return false;
    }
    if (!dst->hasProperty(inKey, Property::InputProperty)) {
        qWarning("Node '%s' doesn't have input property named '%s'",
                 qPrintable(dst->name()), qPrintable(inKey));
        return false;
    }

    QVariant::Type outType = propertyType(outKey);
    QVariant::Type inType  = dst->propertyType(inKey);

//     qDebug()<<outVar.type()<< " === "<< inVar.type();
    if (!QVariant(outType).canConvert(inType)) {
        qWarning("Can't connect property %s to %s. The types are incompatible (%s and %s).",
                 qPrintable(outKey), qPrintable(inKey), QVariant::typeToName(outType),
                 QVariant::typeToName(inType));
        return false;
    }

    if (dst->d_ptr->senders.contains(inKey)) {
        Connection con = dst->d_ptr->senders[inKey];
        dst->d_ptr->senders.remove(inKey);
        QMultiHash<QString, Connection>::iterator itr;
        for (itr = con.node->d_ptr->receivers.begin();
             itr != con.node->d_ptr->receivers.end(); ++itr) {
            if (itr.value().node == dst) {
                con.node->d_ptr->receivers.erase(itr);
                break;
            }
        }
    }

    if (dst->d_ptr->senders.count() >= 1) {
        qWarning("FIXME: Currently only linear layout is supported!");
        return false;
    }

    dst->d_ptr->senders.insert(inKey,
                               Connection(outKey, this));
    d->receivers.insert(outKey,
                        Connection(inKey, dst));

    return true;
}

bool NodePrivate::removeFromReceivers(const QString &outKey,
                                      Node *receiver,
                                      const QString &inKey)
{
    QList<Node::Connection> r =  receivers.values(outKey);
    bool ret = false;
    QList<Node::Connection>::iterator itr;
    itr = r.begin();
    while (itr != r.end()) {
        if (receiver) {
            if ((receiver == (*itr).node) &&
                (inKey.isEmpty() || inKey == (*itr).key)) {
                (*itr).node->d_ptr->removeSender(q_ptr, outKey);
                itr = r.erase(itr);
                ret = true;
            } else {
                ++itr;
            }
        } else {
            (*itr).node->d_ptr->removeSender(q_ptr, outKey);
            itr = r.erase(itr);
            ret = true;
        }
    }
    receivers.remove(outKey);
    if (!r.isEmpty()) {
        foreach(Node::Connection con, r) {
            receivers.insert(outKey, con);
        }
    }
    return ret;
}

void NodePrivate::removeSender(Node *n, const QString &outKey)
{
    QHash<QString, Node::Connection>::iterator itr;
    itr = senders.begin();
    while (itr != senders.end()) {
        if (itr.value().node == n) {
            if (outKey.isEmpty() ||
                itr.value().key == outKey) {
                itr = senders.erase(itr);
                continue;
            }
        }
        ++itr;
    }
}

bool NodePrivate::removeFromSenders(const QString &outKey)
{
    if (senders.isEmpty())
        return false;

    if (!outKey.isEmpty()) {
        if (!senders.contains(outKey))
            return false;
        Node::Connection con = senders[outKey];
        senders.remove(outKey);
        con.node->d_ptr->removeReceiver(con.key, q_ptr, outKey);
    } else {
        QHash<QString, Node::Connection>::const_iterator itr;
        for (itr = senders.constBegin();
             itr != senders.constEnd(); ++itr) {
            Node::Connection con = itr.value();
            con.node->d_ptr->removeReceiver(con.key, q_ptr, outKey);
        }
    }
    return true;
}


void NodePrivate::removeReceiver(const QString &outKey,
                                 Node *n, const QString &inKey)
{
    receivers.remove(outKey, Node::Connection(inKey, n));
}

bool Node::unlink(const QString &outKey,
                  Node *other, const QString &inKey)
{
    Q_D(Node);

    bool ret = false;
    if (!outKey.isEmpty()) {
        if (!d->properties.contains(outKey)) {
            qWarning("Property '%s' doesn't exist",
                     qPrintable(outKey));
            return false;
        }
        Property prop = d->properties[outKey];
        if (prop.kind == Property::OutputProperty)
            ret = d->removeFromReceivers(outKey, other, inKey);
        else
            ret = d->removeFromSenders(outKey);
    } else {
        QHash<QString, Property>::const_iterator itr;
        for (itr = d->properties.constBegin();
             itr != d->properties.constEnd();
             ++itr) {
            bool r = false;
            if (itr.value().kind == Property::OutputProperty) {
                r = d->removeFromReceivers(itr.key(), other, inKey);
            } else {
                r = d->removeFromSenders(itr.key());
            }
            if (!ret)
                ret = r;
        }
    }

    return ret;
}

QList<Node::Connection> Node::outputNodes(const QString &key) const
{
    Q_D(const Node);
    QList<Connection> cons = d->receivers.values(key);
    return cons;
}


QMultiHash<QString, Node::Connection> Node::outputNodes() const
{
    Q_D(const Node);
    return d->receivers;
}


void Node::informReceivers()
{
    Q_D(const Node);
    QMultiHash<QString, Node::Connection>::const_iterator itr;
    for (itr = d->receivers.constBegin();
         itr != d->receivers.constEnd(); ++itr) {
        QString key = itr.key();
        const Node::Connection &con = itr.value();
        con.node->setPropertyValue(con.key,
                                   propertyValue(key));
    }
}

Node::ExecutionRestrain Node::executionRestrain() const
{
    return NoRestrain;
}

QString Node::name() const
{
    Q_D(const Node);
    return d->name;
}

bool Node::outputReady(const QString &key,
                       const QVariant &var)
{
    Q_D(Node);
    Property &prop = d->properties[key];

    if (prop.kind != Property::OutputProperty) {
        emit warning(tr("Property '%1' is not an output property.").arg(key));
        return false;
    }
    return d->setPropertyPrivate(prop, var);
}

QHash<QString, Node::Connection> Node::inputNodes() const
{
    Q_D(const Node);
    return d->senders;
}

bool Node::isDirty() const
{
    Q_D(const Node);
    return d->dirty;
}

bool Node::isFilterNode() const
{
    return false;
}

bool Node::isGraphicsNode() const
{
    return false;
}

bool Node::isRendererNode() const
{
    return false;
}

bool Node::isDependentOn(Node *node) const
{
    QHash<QString, Node::Connection>::const_iterator itr;

    for (itr = d_ptr->senders.constBegin(); itr != d_ptr->senders.constEnd(); ++itr) {
        Node::Connection c = itr.value();
        if (c.node == node)
            return true;
        else if (c.node->isDependentOn(node))
            return true;
    }
    return false;
}

bool Node::isIndependentOf(Node *node) const
{
    return !isDependentOn(node);
}

QList<Node*> Node::children() const
{
    QMultiHash<QString, Node::Connection>::const_iterator itr;
    QList<Node*> nodes;
    for (itr = d_ptr->receivers.constBegin(); itr != d_ptr->receivers.constEnd();
         ++itr) {
        nodes.append(itr.value().node);
    }
    return nodes;
}

QList<Node*> Node::parents() const
{
    QHash<QString, Node::Connection>::const_iterator itr;
    QList<Node*> nodes;
    for (itr = d_ptr->senders.constBegin(); itr != d_ptr->senders.constEnd();
         ++itr) {
        nodes.append(itr.value().node);
    }
    return nodes;
}

QList<Node*> Node::receiversFor(const QString &prop) const
{
    QList<Node*> nodes;
    foreach(Node::Connection con, d_ptr->receivers.values(prop)) {
        nodes.append(con.node);
    }
    return nodes;
}

Node * Node::senderTo(const QString &prop) const
{
    return d_ptr->senders[prop].node;
}


bool Node::hasInputs() const
{
    return !d_ptr->senders.isEmpty();
}

bool Node::hasOutputs() const
{
    return !d_ptr->receivers.isEmpty();
}

void Node::prepare()
{
}

void Node::propertyChanged(const QString &name,
                           const QVariant &val)
{
    Q_UNUSED(name);
    Q_UNUSED(val);
}

}


#include "moc_node.cpp"
