#include "texture.h"

#include <QGLContext>
#include <QByteArray>
#include <QSharedData>
#include <qdebug.h>

namespace Quasar
{

class Texture::Private : public QSharedData
{
public:
    inline Private()
        : texture(0),
          width(0),
          height(0),
          format(0)
    {
    }
    inline Private(const Private &other)
        : QSharedData(other),
          texture(other.texture),
          width(other.width),
          height(other.height),
          format(other.format)
    {
    }
    inline ~Private()
    {
        //TODO: if we own the texture we should free it
        //glDeleteTextures(1, &texture);
    }

    /* it's worth noting that GL_BGRA vs GL_RGBA for f makes
     * a big difference. due to GDI+ hardware usually uses
     * GL_BGRA as a native format so GL_RGBA needs to be
     * swizzled */
    void init(const char *data, int w, int h, GLenum f)
    {
        width = w;
        height = h;
        format = GL_RGBA;

        glGenTextures(1, &texture);
        glBindTexture(GL_TEXTURE_2D, texture);
        glTexImage2D(GL_TEXTURE_2D, 0, format,
                     width, height, 0, f,
                     GL_UNSIGNED_BYTE, data);
    }

    inline void init (const QByteArray &data, int w, int h, GLenum f)
    {
        init(data.constData(), w, h, f);
    }

    inline bool isNull() const
    {
        return texture == 0;
    }

    inline void update(int x, int y, int w, int h, const char *data, GLenum f)
    {
        glBindTexture(GL_TEXTURE_2D, texture);
        glTexSubImage2D(GL_TEXTURE_2D, 0,
                        x, y, w, h,
                        f, GL_UNSIGNED_BYTE, data);
    }
    inline void update(int x, int y, int w, int h,
                       const QByteArray &data, GLenum f)
    {
        update(x, y, w, h, data.constData(), f);
    }


public:
    GLuint  texture;
    GLint   width;
    GLint   height;
    GLenum  format;
};

Texture::Texture()
    : d(new Private)
{
}

Texture::Texture(GLint id)
    :d(new Private)
{
    d->texture = id;
    glBindTexture(GL_TEXTURE_2D, d->texture);
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0,
                             GL_TEXTURE_WIDTH,&d->width);
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0,
                             GL_TEXTURE_HEIGHT,&d->height);
    glGetTexLevelParameteriv(GL_TEXTURE_2D, 0,
                             GL_TEXTURE_INTERNAL_FORMAT,(GLint*)&d->format);
}

Texture::Texture(const QByteArray &data,
                 int w, int h,
                 GLenum f)
    : d(new Private)
{
    d->init(data, w, h, f);
}

Texture::Texture(const char *data,
                 int w, int h, GLenum f)
    : d(new Private)
{
    d->init(data, w, h, f);
}

bool Texture::isNull() const
{
    return d->isNull();
}

GLuint Texture::id() const
{
    return d->texture;
}

int Texture::width() const
{
    return d->width;
}

int Texture::height() const
{
    return d->height;
}

GLenum Texture::format() const
{
    return d->format;
}

Texture::~Texture()
{
    d = 0;
}

Texture::Texture(const Texture &other)
{
    d = other.d;
}

Texture & Texture::operator=(const Texture &other)
{
    d = other.d;
    return *this;
}


Texture::operator QVariant() const
{
    return QVariant::fromValue(*this);
}

void Texture::updateRect(const QRect &rect, const QByteArray &data,
                         GLenum f)
{
    d->update(rect.x(), rect.y(), rect.width(), rect.height(),
              data, f);
}

void Texture::updateRect(const QRect &rect, const char *data,
                         GLenum f)
{
    d->update(rect.x(), rect.y(), rect.width(), rect.height(),
              data, f);
}


QImage Texture::toImage() const
{
    QImage::Format imageFormat = QImage::Format_RGB32;
    if (QGLContext::currentContext()->format().alpha())
        imageFormat = QImage::Format_ARGB32_Premultiplied;
    QImage img(width(), height(), imageFormat);
    glBindTexture(GL_TEXTURE_2D, id());
    glGetTexImage(GL_TEXTURE_2D, 0, format(), GL_UNSIGNED_BYTE, img.bits());
    if (QSysInfo::ByteOrder == QSysInfo::BigEndian) {
        // OpenGL gives RGBA; Qt wants ARGB
        uint *p = (uint*)img.bits();
        uint *end = p + width()*height();
        while (p < end) {
            uint a = *p << 24;
            *p = (*p >> 8) | a;
            p++;
        }
    } else {
        // OpenGL gives ABGR (i.e. RGBA backwards); Qt wants ARGB
        img = img.rgbSwapped();
    }
    return img.mirrored();
}

}
