#define GL_GLEXT_PROTOTYPES
#include "convolutionnode_p.h"

#include <QColor>
#include <QVariant>
#include <QDebug>

#include <GL/glext.h>
#include <GL/gl.h>

using namespace Quasar;


static const char *vertShaderText =
    "void main() {\n"
    "   gl_FrontColor = gl_Color;\n"
    "   gl_TexCoord[0] = gl_MultiTexCoord0;\n"
    "   gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
    "}\n";

static const char *fragShaderText =
    "//max kernel size supported\n"
    "const int KernelSize = 9;\n"
    "//texture offsets\n"
    "uniform vec2 Offset[KernelSize];\n"
    "//convolution kernel\n"
    "uniform vec4 KernelValue[KernelSize];\n"
    "uniform sampler2D srcTex;\n"
    "uniform vec4 ScaleFactor;\n"
    "uniform vec4 BaseColor;\n"
    "void main(void)\n"
    "{\n"
    "    int i;\n"
    "    vec4 sum = vec4(0.0);\n"
    "    for (i = 0; i < KernelSize; ++i) {\n"
    "        vec4 tmp = texture2D(srcTex, gl_TexCoord[0].st + Offset[i]);\n"
    "        sum += tmp * KernelValue[i];\n"
    "    }\n"
    "    gl_FragColor = sum * ScaleFactor + BaseColor;\n"
    "}\n";

ConvolutionNode::ConvolutionNode(const QString &name)
    : FilterNode(name),
      m_kernelSize(9)
{
    setShaderSources(vertShaderText, fragShaderText);
}

QRectF ConvolutionNode::domainOfDefinition(const QRectF &rect) const
{
    return rect;
}

QRectF ConvolutionNode::regionOfInterest(const QRectF &rect) const
{
    return rect;
}

void ConvolutionNode::aboutToExecute()
{
}

void ConvolutionNode::programChanged(GLuint prog)
{
    glUseProgram(prog);
    GLuint texLoc = glGetUniformLocationARB(prog, "srcTex");
    glUniform1iARB(texLoc, 0);
    setupConvolution();
}

void ConvolutionNode::setupConvolution()
{
    GLint *kernel = (GLint*)malloc(sizeof(GLint) * m_kernelSize);
    GLfloat scale;
    QColor baseColor = QColor(0, 0, 0, 0);
    fillConvolution(kernel, &scale, &baseColor);
    //vector of 4
    GLfloat *vecKer = (GLfloat*)malloc(sizeof(GLfloat) * m_kernelSize * 4);
    for (GLuint i = 0; i < m_kernelSize; ++i) {
        vecKer[i*4 + 0] = kernel[i];
        vecKer[i*4 + 1] = kernel[i];
        vecKer[i*4 + 2] = kernel[i];
        vecKer[i*4 + 3] = kernel[i];
    }

    GLuint loc = glGetUniformLocationARB(programId(), "KernelValue");
    glUniform4fv(loc, m_kernelSize, vecKer);

    loc = glGetUniformLocationARB(programId(), "ScaleFactor");
    glUniform4f(loc, scale, scale, scale, scale);

    loc = glGetUniformLocationARB(programId(), "BaseColor");
    glUniform4f(loc, baseColor.redF(), baseColor.greenF(),
                baseColor.blueF(), baseColor.alphaF());

    free(vecKer);
    free(kernel);
}

void ConvolutionNode::setupOffsets(const Texture &tex)
{
    //FIXME: only kernel of size 9 supported
    Q_ASSERT(m_kernelSize == 9);
    float offsets[] = { 1.0 / tex.width(),  1.0 / tex.height(),
                        0.0              ,  1.0 / tex.height(),
                        -1.0 / tex.width(),  1.0 / tex.height(),
                        1.0 / tex.width(),  0.0,
                        0.0              ,  0.0,
                        -1.0 / tex.width(),  0.0,
                        1.0 / tex.width(), -1.0 / tex.height(),
                        0.0              , -1.0 / tex.height(),
                        -1.0 / tex.width(), -1.0 / tex.height() };
    GLuint offsetLoc = glGetUniformLocationARB(programId(), "Offset");
    glUniform2fv(offsetLoc, m_kernelSize, offsets);
}
