/*
    This file is part of the KDE project.
    Copyright (c) 2005 Friedrich W. H. Kossebau <Friedrich.W.H@Kossebau.de>
    Based upon several kicker files.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License version 2 as published by the Free Software Foundation.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301  USA
*/

#include <kdebug.h>

// kde specific
#include <kiconloader.h>
#include <klocale.h>
#include <kicontheme.h>
#include <kpopupmenu.h>
// applet specific
#include "kuserface.h"
#include "ksessionmenu.h"
#include "ksessionmenu.moc"


static const int LockOnSwitchId = 50;
static const int NewSessionId = 100;
static const int BaseSessionId = NewSessionId + 1;

KSessionMenu::KSessionMenu( QWidget* Parent, const char* Name )
: QPopupMenu( Parent, Name ),
  SessionManager( kapp->desktop()->screen(kapp->desktop()->screenNumber( this )), qt_xscreen() )
{
  connect( this, SIGNAL(aboutToShow()), SLOT(populateMenu()) );
  connect( this, SIGNAL(activated(int)), SLOT(slotSessionActivated(int)) );
}


void KSessionMenu::populateMenu()
{
  clear();

  if( SessionManager.canSwitchSession() )
  {
    // other sessions
    KSessionList Sessions = SessionManager.localSessions();
    if( Sessions.size()>0 )
    {
      for( KSessionList::ConstIterator it = Sessions.begin(); it != Sessions.end(); ++it )
        insertSessionItem( *it );

      // anything added?
      if( count()>0 )
        insertSeparator();
    }

    // new session
    int p;
    if( SessionManager.canCreateSession() && (p = SessionManager.numberOfReservedSessions()) >= 0 )
    {
      insertItem( SmallIconSet("fork"), i18n("Log In..."), NewSessionId ); // KDE 4 icon name: probably system-switch-user (?), or maybe go-jump... but there is no real "log in" icon
      if( p == 0 )
        setItemEnabled( NewSessionId, false );
      insertSeparator();
    }

    {
      // add section title
      KPopupTitle *Title = new KPopupTitle(this);
      Title->setTitle(i18n("Other Sessions"));
      insertItem( Title, -1, 0 );

      if( SessionManager.canLockSession() )
      {
        // add toggle for locking
        insertItem( i18n("Lock Current Session on Switch"), this, SLOT(slotToggleLockOnSwitch()), 0, LockOnSwitchId );
        setItemChecked( LockOnSwitchId, SessionManager.doesLockOnSwitch() );
      }
    }
  }

  int CountBefore = count();

  // save session TODO: does not exist in KRootWM!!!!
  if( SessionManager.canSaveSession() )
     insertItem( SmallIconSet("filesave"), i18n("Save Session"), &SessionManager, SLOT(saveSession()) );
  // lock session
  if( SessionManager.canLockSession() )
      insertItem( SmallIconSet("lock"), i18n("Lock Session"), &SessionManager, SLOT(lockSession()) );
  // end session
  if( SessionManager.canLogoutSession() )
      insertItem( SmallIconSet("exit"), i18n("Log Out..."), &SessionManager, SLOT(logoutSession()) );

  // anything added?
  if( count()>CountBefore )
  {
    // add section title
    KPopupTitle *Title = new KPopupTitle(this);
    Title->setTitle( i18n("Current Session") );
    insertItem( Title,-1,CountBefore );
  }
}


void KSessionMenu::slotToggleLockOnSwitch()
{
  bool NewState = !SessionManager.doesLockOnSwitch();
  SessionManager.setLockOnSwitch( NewState );
  setItemChecked( LockOnSwitchId, NewState );
}


void KSessionMenu::slotSessionActivated( int EntryId )
{
  if( EntryId == NewSessionId )
    SessionManager.createSession();
  else if( EntryId>=BaseSessionId /*&& !isItemChecked(EntryId)*/ )
    SessionManager.switchToSession( EntryId-BaseSessionId );
}

/*
Different listings:
a) simple one:
only other graphical sessions, show as users
b)

TODO: make lock when switching a check item, save setting globally
*/

static QString renderSessionDescription( KUser User, const KSession &Session, const QString &Pattern )
{
  QString Description;
  int l = Pattern.length();
  for( int i = 0; i < l; ++i )
  {
    if (Pattern[i] == '%')
    {
      switch (Pattern[++i].latin1())
      {
        //case 'U': Description += User.displayedName(); break;
        case 'F': Description += User.fullName(); break;
        case 'L': Description += User.loginName(); break;
        case 'S': Description += Session.Session; break;
        case 'D': Description += Session.DisplayName; break;
        case 'V': Description += Session.VTId; break;
        case '%': ;
        default:  Description += Pattern[i];
      }
    }
    else
      Description += Pattern[i];
  }
  return Description;
}


static bool ListPlainTerminals = false;
static bool ListCurrent = false;
static const QString DefaultLocalPattern("%F");

void KSessionMenu::insertSessionItem( const KSession &Session )
{
  // ignore ttys and current session
  if( Session.IsTty )
    return;
  if( !ListCurrent && Session.Current )
    return;

  int VTId = Session.VTId;
  QString Text;
  QIconSet IconSet;
  KUser User( Session.LoginName );
  if( User.isValid() )
  {
    Text = renderSessionDescription( User, Session, DefaultLocalPattern );
//     Text = i18n("fullname (loginname): session type", "%1 (%2): %3").arg( User.fullName(), User.loginName(), Session.Session );

//     QString Location = VTId ? QString("%1, vt%2").arg( Session.DisplayName ).arg( VTId ) :
//                     Session.DisplayName;
//     QString Text = i18n( "%1: Terminal on %2" ).arg( UserName, Location );

    IconSet = KUserFaceLoader::self()->loadFaceIconSet( User, KIcon::Small );
  }
  // unknown user
  else
  {
    // TODO: enable pattern here, too
    Text = Session.Session.isEmpty() ?     i18n("Unused") :
                   Session.Session == "<remote>" ? i18n("X login on remote host") : // TODO: which?
                   i18n("X login on %1").arg( Session.Session );
    //QString Location = VTId ? QString("%1, vt%2").arg( Session.DisplayName ).arg( VTId ) :
    //                   Session.DisplayName;
    //QString Text = i18n( "%1: Terminal on %2" ).arg( UserName, Location );

    IconSet = SmallIconSet( "x" );
  }

  int Id = insertItem( IconSet, Text, BaseSessionId+VTId );

  // TODO: in which cases can vt be 0?
  if( VTId == 0 )
    setItemEnabled( Id, false );
}
